/* GSequencer - Advanced GTK Sequencer
 * Copyright (C) 2005-2019 Joël Krähemann
 *
 * This file is part of GSequencer.
 *
 * GSequencer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GSequencer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GSequencer.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ags/audio/ags_audio.h>

#include <ags/libags.h>

#include <ags/audio/ags_preset.h>
#include <ags/audio/ags_notation.h>
#include <ags/audio/ags_automation.h>
#include <ags/audio/ags_wave.h>
#include <ags/audio/ags_midi.h>
#include <ags/audio/ags_pattern.h>
#include <ags/audio/ags_synth_generator.h>
#include <ags/audio/ags_output.h>
#include <ags/audio/ags_input.h>
#include <ags/audio/ags_playback_domain.h>
#include <ags/audio/ags_playback.h>
#include <ags/audio/ags_recall_container.h>
#include <ags/audio/ags_recall.h>
#include <ags/audio/ags_recall_audio.h>
#include <ags/audio/ags_recall_audio_run.h>
#include <ags/audio/ags_recall_id.h>

#include <ags/audio/thread/ags_audio_loop.h>
#include <ags/audio/thread/ags_audio_thread.h>

#include <ags/audio/file/ags_audio_file_link.h>
#include <ags/audio/file/ags_audio_file.h>

#include <ags/audio/midi/ags_midi_file.h>

#include <ags/audio/recall/ags_count_beats_audio_run.h>

#include <libxml/tree.h>

#include <stdlib.h>
#include <string.h>

#include <math.h>

#include <ags/i18n.h>

/**
 * SECTION:ags_audio
 * @short_description: A container of channels organizing them as input or output
 * @title: AgsAudio
 * @section_id:
 * @include: ags/audio/ags_audio.h
 *
 * #AgsAudio organizes #AgsChannel objects either as input or output and
 * is responsible of their alignment. The class can contain #AgsRecall objects
 * in order to perform computation on all channels or in audio context.
 * Therefor exists #AgsRecyclingContext acting as tree context.
 *
 * At least one #AgsRecallID is assigned to it and has one more if
 * %AGS_AUDIO_OUTPUT_HAS_RECYCLING is set as flag.
 *
 * If %AGS_AUDIO_HAS_NOTATION is set as flag one #AgsNotation is allocated per audio
 * channel.
 */

void ags_audio_class_init(AgsAudioClass *audio_class);
void ags_audio_connectable_interface_init(AgsConnectableInterface *connectable);
void ags_audio_init(AgsAudio *audio);
void ags_audio_set_property(GObject *gobject,
			    guint prop_id,
			    const GValue *value,
			    GParamSpec *param_spec);
void ags_audio_get_property(GObject *gobject,
			    guint prop_id,
			    GValue *value,
			    GParamSpec *param_spec);
void ags_audio_dispose(GObject *gobject);
void ags_audio_finalize(GObject *gobject);

AgsUUID* ags_audio_get_uuid(AgsConnectable *connectable);
gboolean ags_audio_has_resource(AgsConnectable *connectable);
gboolean ags_audio_is_ready(AgsConnectable *connectable);
void ags_audio_add_to_registry(AgsConnectable *connectable);
void ags_audio_remove_from_registry(AgsConnectable *connectable);
xmlNode* ags_audio_list_resource(AgsConnectable *connectable);
xmlNode* ags_audio_xml_compose(AgsConnectable *connectable);
void ags_audio_xml_parse(AgsConnectable *connectable,
			 xmlNode *node);
gboolean ags_audio_is_connected(AgsConnectable *connectable);
void ags_audio_connect(AgsConnectable *connectable);
void ags_audio_disconnect(AgsConnectable *connectable);

void ags_audio_real_set_audio_channels(AgsAudio *audio,
				       guint audio_channels, guint audio_channels_old);
void ags_audio_real_set_pads(AgsAudio *audio,
			     GType type,
			     guint channels, guint channels_old);

void ags_audio_real_duplicate_recall(AgsAudio *audio,
				     AgsRecallID *recall_id,
				     guint pad, guint audio_channel,
				     guint line);
void ags_audio_real_resolve_recall(AgsAudio *audio,
				   AgsRecallID *recall_id);

void ags_audio_real_init_recall(AgsAudio *audio,
				AgsRecallID *recall_id, guint staging_flags);
void ags_audio_real_play_recall(AgsAudio *audio,
				AgsRecallID *recall_id, guint staging_flags);

void ags_audio_real_done_recall(AgsAudio *audio,
				AgsRecallID *recall_id);
void ags_audio_real_cancel_recall(AgsAudio *audio,
				  AgsRecallID *recall_id);

void ags_audio_real_cleanup_recall(AgsAudio *audio,
				   AgsRecallID *recall_id);

void ags_audio_recall_done_callback(AgsRecall *recall,
				    AgsAudio *audio);

GList* ags_audio_real_start(AgsAudio *audio,
			    gint sound_scope);
void ags_audio_real_stop(AgsAudio *audio,
			 GList *recall_id, gint sound_scope);

GList* ags_audio_real_check_scope(AgsAudio *audio, gint sound_scope);

void ags_audio_real_recursive_run_stage(AgsAudio *audio,
					gint sound_scope, guint stage);

enum{
  SET_AUDIO_CHANNELS,
  SET_PADS,
  DUPLICATE_RECALL,
  RESOLVE_RECALL,
  INIT_RECALL,
  PLAY_RECALL,
  DONE_RECALL,
  CANCEL_RECALL,
  CLEANUP_RECALL,
  START,
  STOP,
  CHECK_SCOPE,
  RECURSIVE_RUN_STAGE,
  LAST_SIGNAL,
};

enum{
  PROP_0,
  PROP_AUDIO_NAME,
  PROP_OUTPUT_SOUNDCARD,
  PROP_INPUT_SOUNDCARD,
  PROP_OUTPUT_SEQUENCER,
  PROP_INPUT_SEQUENCER,
  PROP_SAMPLERATE,
  PROP_BUFFER_SIZE,
  PROP_FORMAT,
  PROP_BPM,
  PROP_MIN_AUDIO_CHANNELS,
  PROP_MAX_AUDIO_CHANNELS,
  PROP_MIN_OUTPUT_PADS,
  PROP_MAX_OUTPUT_PADS,
  PROP_MIN_INPUT_PADS,
  PROP_MAX_INPUT_PADS,
  PROP_AUDIO_CHANNELS,
  PROP_OUTPUT_PADS,
  PROP_OUTPUT_LINES,
  PROP_INPUT_PADS,
  PROP_INPUT_LINES,
  PROP_AUDIO_START_MAPPING,
  PROP_AUDIO_END_MAPPING,
  PROP_MIDI_START_MAPPING,
  PROP_MIDI_END_MAPPING,
  PROP_MIDI_CHANNEL,
  PROP_NUMERATOR,
  PROP_DENOMINATOR,
  PROP_TIME_SIGNATURE,
  PROP_IS_MINOR,
  PROP_SHARP_FLATS,
  PROP_OCTAVE,
  PROP_KEY,
  PROP_ABSOLUTE_KEY,
  PROP_LOOP_START,
  PROP_LOOP_END,
  PROP_OFFSET,
  PROP_OUTPUT,
  PROP_INPUT,
  PROP_PRESET,
  PROP_SYNTH_GENERATOR,
  PROP_PLAYBACK_DOMAIN,
  PROP_CURSOR,
  PROP_NOTATION,
  PROP_AUTOMATION,
  PROP_WAVE,
  PROP_OUTPUT_AUDIO_FILE,
  PROP_INPUT_AUDIO_FILE,
  PROP_MIDI,
  PROP_OUTPUT_MIDI_FILE,
  PROP_INPUT_MIDI_FILE,
  PROP_RECALL_ID,
  PROP_RECYCLING_CONTEXT,
  PROP_RECALL_CONTAINER,
  PROP_PLAY,
  PROP_RECALL,
};

static gpointer ags_audio_parent_class = NULL;
static guint audio_signals[LAST_SIGNAL];

static pthread_mutex_t ags_audio_class_mutex = PTHREAD_MUTEX_INITIALIZER;

GType
ags_audio_get_type (void)
{
  static volatile gsize g_define_type_id__volatile = 0;

  if(g_once_init_enter (&g_define_type_id__volatile)){
    GType ags_type_audio = 0;

    static const GTypeInfo ags_audio_info = {
      sizeof(AgsAudioClass),
      NULL, /* base_init */
      NULL, /* base_finalize */
      (GClassInitFunc) ags_audio_class_init,
      NULL, /* class_finalize */
      NULL, /* class_data */
      sizeof(AgsAudio),
      0,    /* n_preallocs */
      (GInstanceInitFunc) ags_audio_init,
    };

    static const GInterfaceInfo ags_connectable_interface_info = {
      (GInterfaceInitFunc) ags_audio_connectable_interface_init,
      NULL, /* interface_finalize */
      NULL, /* interface_data */
    };

    ags_type_audio = g_type_register_static(G_TYPE_OBJECT,
					    "AgsAudio", &ags_audio_info,
					    0);

    g_type_add_interface_static(ags_type_audio,
				AGS_TYPE_CONNECTABLE,
				&ags_connectable_interface_info);

    g_once_init_leave(&g_define_type_id__volatile, ags_type_audio);
  }

  return g_define_type_id__volatile;
}

void
ags_audio_class_init(AgsAudioClass *audio)
{
  GObjectClass *gobject;
  GParamSpec *param_spec;

  ags_audio_parent_class = g_type_class_peek_parent(audio);

  /* GObjectClass */
  gobject = (GObjectClass *) audio;

  gobject->set_property = ags_audio_set_property;
  gobject->get_property = ags_audio_get_property;

  gobject->dispose = ags_audio_dispose;
  gobject->finalize = ags_audio_finalize;

  /* properties */
  /**
   * AgsAudio:audio-name:
   *
   * The name of audio object.
   * 
   * Since: 2.1.0
   */
  param_spec = g_param_spec_string("audio-name",
				   i18n_pspec("assigned name"),
				   i18n_pspec("The name of audio"),
				   NULL,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_AUDIO_NAME,
				  param_spec);

  /**
   * AgsAudio:output-soundcard:
   *
   * The assigned #AgsSoundcard acting as default sink.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("output-soundcard",
				   i18n_pspec("assigned output soundcard"),
				   i18n_pspec("The output soundcard it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_OUTPUT_SOUNDCARD,
				  param_spec);

  /**
   * AgsAudio:input-soundcard:
   *
   * The assigned #AgsSoundcard acting as default source.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("input-soundcard",
				   i18n_pspec("assigned input soundcard"),
				   i18n_pspec("The input soundcard it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_INPUT_SOUNDCARD,
				  param_spec);
  
  /**
   * AgsAudio:output-sequencer:
   *
   * The assigned #AgsSequencer acting as default source.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("output-sequencer",
				   i18n_pspec("assigned output sequencer"),
				   i18n_pspec("The output sequencer it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_OUTPUT_SEQUENCER,
				  param_spec);

  /**
   * AgsAudio:input-sequencer:
   *
   * The assigned #AgsSequencer acting as default source.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("input-sequencer",
				   i18n_pspec("assigned input sequencer"),
				   i18n_pspec("The input sequencer it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_INPUT_SEQUENCER,
				  param_spec);
  
  /**
   * AgsAudio:samplerate:
   *
   * The samplerate.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("samplerate",
				 i18n_pspec("samplerate"),
				 i18n_pspec("The samplerate"),
				 0,
				 G_MAXUINT32,
				 AGS_SOUNDCARD_DEFAULT_SAMPLERATE,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_SAMPLERATE,
				  param_spec);

  /**
   * AgsAudio:buffer-size:
   *
   * The buffer length.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("buffer-size",
				 i18n_pspec("buffer size"),
				 i18n_pspec("The buffer size"),
				 0,
				 G_MAXUINT32,
				 AGS_SOUNDCARD_DEFAULT_BUFFER_SIZE,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_BUFFER_SIZE,
				  param_spec);

  /**
   * AgsAudio:format:
   *
   * The format.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("format",
				 i18n_pspec("format"),
				 i18n_pspec("The format"),
				 0,
				 G_MAXUINT32,
				 AGS_SOUNDCARD_DEFAULT_FORMAT,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_FORMAT,
				  param_spec);

  /**
   * AgsAudio:bpm:
   *
   * The bpm.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_double("bpm",
				   i18n_pspec("bpm"),
				   i18n_pspec("The bpm"),
				   0.0,
				   G_MAXDOUBLE,
				   AGS_SOUNDCARD_DEFAULT_BPM,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_BPM,
				  param_spec);

  /**
   * AgsAudio:min-audio-channels:
   *
   * The minimum audio channels count.
   * 
   * Since: 2.0.31
   */
  param_spec = g_param_spec_uint("min-audio-channels",
				 i18n_pspec("minimum audio channels count"),
				 i18n_pspec("The minimum count of audio channels of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MIN_AUDIO_CHANNELS,
				  param_spec);

  /**
   * AgsAudio:max-audio-channels:
   *
   * The maximum audio channels count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("max-audio-channels",
				 i18n_pspec("maximum audio channels count"),
				 i18n_pspec("The maximum count of audio channels of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MAX_AUDIO_CHANNELS,
				  param_spec);

  /**
   * AgsAudio:min-output-pads:
   *
   * The minimum output pads count.
   * 
   * Since: 2.0.31
   */
  param_spec = g_param_spec_uint("min-output-pads",
				 i18n_pspec("minimum output pads count"),
				 i18n_pspec("The minimum count of output pads of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MIN_OUTPUT_PADS,
				  param_spec);

  /**
   * AgsAudio:max-output-pads:
   *
   * The maximum output pads count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("max-output-pads",
				 i18n_pspec("maximum output pads count"),
				 i18n_pspec("The maximum count of output pads of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MAX_OUTPUT_PADS,
				  param_spec);

  /**
   * AgsAudio:min-input-pads:
   *
   * The minimum input pads count.
   * 
   * Since: 2.0.31
   */
  param_spec = g_param_spec_uint("min-input-pads",
				 i18n_pspec("minimum input pads count"),
				 i18n_pspec("The minimum count of input pads of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MIN_INPUT_PADS,
				  param_spec);

  /**
   * AgsAudio:max-input-pads:
   *
   * The maximum input pads count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("max-input-pads",
				 i18n_pspec("maximum input pads count"),
				 i18n_pspec("The maximum count of input pads of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MAX_INPUT_PADS,
				  param_spec);
  
  /**
   * AgsAudio:audio-channels:
   *
   * The audio channels count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("audio-channels",
				 i18n_pspec("audio channels count"),
				 i18n_pspec("The count of audio channels of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_AUDIO_CHANNELS,
				  param_spec);

  /**
   * AgsAudio:output-pads:
   *
   * The output pads count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("output-pads",
				 i18n_pspec("output pads count"),
				 i18n_pspec("The count of output pads of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_OUTPUT_PADS,
				  param_spec);

  /**
   * AgsAudio:output-lines:
   *
   * The output lines count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("output-lines",
				 i18n_pspec("output lines count"),
				 i18n_pspec("The count of output lines of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE);
  g_object_class_install_property(gobject,
				  PROP_OUTPUT_LINES,
				  param_spec);

  /**
   * AgsAudio:input-pads:
   *
   * The input pads count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("input-pads",
				 i18n_pspec("input pads count"),
				 i18n_pspec("The count of input pads of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_INPUT_PADS,
				  param_spec);

  /**
   * AgsAudio:input-lines:
   *
   * The input lines count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("input-lines",
				 i18n_pspec("input lines count"),
				 i18n_pspec("The count of input lines of audio"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE);
  g_object_class_install_property(gobject,
				  PROP_INPUT_LINES,
				  param_spec);

  /**
   * AgsAudio:audio-start-mapping:
   *
   * The audio start mapping.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("audio-start-mapping",
				 i18n_pspec("audio start mapping"),
				 i18n_pspec("The audio start mapping"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_AUDIO_START_MAPPING,
				  param_spec);

  /**
   * AgsAudio:audio-end-mapping:
   *
   * The audio end mapping.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("audio-end-mapping",
				 i18n_pspec("audio end mapping"),
				 i18n_pspec("The audio end mapping"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_AUDIO_END_MAPPING,
				  param_spec);

  /**
   * AgsAudio:midi-start-mapping:
   *
   * The midi start mapping.
   * 
   * Since: 2.0.0
   */
  param_spec =  g_param_spec_uint("midi-start-mapping",
				  i18n_pspec("midi start mapping range"),
				  i18n_pspec("The midi mapping range's start"),
				  0,
				  G_MAXUINT32,
				  0,
				  G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MIDI_START_MAPPING,
				  param_spec);

  /**
   * AgsAudio:midi-end-mapping:
   *
   * The midi end mapping.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("midi-end-mapping",
				 i18n_pspec("midi end mapping range"),
				 i18n_pspec("The midi mapping range's start"),
				 0,
				 G_MAXUINT32,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MIDI_END_MAPPING,
				  param_spec);

  /**
   * AgsAudio:midi-channel:
   *
   * The midi channel.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("midi-channel",
				 i18n_pspec("midi channel"),
				 i18n_pspec("The midi channel"),
				 0,
				 16,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MIDI_CHANNEL,
				  param_spec);

  /**
   * AgsAudio:numerator:
   *
   * The numerator of time signature.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("numerator",
				 i18n_pspec("numerator"),
				 i18n_pspec("The numerator"),
				 0,
				 32,
				 4,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_NUMERATOR,
				  param_spec);
  
  /**
   * AgsAudio:denominator:
   *
   * The denominator of time signature.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("denominator",
				 i18n_pspec("denominator"),
				 i18n_pspec("The denominator"),
				 0,
				 32,
				 4,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_DENOMINATOR,
				  param_spec);

  /**
   * AgsAudio:time-signature:
   *
   * The time signature.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_string("time-signature",
				   i18n_pspec("time signature"),
				   i18n_pspec("The time signature"),
				   "4/4",
				   G_PARAM_READABLE);
  g_object_class_install_property(gobject,
				  PROP_TIME_SIGNATURE,
				  param_spec);

  /**
   * AgsAudio:is-minor:
   *
   * Is minor.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_boolean("is-minor",
				    i18n_pspec("is minor"),
				    i18n_pspec("Is minor"),
				    FALSE,
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_IS_MINOR,
				  param_spec);

  /**
   * AgsAudio:sharp-flats:
   *
   * The sharp/flats count.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("sharp-flats",
				 i18n_pspec("sharp/flats"),
				 i18n_pspec("The sharp/flats count"),
				 0,
				 12,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_SHARP_FLATS,
				  param_spec);
  
  /**
   * AgsAudio:octave:
   *
   * The octave lower.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_int("octave",
				i18n_pspec("octave"),
				i18n_pspec("The octave lower"),
				0,
				10,
				0,
				G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_OCTAVE,
				  param_spec);

  /**
   * AgsAudio:key:
   *
   * The key relative to octave.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_uint("key",
				 i18n_pspec("relative key"),
				 i18n_pspec("The key relative to octave"),
				 0,
				 12,
				 0,
				 G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_KEY,
				  param_spec);

  /**
   * AgsAudio:absolute-key:
   *
   * The absolute key lower.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_int("absolute-key",
				i18n_pspec("absolute key"),
				i18n_pspec("The absolute key lower"),
				0,
				128,
				0,
				G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_ABSOLUTE_KEY,
				  param_spec);


  /**
   * AgsAudio:loop-start:
   *
   * The audio's loop start.
   * 
   * Since: 2.0.0
   */
  param_spec =  g_param_spec_uint64("loop-start",
				    i18n_pspec("loop start of audio"),
				    i18n_pspec("The loop start of audio"),
				    0.0,
				    G_MAXUINT64,
				    0.0,
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_LOOP_START,
				  param_spec);

  /**
   * AgsAudio:loop-end:
   *
   * The audio's loop end.
   * 
   * Since: 2.0.0
   */
  param_spec =  g_param_spec_uint64("loop-end",
				    i18n_pspec("loop end of audio"),
				    i18n_pspec("The loop end of audio"),
				    0.0,
				    G_MAXUINT64,
				    0.0,
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_LOOP_END,
				  param_spec);

  /**
   * AgsAudio:offset:
   *
   * The audio's offset.
   * 
   * Since: 2.0.0
   */
  param_spec =  g_param_spec_uint64("offset",
				    i18n_pspec("offset of audio"),
				    i18n_pspec("The offset of audio"),
				    0.0,
				    G_MAXUINT64,
				    0.0,
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_OFFSET,
				  param_spec);

  /**
   * AgsAudio:output:
   *
   * The #AgsOutput it contains.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("output",
				   i18n_pspec("containing output"),
				   i18n_pspec("The output it contains"),
				   AGS_TYPE_OUTPUT,
				   G_PARAM_READABLE);
  g_object_class_install_property(gobject,
				  PROP_OUTPUT,
				  param_spec);

  /**
   * AgsAudio:input:
   *
   * The #AgsInput it contains.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("input",
				   i18n_pspec("containing input"),
				   i18n_pspec("The input it contains"),
				   AGS_TYPE_INPUT,
				   G_PARAM_READABLE);
  g_object_class_install_property(gobject,
				  PROP_INPUT,
				  param_spec);

  /**
   * AgsAudio:preset:
   *
   * The assigned #GList-struct containing #AgsPreset information.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("preset",
				    i18n_pspec("preset"),
				    i18n_pspec("The preset"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_PRESET,
				  param_spec);

  /**
   * AgsAudio:synth_generator:
   *
   * The assigned #GList-struct containing #AgsSynthGenerator information.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("synth-generator",
				    i18n_pspec("synth generator"),
				    i18n_pspec("The synth generator"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_SYNTH_GENERATOR,
				  param_spec);
  
  /**
   * AgsAudio:playback-domain:
   *
   * The assigned #AgsPlaybackDomain.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("playback-domain",
				   i18n_pspec("assigned playback domain"),
				   i18n_pspec("The assigned playback domain"),
				   AGS_TYPE_PLAYBACK_DOMAIN,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_PLAYBACK_DOMAIN,
				  param_spec);

  /**
   * AgsAudio:cursor:
   *
   * The #GObject implementing #AgsCursor interface.
   * 
   * Since: 2.2.0
   */
  param_spec = g_param_spec_pointer("cursor",
				    i18n_pspec("cursor"),
				    i18n_pspec("The cursor object"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_CURSOR,
				  param_spec);

  /**
   * AgsAudio:notation:
   *
   * The #AgsNotation it contains.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("notation",
				    i18n_pspec("containing notation"),
				    i18n_pspec("The notation it contains"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_NOTATION,
				  param_spec);

  /**
   * AgsAudio:automation:
   *
   * The #AgsAutomation it contains.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("automation",
				    i18n_pspec("containing automation"),
				    i18n_pspec("The automation it contains"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_AUTOMATION,
				  param_spec);
  
  /**
   * AgsAudio:wave:
   *
   * The #AgsWave it contains.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("wave",
				    i18n_pspec("containing wave"),
				    i18n_pspec("The wave it contains"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_WAVE,
				  param_spec);

  /**
   * AgsAudio:output-audio-file:
   *
   * The assigned #AgsAudioFile acting as default sink.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("output-audio-file",
				   i18n_pspec("assigned output audio file"),
				   i18n_pspec("The output audio file it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_OUTPUT_AUDIO_FILE,
				  param_spec);
  
  /**
   * AgsAudio:input-audio-file:
   *
   * The assigned #AgsAudioFile acting as default sink.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("input-audio-file",
				   i18n_pspec("assigned input audio file"),
				   i18n_pspec("The input audio file it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_INPUT_AUDIO_FILE,
				  param_spec);

  /**
   * AgsAudio:midi:
   *
   * The #AgsMidi it contains.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("midi",
				    i18n_pspec("containing midi"),
				    i18n_pspec("The midi it contains"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_MIDI,
				  param_spec);
  
  /**
   * AgsAudio:output-midi-file:
   *
   * The assigned #AgsMidiFile acting as default sink.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("output-midi-file",
				   i18n_pspec("assigned output midi file"),
				   i18n_pspec("The output midi file it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_OUTPUT_MIDI_FILE,
				  param_spec);
  
  /**
   * AgsAudio:input-midi-file:
   *
   * The assigned #AgsMidiFile acting as default sink.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_object("input-midi-file",
				   i18n_pspec("assigned input midi file"),
				   i18n_pspec("The input midi file it is assigned with"),
				   G_TYPE_OBJECT,
				   G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_INPUT_MIDI_FILE,
				  param_spec);
  
  /**
   * AgsAudio:recall-id:
   *
   * The assigned #AgsRecallID.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("recall-id",
				    i18n_pspec("assigned recall id"),
				    i18n_pspec("The assigned recall id"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_RECALL_ID,
				  param_spec);

  /**
   * AgsAudio:recycling-context:
   *
   * The assigned #AgsRecyclingContext.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("recycling-context",
				    i18n_pspec("assigned recycling context"),
				    i18n_pspec("The assigned recall id"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_RECYCLING_CONTEXT,
				  param_spec);

  /**
   * AgsAudio:recall-container:
   *
   * The #AgsRecallContainer it contains in container-context.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("recall-container",
				    i18n_pspec("containing recall container"),
				    i18n_pspec("The recall container it contains"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_RECALL_CONTAINER,
				  param_spec);

  /**
   * AgsAudio:play:
   *
   * The #AgsRecall it contains in play-context.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("play",
				    i18n_pspec("containing play"),
				    i18n_pspec("The play it contains"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_PLAY,
				  param_spec);

  /**
   * AgsAudio:recall:
   *
   * The #AgsRecall it contains in recall-context.
   * 
   * Since: 2.0.0
   */
  param_spec = g_param_spec_pointer("recall",
				    i18n_pspec("containing recall"),
				    i18n_pspec("The recall it contains"),
				    G_PARAM_READABLE | G_PARAM_WRITABLE);
  g_object_class_install_property(gobject,
				  PROP_RECALL,
				  param_spec);

  
  /* AgsAudioClass */
  audio->set_audio_channels = ags_audio_real_set_audio_channels;
  audio->set_pads = ags_audio_real_set_pads;

  audio->duplicate_recall = ags_audio_real_duplicate_recall;
  audio->resolve_recall = ags_audio_real_resolve_recall;

  audio->init_recall = ags_audio_real_init_recall;
  audio->play_recall = ags_audio_real_play_recall;
  
  audio->done_recall = ags_audio_real_done_recall;
  audio->cancel_recall = ags_audio_real_cancel_recall;

  audio->cleanup_recall = ags_audio_real_cleanup_recall;

  audio->start = ags_audio_real_start;
  audio->stop = ags_audio_real_stop;

  audio->check_scope = ags_audio_real_check_scope;
  audio->recursive_run_stage = ags_audio_real_recursive_run_stage;
  
  /* signals */
  /**
   * AgsAudio::set-audio-channels:
   * @audio: the object to adjust the channels.
   * @audio_channels_new: new audio channel count
   * @audio_channels_old: old audio channel count
   *
   * The ::set-audio-channels signal notifies about changes in channel
   * alignment.
   *
   * Since: 2.0.0
   */
  audio_signals[SET_AUDIO_CHANNELS] = 
    g_signal_new("set-audio-channels",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, set_audio_channels),
		 NULL, NULL,
		 ags_cclosure_marshal_VOID__UINT_UINT,
		 G_TYPE_NONE, 2,
		 G_TYPE_UINT, G_TYPE_UINT);

  /**
   * AgsAudio::set-pads:
   * @audio: the object to adjust pads.
   * @channel_type: either #AGS_TYPE_INPUT or #AGS_TYPE_OUTPUT
   * @pads_new: new pad count
   * @pads_old: old pad count
   *
   * The ::set-pads signal notifies about changes in channel
   * alignment.
   *
   * Since: 2.0.0
   */
  audio_signals[SET_PADS] = 
    g_signal_new("set-pads",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, set_pads),
		 NULL, NULL,
		 ags_cclosure_marshal_VOID__ULONG_UINT_UINT,
		 G_TYPE_NONE, 3,
		 G_TYPE_ULONG,
		 G_TYPE_UINT, G_TYPE_UINT);

  /**
   * AgsAudio::duplicate-recall:
   * @audio: the #AgsAudio
   * @recall_id: the #AgsRecallID
   * @pad: the pad
   * @audio_channel: the audio channel
   * @line: the line
   *
   * The ::duplicate-recall signal notifies about duplicated recalls.
   *
   * Since: 2.0.0
   */
  audio_signals[DUPLICATE_RECALL] = 
    g_signal_new("duplicate-recall",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, duplicate_recall),
		 NULL, NULL,
		 ags_cclosure_marshal_VOID__OBJECT_UINT_UINT_UINT,
		 G_TYPE_NONE, 4,
		 G_TYPE_OBJECT,
		 G_TYPE_UINT, G_TYPE_UINT,
		 G_TYPE_UINT);

  /**
   * AgsAudio::resolve-recall:
   * @audio: the #AgsAudio
   * @recall_id: the #AgsRecallID
   *
   * The ::resolve-recall signal notifies about resolved recalls.
   *
   * Since: 2.0.0
   */
  audio_signals[RESOLVE_RECALL] = 
    g_signal_new("resolve-recall",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, resolve_recall),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 G_TYPE_OBJECT);

  /**
   * AgsAudio::init-recall:
   * @audio: the #AgsAudio
   * @recall_id: the #AgsRecallID
   * @staging_flags: the staging flags
   *
   * The ::init-recall signal notifies about initd recalls.
   *
   * Since: 2.0.0
   */
  audio_signals[INIT_RECALL] = 
    g_signal_new("init-recall",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, init_recall),
		 NULL, NULL,
		 ags_cclosure_marshal_VOID__OBJECT_UINT,
		 G_TYPE_NONE, 2,
		 G_TYPE_OBJECT,
		 G_TYPE_UINT);
 
  /**
   * AgsAudio::play-recall:
   * @audio: the #AgsAudio
   * @recall_id: the #AgsRecallID
   * @staging_flags: the staging flags
   *
   * The ::play-recall signal notifies about playd recalls.
   *
   * Since: 2.0.0
   */
  audio_signals[PLAY_RECALL] = 
    g_signal_new("play-recall",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, play_recall),
		 NULL, NULL,
		 ags_cclosure_marshal_VOID__OBJECT_UINT,
		 G_TYPE_NONE, 2,
		 G_TYPE_OBJECT,
		 G_TYPE_UINT);

  /**
   * AgsAudio::done-recall:
   * @audio: the #AgsAudio
   * @recall_id: the #AgsRecallID
   *
   * The ::done-recall signal notifies about doned recalls.
   *
   * Since: 2.0.0
   */
  audio_signals[DONE_RECALL] = 
    g_signal_new("done-recall",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, done_recall),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 G_TYPE_OBJECT);

  /**
   * AgsAudio::cancel-recall:
   * @audio: the #AgsAudio
   * @recall_id: the #AgsRecallID
   *
   * The ::cancel-recall signal notifies about canceld recalls.
   *
   * Since: 2.0.0
   */
  audio_signals[CANCEL_RECALL] = 
    g_signal_new("cancel-recall",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, cancel_recall),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 G_TYPE_OBJECT);

  /**
   * AgsAudio::cleanup-recall:
   * @audio: the #AgsAudio
   * @recall_id: the #AgsRecallID
   *
   * The ::cleanup-recall signal notifies about cleanup recalls.
   *
   * Since: 2.0.0
   */
  audio_signals[CLEANUP_RECALL] = 
    g_signal_new("cleanup-recall",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, cleanup_recall),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__OBJECT,
		 G_TYPE_NONE, 1,
		 G_TYPE_OBJECT);

  /**
   * AgsAudio::start:
   * @audio: the #AgsAudio starts playing
   * @sound_scope: the affected scope
   *
   * The ::start signal is invoked while starting playback
   * of @audio.
   *
   * Returns: the #GList-struct containing #AgsRecallID
   * 
   * Since: 2.0.0
   */
  audio_signals[START] = 
    g_signal_new("start",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, start),
		 NULL, NULL,
		 ags_cclosure_marshal_POINTER__INT,
		 G_TYPE_POINTER, 1,
		 G_TYPE_INT);

  /**
   * AgsAudio::stop:
   * @audio: the #AgsAudio stops playing
   * @recall_id: a #GList-struct containing #AgsRecallID
   * @sound_scope: the affected scope
   *
   * The ::stop signal is invoked while stoping playback
   * of @audio.
   *
   * Since: 2.0.0
   */
  audio_signals[STOP] = 
    g_signal_new("stop",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, stop),
		 NULL, NULL,
		 ags_cclosure_marshal_VOID__POINTER_INT,
		 G_TYPE_NONE, 2,
		 G_TYPE_POINTER,
		 G_TYPE_INT);

  /**
   * AgsAudio::check-scope:
   * @audio: the #AgsAudio to check the scopes
   * @sound_scope: the affected scope
   *
   * The ::check-scope method returns the appropriate recall id of @sound_scope.
   *
   * Returns: the #GList-struct containing #AgsRecallID
   * 
   * Since: 2.0.0
   */
  audio_signals[CHECK_SCOPE] = 
    g_signal_new("check-scope",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, check_scope),
		 NULL, NULL,
		 ags_cclosure_marshal_POINTER__INT,
		 G_TYPE_POINTER, 1,
		 G_TYPE_INT);

  /**
   * AgsAudio::recursive-run-stage:
   * @audio: the #AgsAudio to run
   * @sound_scope: the affected scope
   * @staging_flags: the flags to set
   *
   * The ::recursive-run-stage signal is invoked while run staging
   * of @audio for @sound_scope.
   * 
   * Since: 2.0.0
   */
  audio_signals[RECURSIVE_RUN_STAGE] = 
    g_signal_new("recursive-run-stage",
		 G_TYPE_FROM_CLASS(audio),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(AgsAudioClass, recursive_run_stage),
		 NULL, NULL,
		 ags_cclosure_marshal_VOID__INT_UINT,
		 G_TYPE_NONE, 2,
		 G_TYPE_INT,
		 G_TYPE_UINT);
}

void
ags_audio_connectable_interface_init(AgsConnectableInterface *connectable)
{
  connectable->get_uuid = ags_audio_get_uuid;
  connectable->has_resource = ags_audio_has_resource;

  connectable->is_ready = ags_audio_is_ready;
  connectable->add_to_registry = ags_audio_add_to_registry;
  connectable->remove_from_registry = ags_audio_remove_from_registry;

  connectable->list_resource = ags_audio_list_resource;
  connectable->xml_compose = ags_audio_xml_compose;
  connectable->xml_parse = ags_audio_xml_parse;

  connectable->is_connected = ags_audio_is_connected;
  connectable->connect = ags_audio_connect;
  connectable->disconnect = ags_audio_disconnect;

  connectable->connect_connection = NULL;
  connectable->disconnect_connection = NULL;
}

void
ags_audio_init(AgsAudio *audio)
{
  AgsConfig *config;
  
  gchar *str;
  gchar *str0, *str1;

  pthread_mutex_t *mutex;
  pthread_mutexattr_t *attr;

  audio->flags = 0;
  audio->ability_flags = 0;
  audio->behaviour_flags = 0;
  memset(audio->staging_flags, 0, (AGS_SOUND_SCOPE_LAST + 1) * sizeof(guint));

  /* add audio mutex */
  audio->obj_mutexattr = 
    attr = (pthread_mutexattr_t *) malloc(sizeof(pthread_mutexattr_t));
  pthread_mutexattr_init(attr);
  pthread_mutexattr_settype(attr,
			    PTHREAD_MUTEX_RECURSIVE);

#ifdef __linux__
  pthread_mutexattr_setprotocol(attr,
				PTHREAD_PRIO_INHERIT);
#endif

  audio->obj_mutex = 
    mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(mutex,
		     attr);

  /* uuid */
  audio->uuid = ags_uuid_alloc();
  ags_uuid_generate(audio->uuid);
  
  audio->audio_name = NULL;

  /* config */
  config = ags_config_get_instance();
  
  /* base init */
  audio->output_soundcard = NULL;
  audio->output_soundcard_channel_map = NULL;

  audio->input_soundcard = NULL;
  audio->input_soundcard_channel_map = NULL;

  audio->output_sequencer = NULL;
  audio->input_sequencer = NULL;
  
  audio->samplerate = ags_soundcard_helper_config_get_samplerate(config);
  audio->buffer_size = ags_soundcard_helper_config_get_buffer_size(config);
  audio->format = ags_soundcard_helper_config_get_format(config);

  audio->bpm = AGS_SOUNDCARD_DEFAULT_BPM;

  /* bank */
  audio->bank_dim[0] = 0;
  audio->bank_dim[1] = 0;
  audio->bank_dim[2] = 0;

  /* channel allocation */
  audio->max_audio_channels = G_MAXUINT32;
  audio->min_audio_channels = 0;

  audio->max_output_pads = G_MAXUINT32;
  audio->min_output_pads = 0;

  audio->max_input_pads = G_MAXUINT32;
  audio->min_input_pads = 0;

  audio->audio_channels = 0;

  audio->output_pads = 0;
  audio->output_lines = 0;
  
  audio->input_pads = 0;
  audio->input_lines = 0;
  
  /* midi mapping */
  audio->audio_start_mapping = 0;
  audio->audio_end_mapping = 0;

  audio->midi_start_mapping = 0;
  audio->midi_end_mapping = 0;

  audio->midi_channel = 0;

  /* time-signature */
  audio->numerator = 4;
  audio->denominator = 4;

  audio->time_signature = g_strdup("4/4");

  /* sharp/flats */
  audio->is_minor = FALSE;

  audio->sharp_flats = 0;

  /* octave */
  audio->octave = 0;
  audio->key = 0;
  
  audio->absolute_key = 0;

  /* loop */
  audio->loop_start = 0;
  audio->loop_end = 0;
  audio->offset = 0;
  
  /* channels */
  audio->output = NULL;
  audio->input = NULL;

  /* preset */
  audio->preset = NULL;

  /* playback domain */
  audio->playback_domain = (GObject *) ags_playback_domain_new((GObject *) audio);
  g_object_ref(audio->playback_domain);

  /* synth generator */
  audio->synth_generator = NULL;

  /*
   * Storage objects
   */
  /* cursor */
  audio->cursor = NULL;

  /* notation */
  audio->notation = NULL;

  /* automation */
  audio->automation = NULL;
  
  /* wave */
  audio->wave = NULL;

  audio->output_audio_file = NULL;
  audio->input_audio_file = NULL;

  /* midi */
  audio->midi = NULL;

  audio->output_midi_file = NULL;
  audio->input_midi_file = NULL;
  
  /* recycling context */
  audio->recall_id = NULL;
  audio->recycling_context = NULL;

  audio->recall_container = NULL;

  /* play context */
  audio->play_mutexattr = 
    attr = (pthread_mutexattr_t *) malloc(sizeof(pthread_mutexattr_t));
  pthread_mutexattr_init(attr);
  pthread_mutexattr_settype(attr,
			    PTHREAD_MUTEX_RECURSIVE);

  audio->play_mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(audio->play_mutex,
		     attr);

  audio->play = NULL;

  /* recall context */
  audio->recall_mutexattr = 
    attr = (pthread_mutexattr_t *) malloc(sizeof(pthread_mutexattr_t));
  pthread_mutexattr_init(attr);
  pthread_mutexattr_settype(attr,
			    PTHREAD_MUTEX_RECURSIVE);

  audio->recall_mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(audio->recall_mutex,
		     attr);

  audio->recall = NULL;

  /* data */
  audio->machine_widget = NULL;
}

void
ags_audio_set_property(GObject *gobject,
		       guint prop_id,
		       const GValue *value,
		       GParamSpec *param_spec)
{
  AgsAudio *audio;

  pthread_mutex_t *audio_mutex;

  audio = AGS_AUDIO(gobject);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  switch(prop_id){
  case PROP_AUDIO_NAME:
    {
      gchar *audio_name;

      audio_name = g_value_get_string(value);

      pthread_mutex_lock(audio_mutex);
      
      audio->audio_name = g_strdup(audio_name);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT_SOUNDCARD:
    {
      GObject *soundcard;

      soundcard = g_value_get_object(value);

      ags_audio_set_output_soundcard(audio,
				     (GObject *) soundcard);
    }
    break;
  case PROP_INPUT_SOUNDCARD:
    {
      GObject *soundcard;

      soundcard = g_value_get_object(value);

      ags_audio_set_input_soundcard(audio,
				    (GObject *) soundcard);
    }
    break;
  case PROP_OUTPUT_SEQUENCER:
    {
      GObject *sequencer;

      sequencer = g_value_get_object(value);

      ags_audio_set_output_sequencer(audio,
				     (GObject *) sequencer);
    }
    break;
  case PROP_INPUT_SEQUENCER:
    {
      GObject *sequencer;

      sequencer = g_value_get_object(value);

      ags_audio_set_input_sequencer(audio,
				    (GObject *) sequencer);
    }
    break;
  case PROP_SAMPLERATE:
    {
      guint samplerate;

      samplerate = g_value_get_uint(value);

      ags_audio_set_samplerate(audio,
			       samplerate);
    }
    break;
  case PROP_BUFFER_SIZE:
    {
      guint buffer_size;

      buffer_size = g_value_get_uint(value);

      ags_audio_set_buffer_size(audio,
				buffer_size);
    }
    break;
  case PROP_FORMAT:
    {
      guint format;

      format = g_value_get_uint(value);

      ags_audio_set_format(audio,
			   format);
    }
    break;
  case PROP_BPM:
    {
      gdouble bpm;

      bpm = g_value_get_double(value);

      pthread_mutex_lock(audio_mutex);
      
      audio->bpm = bpm;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIN_AUDIO_CHANNELS:
    {
      guint min_audio_channels;

      min_audio_channels = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);
      
      audio->min_audio_channels = min_audio_channels;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MAX_AUDIO_CHANNELS:
    {
      guint max_audio_channels;

      max_audio_channels = g_value_get_uint(value);

      ags_audio_set_max_audio_channels(audio,
				       max_audio_channels);
    }
    break;
  case PROP_MIN_OUTPUT_PADS:
    {
      guint min_output_pads;

      min_output_pads = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);
      
      audio->min_output_pads = min_output_pads;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MAX_OUTPUT_PADS:
    {
      guint max_output_pads;

      max_output_pads = g_value_get_uint(value);

      ags_audio_set_max_pads(audio,
			     AGS_TYPE_OUTPUT,
			     max_output_pads);
    }
    break;
  case PROP_MIN_INPUT_PADS:
    {
      guint min_input_pads;

      min_input_pads = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);
      
      audio->min_input_pads = min_input_pads;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MAX_INPUT_PADS:
    {
      guint max_input_pads;

      max_input_pads = g_value_get_uint(value);

      ags_audio_set_max_pads(audio,
			     AGS_TYPE_INPUT,
			     max_input_pads);
    }
    break;
  case PROP_AUDIO_CHANNELS:
    {
      guint audio_channels;

      audio_channels = g_value_get_uint(value);

      ags_audio_set_audio_channels(audio,
				   audio_channels, 0);
    }
    break;
  case PROP_OUTPUT_PADS:
    {
      guint output_pads;

      output_pads = g_value_get_uint(value);

      ags_audio_set_pads(audio,
			 AGS_TYPE_OUTPUT,
			 output_pads, 0);
    }
    break;
  case PROP_INPUT_PADS:
    {
      guint input_pads;

      input_pads = g_value_get_uint(value);

      ags_audio_set_pads(audio,
			 AGS_TYPE_INPUT,
			 input_pads, 0);
    }
    break;
  case PROP_AUDIO_START_MAPPING:
    {
      guint audio_start_mapping;

      audio_start_mapping = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);
      
      audio->audio_start_mapping = audio_start_mapping;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_AUDIO_END_MAPPING:
    {
      guint audio_end_mapping;

      audio_end_mapping = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);

      audio->audio_end_mapping = audio_end_mapping;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIDI_START_MAPPING:
    {
      guint midi_start_mapping;

      midi_start_mapping = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);

      audio->midi_start_mapping = midi_start_mapping;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIDI_END_MAPPING:
    {
      guint midi_end_mapping;

      midi_end_mapping = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);

      audio->midi_end_mapping = midi_end_mapping;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIDI_CHANNEL:
    {
      guint midi_channel;

      midi_channel = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);

      audio->midi_channel = midi_channel;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_NUMERATOR:
    {
      guint numerator;

      numerator = g_value_get_uint(value);

      /* numerator and time signature */
      pthread_mutex_lock(audio_mutex);

      audio->numerator = numerator;

      g_free(audio->time_signature);
      audio->time_signature = g_strdup_printf("%u/%u",
					      audio->numerator,
					      audio->denominator);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_DENOMINATOR:
    {
      guint denominator;

      denominator = g_value_get_uint(value);

      /* denominator and time signature */
      pthread_mutex_lock(audio_mutex);

      audio->denominator = denominator;

      g_free(audio->time_signature);
      audio->time_signature = g_strdup_printf("%u/%u",
					      audio->numerator,
					      audio->denominator);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_IS_MINOR:
    {
      gboolean is_minor;

      is_minor = g_value_get_boolean(value);

      pthread_mutex_lock(audio_mutex);

      audio->is_minor = is_minor;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_SHARP_FLATS:
    {
      guint sharp_flats;

      sharp_flats = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);

      audio->sharp_flats = sharp_flats;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OCTAVE:
    {
      gint octave;

      octave = g_value_get_int(value);

      pthread_mutex_lock(audio_mutex);

      audio->octave = octave;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_KEY:
    {
      guint key;

      key = g_value_get_uint(value);

      pthread_mutex_lock(audio_mutex);

      audio->key = key;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_ABSOLUTE_KEY:
    {
      gint absolute_key;

      absolute_key = g_value_get_int(value);

      pthread_mutex_lock(audio_mutex);

      audio->absolute_key = absolute_key;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_LOOP_START:
    {
      guint64 loop_start;

      loop_start = g_value_get_uint64(value);

      pthread_mutex_lock(audio_mutex);

      audio->loop_start = loop_start;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_LOOP_END:
    {
      guint64 loop_end;

      loop_end = g_value_get_uint64(value);

      pthread_mutex_lock(audio_mutex);

      audio->loop_end = loop_end;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OFFSET:
    {
      guint64 offset;

      offset = g_value_get_uint64(value);

      pthread_mutex_lock(audio_mutex);

      audio->offset = offset;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_PRESET:
    {
      AgsPreset *preset;

      preset = (AgsPreset *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(preset == NULL ||
	 g_list_find(audio->preset,
		     preset) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_preset(audio,
			   (GObject *) preset);
    }
    break;
  case PROP_SYNTH_GENERATOR:
    {
      AgsSynthGenerator *synth_generator;

      synth_generator = (AgsSynthGenerator *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(synth_generator == NULL ||
	 g_list_find(audio->synth_generator,
		     synth_generator) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_synth_generator(audio,
			   (GObject *) synth_generator);
    }
    break;
  case PROP_PLAYBACK_DOMAIN:
    {
      AgsPlaybackDomain *playback_domain;

      playback_domain = (AgsPlaybackDomain *) g_value_get_object(value);

      pthread_mutex_lock(audio_mutex);

      if(audio->playback_domain == (GObject *) playback_domain){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      if(audio->playback_domain != NULL){
	g_object_unref(audio->playback_domain);
      }

      if(playback_domain != NULL){
	g_object_ref(playback_domain);
      }

      audio->playback_domain = (GObject *) playback_domain;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_CURSOR:
    {
      GObject *cursor;

      cursor = (GObject *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(cursor == NULL ||
	 g_list_find(audio->cursor,
		     cursor) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_cursor(audio,
			   (GObject *) cursor);
    }
    break;
  case PROP_NOTATION:
    {
      AgsNotation *notation;

      notation = (AgsNotation *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(notation == NULL ||
	 g_list_find(audio->notation,
		     notation) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_notation(audio,
			     (GObject *) notation);
    }
    break;
  case PROP_AUTOMATION:
    {
      AgsAutomation *automation;

      automation = (AgsAutomation *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(automation == NULL ||
	 g_list_find(audio->automation,
		     automation) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_automation(audio,
			       (GObject *) automation);
    }
    break;
  case PROP_WAVE:
    {
      AgsWave *wave;

      wave = (AgsWave *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(wave == NULL ||
	 g_list_find(audio->wave, wave) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_wave(audio,
			 (GObject *) wave);
    }
    break;
  case PROP_OUTPUT_AUDIO_FILE:
    {
      AgsAudioFile *output_audio_file;

      output_audio_file = (AgsAudioFile *) g_value_get_object(value);

      pthread_mutex_lock(audio_mutex);

      if(audio->output_audio_file == (GObject *) output_audio_file){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      if(audio->output_audio_file != NULL){
	g_object_unref(audio->output_audio_file);
      }

      if(output_audio_file != NULL){
	g_object_ref(output_audio_file);
      }

      audio->output_audio_file = (GObject *) output_audio_file;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_AUDIO_FILE:
    {
      AgsAudioFile *input_audio_file;

      input_audio_file = (AgsAudioFile *) g_value_get_object(value);

      pthread_mutex_lock(audio_mutex);

      if(audio->input_audio_file == (GObject *) input_audio_file){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      if(audio->input_audio_file != NULL){
	g_object_unref(audio->input_audio_file);
      }

      if(input_audio_file != NULL){
	g_object_ref(input_audio_file);
      }

      audio->input_audio_file = (GObject *) input_audio_file;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIDI:
    {
      AgsMidi *midi;

      midi = (AgsMidi *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(midi == NULL ||
	 g_list_find(audio->midi, midi) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_midi(audio,
			 (GObject *) midi);
    }
    break;
  case PROP_OUTPUT_MIDI_FILE:
    {
      AgsMidiFile *output_midi_file;

      output_midi_file = g_value_get_object(value);

      pthread_mutex_lock(audio_mutex);

      if((AgsMidiFile *) audio->output_midi_file == output_midi_file){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      if(audio->output_midi_file != NULL){
	g_object_unref(audio->output_midi_file);
      }

      if(output_midi_file != NULL) {
	g_object_ref(output_midi_file);
      }

      audio->output_midi_file = (GObject *) output_midi_file;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_MIDI_FILE:
    {
      AgsMidiFile *input_midi_file;

      input_midi_file = g_value_get_object(value);

      pthread_mutex_lock(audio_mutex);

      if((AgsMidiFile *) audio->input_midi_file == input_midi_file){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      if(audio->input_midi_file != NULL){
	g_object_unref(audio->input_midi_file);
      }

      if(input_midi_file != NULL) {
	g_object_ref(input_midi_file);
      }

      audio->input_midi_file = (GObject *) input_midi_file;

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_RECALL_ID:
    {
      AgsRecallID *recall_id;

      recall_id = (AgsRecallID *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(recall_id == NULL ||
	 g_list_find(audio->recall_id, recall_id) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_recall_id(audio,
			      (GObject *) recall_id);
    }
    break;
  case PROP_RECYCLING_CONTEXT:
    {
      AgsRecyclingContext *recycling_context;

      recycling_context = (AgsRecyclingContext *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(recycling_context == NULL ||
	 g_list_find(audio->recycling_context, recycling_context) != NULL){
	pthread_mutex_unlock(audio_mutex);
	
	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_recycling_context(audio,
				      (GObject *) recycling_context);
    }
    break;
  case PROP_RECALL_CONTAINER:
    {
      AgsRecallContainer *recall_container;

      recall_container = (AgsRecallContainer *) g_value_get_pointer(value);

      pthread_mutex_lock(audio_mutex);

      if(recall_container == NULL ||
	 g_list_find(audio->recall_container,
		     recall_container) != NULL){
	pthread_mutex_unlock(audio_mutex);

	return;
      }

      pthread_mutex_unlock(audio_mutex);

      ags_audio_add_recall_container(audio,
				     (GObject *) recall_container);
    }
    break;
  case PROP_PLAY:
    {
      AgsRecall *recall;

      pthread_mutex_t *play_mutex;

      /* get play mutex */
      play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

      /*  */
      recall = (AgsRecall *) g_value_get_pointer(value);

      pthread_mutex_lock(play_mutex);
      
      if(recall == NULL ||
	 g_list_find(audio->play,
		     recall) != NULL){
	pthread_mutex_unlock(play_mutex);
	
	return;
      }

      pthread_mutex_unlock(play_mutex);

      ags_audio_add_recall(audio,
			   (GObject *) recall,
			   TRUE);
    }
    break;
  case PROP_RECALL:
    {
      AgsRecall *recall;

      pthread_mutex_t *recall_mutex;

      /* get recall mutex */
      recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

      /*  */
      recall = (AgsRecall *) g_value_get_pointer(value);

      pthread_mutex_lock(recall_mutex);
      
      if(recall == NULL ||
	 g_list_find(audio->recall,
		     recall) != NULL){
	pthread_mutex_unlock(recall_mutex);
	
	return;
      }

      pthread_mutex_unlock(recall_mutex);

      ags_audio_add_recall(audio,
			   (GObject *) recall,
			   FALSE);
    }
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, param_spec);
  }
}

void
ags_audio_get_property(GObject *gobject,
		       guint prop_id,
		       GValue *value,
		       GParamSpec *param_spec)
{
  AgsAudio *audio;

  pthread_mutex_t *audio_mutex;

  audio = AGS_AUDIO(gobject);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  switch(prop_id){
  case PROP_AUDIO_NAME:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_string(value,
			 audio->audio_name);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT_SOUNDCARD:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->output_soundcard);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_SOUNDCARD:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->input_soundcard);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT_SEQUENCER:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->output_sequencer);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_SEQUENCER:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->input_sequencer);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_SAMPLERATE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->samplerate);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_BUFFER_SIZE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->buffer_size);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_FORMAT:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->format);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_BPM:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_double(value,
			 audio->bpm);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIN_AUDIO_CHANNELS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->min_audio_channels);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MAX_AUDIO_CHANNELS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->max_audio_channels);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIN_OUTPUT_PADS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->min_output_pads);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MAX_OUTPUT_PADS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->max_output_pads);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIN_INPUT_PADS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->min_input_pads);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MAX_INPUT_PADS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->max_input_pads);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_AUDIO_CHANNELS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->audio_channels);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT_PADS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->output_pads);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT_LINES:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->output_lines);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_PADS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->input_pads);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_LINES:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->input_lines);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_AUDIO_START_MAPPING:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->audio_start_mapping);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_AUDIO_END_MAPPING:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->audio_end_mapping);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIDI_START_MAPPING:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->midi_start_mapping);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIDI_END_MAPPING:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->midi_end_mapping);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_MIDI_CHANNEL:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->midi_channel);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_NUMERATOR:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->numerator);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_DENOMINATOR:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->denominator);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_TIME_SIGNATURE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_string(value,
			 audio->time_signature);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_IS_MINOR:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_boolean(value,
			  audio->is_minor);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_SHARP_FLATS:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->sharp_flats);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OCTAVE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_int(value,
		      audio->octave);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_KEY:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint(value,
		       audio->key);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_ABSOLUTE_KEY:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_int(value,
		      audio->absolute_key);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_LOOP_START:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint64(value,
			 audio->loop_start);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_LOOP_END:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint64(value,
			 audio->loop_end);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OFFSET:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_uint64(value,
			 audio->offset);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->output);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->input);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_PRESET:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->preset,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_SYNTH_GENERATOR:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->synth_generator,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_PLAYBACK_DOMAIN:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->playback_domain);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_CURSOR:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->cursor,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_NOTATION:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->notation,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_AUTOMATION:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->automation,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_WAVE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->wave,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT_AUDIO_FILE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->output_audio_file);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_AUDIO_FILE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->input_audio_file);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_OUTPUT_MIDI_FILE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->output_midi_file);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_INPUT_MIDI_FILE:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_object(value,
			 audio->input_midi_file);

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_RECALL_ID:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->recall_id,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_RECYCLING_CONTEXT:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->recycling_context,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_RECALL_CONTAINER:
    {
      pthread_mutex_lock(audio_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->recall_container,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(audio_mutex);
    }
    break;
  case PROP_PLAY:
    {
      pthread_mutex_t *play_mutex;

      /* get play mutex */
      play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

      /*  */
      pthread_mutex_lock(play_mutex);
      
      g_value_set_pointer(value,
			  g_list_copy_deep(audio->play,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(play_mutex);
    }
    break;
  case PROP_RECALL:
    {
      pthread_mutex_t *recall_mutex;

      /* get recall mutex */
      recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

      /*  */
      pthread_mutex_lock(recall_mutex);

      g_value_set_pointer(value,
			  g_list_copy_deep(audio->recall,
					   (GCopyFunc) g_object_ref,
					   NULL));

      pthread_mutex_unlock(recall_mutex);
    }
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, param_spec);
  }
}

void
ags_audio_dispose(GObject *gobject)
{
  AgsAudio *audio;

  GList *list, *list_next;

  pthread_mutex_t *play_mutex, *recall_mutex;

  audio = AGS_AUDIO(gobject);

  ags_connectable_disconnect(AGS_CONNECTABLE(audio));
  
  /* soundcard */
  if(audio->output_soundcard != NULL){    
    g_object_unref(audio->output_soundcard);

    audio->output_soundcard = NULL;
  }

  if(audio->input_soundcard != NULL){    
    g_object_unref(audio->input_soundcard);

    audio->input_soundcard = NULL;
  }

  /* sequencer */
  if(audio->output_sequencer != NULL){
    g_object_unref(audio->output_sequencer);

    audio->output_sequencer = NULL;
  }

  if(audio->input_sequencer != NULL){
    g_object_unref(audio->input_sequencer);

    audio->input_sequencer = NULL;
  }

  /* channels */
  ags_audio_set_audio_channels(audio,
			       0, 0);

  ags_audio_set_pads(audio,
		     AGS_TYPE_INPUT,
		     0, 0);
  ags_audio_set_pads(audio,
		     AGS_TYPE_OUTPUT,
		     0, 0);
  
  /* preset */
  if(audio->preset != NULL){
    list = audio->preset;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }
  
    g_list_free_full(audio->preset,
		     g_object_unref);

    audio->preset = NULL;
  }

  /* synth_generator */
  if(audio->synth_generator != NULL){
    list = audio->synth_generator;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }
  
    g_list_free_full(audio->synth_generator,
		     g_object_unref);

    audio->synth_generator = NULL;
  }

  /* playback domain */
  if(audio->playback_domain != NULL){
    g_object_run_dispose(audio->playback_domain);

    audio->playback_domain = NULL;
  }
  
  /* notation */
  if(audio->notation != NULL){
    list = audio->notation;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }
  
    g_list_free_full(audio->notation,
		     g_object_unref);

    audio->notation = NULL;
  }
  
  /* automation */
  if(audio->automation != NULL){
    list = audio->automation;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }
  
    g_list_free_full(audio->automation,
		     g_object_unref);

    audio->automation = NULL;
  }

  /* wave */
  if(audio->wave != NULL){
    list = audio->wave;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }
  
    g_list_free_full(audio->wave,
		     g_object_unref);

    audio->wave = NULL;
  }

  /* output audio file */
  if(audio->output_audio_file != NULL){
    g_object_unref(audio->output_audio_file);

    audio->output_audio_file = NULL;
  }
  
  /* input audio file */
  if(audio->input_audio_file != NULL){
    g_object_unref(audio->input_audio_file);

    audio->input_audio_file = NULL;
  }

  /* midi */
  if(audio->midi != NULL){
    list = audio->midi;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }
  
    g_list_free_full(audio->midi,
		     g_object_unref);

    audio->midi = NULL;
  }

  /* output midi file */
  if(audio->output_midi_file != NULL){
    g_object_unref(audio->output_midi_file);

    audio->output_midi_file = NULL;
  }
  
  /* input midi file */
  if(audio->input_midi_file != NULL){
    g_object_unref(audio->input_midi_file);

    audio->input_midi_file = NULL;
  }

  /* recall id */
  if(audio->recall_id != NULL){
    list = audio->recall_id;
    
    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);
      
      list = list_next;
    }

    
    g_list_free_full(audio->recall_id,
		     g_object_unref);

    audio->recall_id = NULL;
  }
  
  /* recycling context */
  if(audio->recycling_context != NULL){
    list = audio->recycling_context;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }
  
    g_list_free_full(audio->recycling_context,
		     g_object_unref);

    audio->recycling_context = NULL;
  }
  
  /* recall container */
  if(audio->recall_container != NULL){
    list = audio->recall_container;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);

      list = list_next;
    }

    g_list_free_full(audio->recall_container,
		     g_object_unref);

    audio->recall_container = NULL;
  }

  /* play */
  if(audio->play != NULL){
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* run dispose and unref */
    pthread_mutex_lock(play_mutex);

    list = audio->play;

    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);
    
      list = list_next;
    }

    g_list_free_full(audio->play,
		     g_object_unref);

    audio->play = NULL;

    pthread_mutex_unlock(play_mutex);
  }

  /* recall */
  if(audio->recall != NULL){
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* run dispose and unref */
    pthread_mutex_lock(recall_mutex);
    
    list = audio->recall;
  
    while(list != NULL){
      list_next = list->next;
      
      g_object_run_dispose(list->data);
    
      list = list_next;
    }

    g_list_free_full(audio->recall,
		     g_object_unref);

    audio->recall = NULL;

    pthread_mutex_unlock(recall_mutex);
  }

  /* call parent */
  G_OBJECT_CLASS(ags_audio_parent_class)->dispose(gobject);
}

void
ags_audio_finalize(GObject *gobject)
{
  AgsAudio *audio;
  AgsChannel *channel;

  GList *list;

  audio = AGS_AUDIO(gobject);

  pthread_mutex_destroy(audio->obj_mutex);
  free(audio->obj_mutex);

  pthread_mutexattr_destroy(audio->obj_mutexattr);
  free(audio->obj_mutexattr);

  ags_uuid_free(audio->uuid);
  
  g_free(audio->audio_name);
  
  /* soundcard */
  if(audio->output_soundcard != NULL){
    g_object_unref(audio->output_soundcard);
  }

  if(audio->input_soundcard != NULL){
    g_object_unref(audio->input_soundcard);
  }

  /* sequencer */
  if(audio->output_sequencer != NULL){
    g_object_unref(audio->output_sequencer);
  }

  if(audio->input_sequencer != NULL){
    g_object_unref(audio->input_sequencer);
  }

  /* channels */
  ags_audio_set_audio_channels(audio,
			       0, 0);

  ags_audio_set_pads(audio,
		     AGS_TYPE_INPUT,
		     0, 0);
  ags_audio_set_pads(audio,
		     AGS_TYPE_OUTPUT,
		     0, 0);
  
  /* preset */
  if(audio->preset != NULL){
    list = audio->preset;

    while(list != NULL){
      g_object_set(list->data,
		   "audio", NULL,
		   NULL);

      list = list->next;
    }

    g_list_free_full(audio->preset,
		     g_object_unref);
  }

  /* synth_generator */
  if(audio->synth_generator != NULL){
    list = audio->synth_generator;

    while(list != NULL){
      g_object_set(list->data,
		   "audio", NULL,
		   NULL);

      list = list->next;
    }

    g_list_free_full(audio->synth_generator,
		     g_object_unref);
  }
  
  /* playback domain */
  if(audio->playback_domain != NULL){
    g_object_unref(audio->playback_domain);
  }
  
  /* notation */
  if(audio->notation != NULL){
    list = audio->notation;

    while(list != NULL){
      g_object_set(list->data,
		   "audio", NULL,
		   NULL);

      list = list->next;
    }

    g_list_free_full(audio->notation,
		     g_object_unref);
  }
  
  /* automation */
  if(audio->automation != NULL){
    list = audio->automation;

    while(list != NULL){
      g_object_set(list->data,
		   "audio", NULL,
		   NULL);

      list = list->next;
    }

    g_list_free_full(audio->automation,
		     g_object_unref);
  }

  /* wave */
  if(audio->wave != NULL){
    list = audio->wave;

    while(list != NULL){
      g_object_set(list->data,
		   "audio", NULL,
		   NULL);

      list = list->next;
    }

    g_list_free_full(audio->wave,
		     g_object_unref);
  }

  /* output audio file */
  if(audio->output_audio_file != NULL){
    g_object_unref(audio->output_audio_file);
  }

  /* input audio file */
  if(audio->input_audio_file != NULL){
    g_object_unref(audio->input_audio_file);
  }
  
  /* midi */
  if(audio->midi != NULL){
    list = audio->midi;

    while(list != NULL){
      g_object_set(list->data,
		   "audio", NULL,
		   NULL);

      list = list->next;
    }

    g_list_free_full(audio->midi,
		     g_object_unref);
  }

  /* output midi file */
  if(audio->output_midi_file != NULL){
    g_object_unref(audio->output_midi_file);
  }

  /* input midi file */
  if(audio->input_midi_file != NULL){
    g_object_unref(audio->input_midi_file);
  }

  /* recall id */
  if(audio->recall_id != NULL){
    g_list_free_full(audio->recall_id,
		     g_object_unref);
  }
  
  /* recycling context */
  if(audio->recycling_context != NULL){
    g_list_free_full(audio->recycling_context,
		     g_object_unref);
  }
  
  /* recall container */
  if(audio->recall_container != NULL){
    g_list_free_full(audio->recall_container,
		     g_object_unref);
  }

  /* play context */
  if(audio->play != NULL){
    list = audio->play;

    while(list != NULL){
      if(AGS_IS_RECALL_AUDIO(list->data) ||
	 AGS_IS_RECALL_AUDIO_RUN(list->data)){
	g_object_set(list->data,
		     "audio", NULL,
		     NULL);
      }
    
      list = list->next;
    }

    g_list_free_full(audio->play,
		     g_object_unref);
  }

  pthread_mutex_destroy(audio->play_mutex);
  free(audio->play_mutex);

  pthread_mutexattr_destroy(audio->play_mutexattr);
  free(audio->play_mutexattr);

  /* recall context */
  if(audio->recall != NULL){
    list = audio->recall;

    while(list != NULL){
      if(AGS_IS_RECALL_AUDIO(list->data) ||
	 AGS_IS_RECALL_AUDIO_RUN(list->data)){
	g_object_set(list->data,
		     "audio", NULL,
		     NULL);
      }
    
      list = list->next;
    }

    g_list_free_full(audio->recall,
		     g_object_unref);
  }

  pthread_mutex_destroy(audio->recall_mutex);
  free(audio->recall_mutex);

  pthread_mutexattr_destroy(audio->recall_mutexattr);
  free(audio->recall_mutexattr);
  
  /* call parent */
  G_OBJECT_CLASS(ags_audio_parent_class)->finalize(gobject);
}

AgsUUID*
ags_audio_get_uuid(AgsConnectable *connectable)
{
  AgsAudio *audio;
  
  AgsUUID *ptr;

  pthread_mutex_t *audio_mutex;

  audio = AGS_AUDIO(connectable);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get UUID */
  pthread_mutex_lock(audio_mutex);

  ptr = audio->uuid;

  pthread_mutex_unlock(audio_mutex);
  
  return(ptr);
}

gboolean
ags_audio_has_resource(AgsConnectable *connectable)
{
  return(TRUE);
}

gboolean
ags_audio_is_ready(AgsConnectable *connectable)
{
  AgsAudio *audio;
  
  gboolean is_ready;

  pthread_mutex_t *audio_mutex;

  audio = AGS_AUDIO(connectable);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* check is added */
  pthread_mutex_lock(audio_mutex);

  is_ready = (((AGS_AUDIO_ADDED_TO_REGISTRY & (audio->flags)) != 0) ? TRUE: FALSE);

  pthread_mutex_unlock(audio_mutex);
  
  return(is_ready);
}

void
ags_audio_add_to_registry(AgsConnectable *connectable)
{
  AgsAudio *audio;
  AgsChannel *channel;

  AgsRegistry *registry;
  AgsRegistryEntry *entry;

  AgsApplicationContext *application_context;

  GList *list;
  
  if(ags_connectable_is_ready(connectable)){
    return;
  }
  
  audio = AGS_AUDIO(connectable);

  ags_audio_set_flags(audio, AGS_AUDIO_ADDED_TO_REGISTRY);

  application_context = ags_application_context_get_instance();

  registry = (AgsRegistry *) ags_service_provider_get_registry(AGS_SERVICE_PROVIDER(application_context));

  if(registry != NULL){
    entry = ags_registry_entry_alloc(registry);

    entry->id = audio->uuid;
    g_value_set_object(entry->entry,
		       (gpointer) audio);
    
    ags_registry_add_entry(registry,
			   entry);
  }
  
  /* add play */
  list = audio->play;

  while(list != NULL){
    ags_connectable_add_to_registry(AGS_CONNECTABLE(list->data));

    list = list->next;
  }
  
  /* add recall */
  list = audio->recall;

  while(list != NULL){
    ags_connectable_add_to_registry(AGS_CONNECTABLE(list->data));

    list = list->next;
  }

  /* add output */
  channel = audio->output;

  while(channel != NULL){
    ags_connectable_add_to_registry(AGS_CONNECTABLE(channel));

    channel = channel->next;
  }

  /* add input */
  channel = audio->input;

  while(channel != NULL){
    ags_connectable_add_to_registry(AGS_CONNECTABLE(channel));

    channel = channel->next;
  }
}

void
ags_audio_remove_from_registry(AgsConnectable *connectable)
{
  if(!ags_connectable_is_ready(connectable)){
    return;
  }

  //TODO:JK: implement me
}

xmlNode*
ags_audio_list_resource(AgsConnectable *connectable)
{
  xmlNode *node;
  
  node = NULL;

  //TODO:JK: implement me
  
  return(node);
}

xmlNode*
ags_audio_xml_compose(AgsConnectable *connectable)
{
  xmlNode *node;
  
  node = NULL;

  //TODO:JK: implement me
  
  return(node);
}

void
ags_audio_xml_parse(AgsConnectable *connectable,
		    xmlNode *node)
{
  //TODO:JK: implement me
}

gboolean
ags_audio_is_connected(AgsConnectable *connectable)
{
  AgsAudio *audio;
  
  gboolean is_connected;

  pthread_mutex_t *audio_mutex;

  audio = AGS_AUDIO(connectable);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* check is connected */
  pthread_mutex_lock(audio_mutex);

  is_connected = (((AGS_AUDIO_CONNECTED & (audio->flags)) != 0) ? TRUE: FALSE);
  
  pthread_mutex_unlock(audio_mutex);
  
  return(is_connected);
}

void
ags_audio_connect(AgsConnectable *connectable)
{
  AgsAudio *audio;
  AgsChannel *channel;

  GList *list_start, *list;

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;
  pthread_mutex_t *recall_mutex;

  if(ags_connectable_is_connected(connectable)){
    return;
  }

  audio = AGS_AUDIO(connectable);

  ags_audio_set_flags(audio, AGS_AUDIO_CONNECTED);

#ifdef AGS_DEBUG
  g_message("connecting audio");
#endif

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* connect channels - output */
  pthread_mutex_lock(audio_mutex);

  channel = audio->output;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* connect */
    ags_connectable_connect(AGS_CONNECTABLE(channel));

    /* iterate */
    pthread_mutex_lock(channel_mutex);
    
    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }

  /* connect channels - input */
  pthread_mutex_lock(audio_mutex);

  channel = audio->input;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* connect */
    ags_connectable_connect(AGS_CONNECTABLE(channel));

    /* iterate */
    pthread_mutex_lock(channel_mutex);
    
    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }

  /* connect recall container */
  pthread_mutex_lock(audio_mutex);

  list =
    list_start = g_list_copy(audio->recall_container);

  pthread_mutex_unlock(audio_mutex);

  while(list != NULL){
    ags_connectable_connect(AGS_CONNECTABLE(list->data));

    list = list->next;
  }

  g_list_free(list_start);

  /* get recall mutex */
  recall_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

  /* connect recall - play context */
  pthread_mutex_lock(recall_mutex);

  list =
    list_start = g_list_copy(audio->play);

  pthread_mutex_unlock(recall_mutex);

  while(list != NULL){
    ags_connectable_connect(AGS_CONNECTABLE(list->data));

    list = list->next;
  }

  g_list_free(list_start);

  /* get recall mutex */
  recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

  /* connect recall - recall context */
  pthread_mutex_lock(recall_mutex);

  list =
    list_start = g_list_copy(audio->recall);

  pthread_mutex_unlock(recall_mutex);

  while(list != NULL){
    ags_connectable_connect(AGS_CONNECTABLE(list->data));

    list = list->next;
  }

  g_list_free(list_start);
}

void
ags_audio_disconnect(AgsConnectable *connectable)
{
  AgsAudio *audio;
  AgsChannel *channel;

  GList *list_start, *list;

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;
  pthread_mutex_t *recall_mutex;

  if(!ags_connectable_is_connected(connectable)){
    return;
  }

  audio = AGS_AUDIO(connectable);

  ags_audio_unset_flags(audio, AGS_AUDIO_CONNECTED);

#ifdef AGS_DEBUG
  g_message("disconnecting audio");
#endif

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* disconnect channels - output */
  pthread_mutex_lock(audio_mutex);

  channel = audio->output;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* disconnect */
    ags_connectable_disconnect(AGS_CONNECTABLE(channel));

    /* iterate */
    pthread_mutex_lock(channel_mutex);
    
    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }

  /* disconnect channels - input */
  pthread_mutex_lock(audio_mutex);

  channel = audio->input;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* disconnect */
    ags_connectable_disconnect(AGS_CONNECTABLE(channel));

    /* iterate */
    pthread_mutex_lock(channel_mutex);
    
    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }

  /* disconnect recall container */
  pthread_mutex_lock(audio_mutex);

  list =
    list_start = g_list_copy(audio->recall_container);

  pthread_mutex_unlock(audio_mutex);

  while(list != NULL){
    ags_connectable_disconnect(AGS_CONNECTABLE(list->data));

    list = list->next;
  }

  g_list_free(list_start);

  /* get recall mutex */
  recall_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

  /* disconnect recall - play context */
  pthread_mutex_lock(recall_mutex);

  list = 
    list_start = g_list_copy(audio->play);

  pthread_mutex_unlock(recall_mutex);

  while(list != NULL){
    ags_connectable_disconnect(AGS_CONNECTABLE(list->data));

    list = list->next;
  }

  g_list_free(list_start);

  /* get recall mutex */
  recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

  /* disconnect recall - recall context */
  pthread_mutex_lock(recall_mutex);

  list = 
    list_start = g_list_copy(audio->recall);

  pthread_mutex_unlock(recall_mutex);

  while(list != NULL){
    ags_connectable_disconnect(AGS_CONNECTABLE(list->data));

    list = list->next;
  }

  g_list_free(list_start);
}

/**
 * ags_audio_get_class_mutex:
 * 
 * Use this function's returned mutex to access mutex fields.
 *
 * Returns: the class mutex
 * 
 * Since: 2.0.0
 */
pthread_mutex_t*
ags_audio_get_class_mutex()
{
  return(&ags_audio_class_mutex);
}

/**
 * ags_audio_test_flags:
 * @audio: the #AgsAudio
 * @flags: the flags
 *
 * Test @flags to be set on @audio.
 * 
 * Returns: %TRUE if flags are set, else %FALSE
 *
 * Since: 2.0.0
 */
gboolean
ags_audio_test_flags(AgsAudio *audio, guint flags)
{
  gboolean retval;  
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return(FALSE);
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);
    
  /* test */
  pthread_mutex_lock(audio_mutex);

  retval = (flags & (audio->flags)) ? TRUE: FALSE;
  
  pthread_mutex_unlock(audio_mutex);

  return(retval);
}

/**
 * ags_audio_set_flags:
 * @audio: the #AgsAudio
 * @flags: see enum AgsAudioFlags
 *
 * Enable a feature of #AgsAudio.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_flags(AgsAudio *audio, guint flags)
{
  guint audio_flags;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* check flags */
  pthread_mutex_lock(audio_mutex);
  
  audio_flags = audio->flags;
  
  pthread_mutex_unlock(audio_mutex);

  if((AGS_AUDIO_NO_OUTPUT & flags) != 0 &&
     (AGS_AUDIO_NO_OUTPUT & audio_flags) == 0){
    ags_audio_set_pads(audio,
		       AGS_TYPE_OUTPUT,
		       0, 0);
  }

  if((AGS_AUDIO_NO_INPUT & flags) != 0 &&
     (AGS_AUDIO_NO_INPUT & audio_flags) == 0){
    ags_audio_set_pads(audio,
		       AGS_TYPE_INPUT,
		       0, 0);
  }
  
  if((AGS_AUDIO_OUTPUT_HAS_RECYCLING & (flags)) != 0 &&
     (AGS_AUDIO_OUTPUT_HAS_RECYCLING & (audio_flags)) == 0){
    AgsChannel *channel;
    AgsRecycling *first_recycling, *last_recycling;

    GObject *soundcard;
    
    pthread_mutex_t *channel_mutex;

    pthread_mutex_lock(audio_mutex);

    channel = audio->output;

    pthread_mutex_unlock(audio_mutex);

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* get some fields */
      pthread_mutex_lock(channel_mutex);

      soundcard = channel->output_soundcard;
      
      pthread_mutex_unlock(channel_mutex);
      
      /* add recycling */
      first_recycling =
	last_recycling = ags_recycling_new(soundcard);
      g_object_ref(first_recycling);
      g_object_set(first_recycling,
		   "channel", channel,
		   NULL);
	  
      ags_channel_reset_recycling(channel,
				  first_recycling, last_recycling);
      
      /* iterate */
      pthread_mutex_lock(channel_mutex);

      channel = channel->next;
      
      pthread_mutex_unlock(channel_mutex);
    }
  }

  if((AGS_AUDIO_INPUT_HAS_RECYCLING & (flags)) != 0 &&
     (AGS_AUDIO_INPUT_HAS_RECYCLING & (audio_flags)) == 0){
    AgsChannel *channel;
    AgsRecycling *recycling;
    AgsRecycling *first_recycling, *last_recycling;

    GObject *soundcard;
    
    pthread_mutex_t *channel_mutex;

    pthread_mutex_lock(audio_mutex);

    channel = audio->output;

    pthread_mutex_unlock(audio_mutex);

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* get some fields */
      pthread_mutex_lock(channel_mutex);

      soundcard = channel->output_soundcard;
      recycling = channel->first_recycling;
      
      pthread_mutex_unlock(channel_mutex);
      
      /* add recycling */
      if(recycling == NULL){
	first_recycling =
	  last_recycling = ags_recycling_new(soundcard);
	g_object_ref(first_recycling);
	g_object_set(first_recycling,
		     "channel", channel,
		     NULL);
	  
	ags_channel_reset_recycling(channel,
				    first_recycling, last_recycling);
      }
      
      /* iterate */
      pthread_mutex_lock(channel_mutex);

      channel = channel->next;
      
      pthread_mutex_unlock(channel_mutex);
    }
  }

  //TODO:JK: add more?

  /* set flags */
  pthread_mutex_lock(audio_mutex);

  audio->flags |= flags;
  
  pthread_mutex_unlock(audio_mutex);
}
    
/**
 * ags_audio_unset_flags:
 * @audio: the #AgsAudio
 * @flags: see enum AgsAudioFlags
 *
 * Disable a feature of AgsAudio.
 *
 * Since: 2.0.0
 */
void
ags_audio_unset_flags(AgsAudio *audio, guint flags)
{
  guint audio_flags;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* check flags */
  pthread_mutex_lock(audio_mutex);
  
  audio_flags = audio->flags;
  
  pthread_mutex_unlock(audio_mutex);

  if((AGS_AUDIO_OUTPUT_HAS_RECYCLING & (audio_flags)) != 0 &&
     (AGS_AUDIO_OUTPUT_HAS_RECYCLING & (flags)) == 0){
    AgsChannel *channel;
    AgsRecycling *first_recycling;
    
    pthread_mutex_t *channel_mutex;

    pthread_mutex_lock(audio_mutex);

    channel = audio->output;

    pthread_mutex_unlock(audio_mutex);

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* unset recycling */
      pthread_mutex_lock(channel_mutex);

      first_recycling = channel->first_recycling;

      channel->first_recycling =
	channel->last_recycling = NULL;
      
      pthread_mutex_unlock(channel_mutex);
      
      /* remove recycling */
      if(first_recycling != NULL){
	g_object_run_dispose((GObject *) first_recycling);
	g_object_unref(first_recycling);
      }
      
      ags_channel_reset_recycling(channel,
				  NULL, NULL);
      
      /* iterate */
      pthread_mutex_lock(channel_mutex);

      channel = channel->next;
      
      pthread_mutex_unlock(channel_mutex);
    }
  }
  
  if((AGS_AUDIO_INPUT_HAS_RECYCLING & (audio_flags)) != 0 &&
     (AGS_AUDIO_INPUT_HAS_RECYCLING & (flags)) == 0){
    AgsChannel *channel, *link;
    AgsRecycling *first_recycling;
    
    pthread_mutex_t *channel_mutex;

    pthread_mutex_lock(audio_mutex);

    channel = audio->output;

    pthread_mutex_unlock(audio_mutex);

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* get some fields */
      pthread_mutex_lock(channel_mutex);

      link = channel->link;
      
      pthread_mutex_unlock(channel_mutex);
      
      /* unset recycling */
      if(link == NULL){
	pthread_mutex_lock(channel_mutex);

	first_recycling = channel->first_recycling;

	channel->first_recycling =
	  channel->last_recycling = NULL;
      
	pthread_mutex_unlock(channel_mutex);
      
	/* remove recycling */
	if(first_recycling != NULL){
	  g_object_run_dispose((GObject *) first_recycling);
	  g_object_unref(first_recycling);
	}
	
	ags_channel_reset_recycling(channel,
				    NULL, NULL);
      }
      
      /* iterate */
      pthread_mutex_lock(channel_mutex);

      channel = channel->next;
      
      pthread_mutex_unlock(channel_mutex);
    }
  }

  //TODO:JK: add more?

  /* unset flags */
  pthread_mutex_lock(audio_mutex);

  audio->flags &= (~flags);
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_test_ability_flags:
 * @audio: the #AgsAudio
 * @ability_flags: the ability flags
 *
 * Test @ability_flags to be set on @audio.
 * 
 * Returns: %TRUE if flags are set, else %FALSE
 *
 * Since: 2.0.0
 */
gboolean
ags_audio_test_ability_flags(AgsAudio *audio, guint ability_flags)
{
  gboolean retval;  
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return(FALSE);
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* test */
  pthread_mutex_lock(audio_mutex);

  retval = (ability_flags & (audio->ability_flags)) ? TRUE: FALSE;
  
  pthread_mutex_unlock(audio_mutex);

  return(retval);
}

/**
 * ags_audio_set_ability_flags:
 * @audio: the #AgsAudio
 * @ability_flags: see enum AgsSoundAbilityFlags
 *
 * Enable an ability of AgsAudio.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_ability_flags(AgsAudio *audio, guint ability_flags)
{
  AgsChannel *output, *input;  
  AgsPlaybackDomain *playback_domain;
  
  AgsThread *main_loop;
  AgsAudioLoop *audio_loop;
  
  AgsApplicationContext *application_context;
  
  GObject *soundcard;
  
  guint samplerate, buffer_size;
  guint audio_ability_flags;
  gboolean super_threaded_audio;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *playback_domain_mutex;

  auto void ags_audio_set_ability_flags_channel(AgsChannel *channel, guint ability_flags);

  void ags_audio_set_ability_flags_channel(AgsChannel *channel, guint ability_flags){
    pthread_mutex_t *channel_mutex;

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* set ability flags */
      ags_channel_set_ability_flags(channel, ability_flags);

      /* iterate */
      pthread_mutex_lock(channel_mutex);

      channel = channel->next;
      
      pthread_mutex_unlock(channel_mutex);
    }
  }
  
  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  application_context = ags_application_context_get_instance();

  /* get main loop */
  main_loop = ags_concurrency_provider_get_main_loop(AGS_CONCURRENCY_PROVIDER(application_context));
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* check flags */
  pthread_mutex_lock(audio_mutex);

  audio_ability_flags = audio->ability_flags;

  soundcard = audio->output_soundcard;
  
  samplerate = audio->samplerate;
  buffer_size = audio->buffer_size;

  output = audio->output;
  input = audio->input;
  
  playback_domain = AGS_PLAYBACK_DOMAIN(audio->playback_domain);
  
  pthread_mutex_unlock(audio_mutex);

  /* get audio mutex */
  playback_domain_mutex = AGS_PLAYBACK_DOMAIN_GET_OBJ_MUTEX(playback_domain);

  /* get super-threaded flags */
  pthread_mutex_lock(playback_domain_mutex);

  super_threaded_audio = ((AGS_PLAYBACK_DOMAIN_SUPER_THREADED_AUDIO & (g_atomic_int_get(&(playback_domain->flags)))) != 0) ? TRUE: FALSE;

  pthread_mutex_unlock(playback_domain_mutex);
  
  /* notation ability */  
  if(super_threaded_audio){
    /* find audio loop */
    audio_loop = (AgsAudioLoop *) ags_thread_find_type(main_loop,
						       AGS_TYPE_AUDIO_LOOP);
    
    /* playback ability */
    if((AGS_SOUND_ABILITY_PLAYBACK & (ability_flags)) != 0 &&
       (AGS_SOUND_ABILITY_PLAYBACK & (audio_ability_flags)) == 0){
      AgsAudioThread *audio_thread;

      audio_thread = ags_audio_thread_new(soundcard,
					  (GObject *) audio);
      ags_audio_thread_set_sound_scope(audio_thread,
				       AGS_SOUND_SCOPE_PLAYBACK);
      
      g_object_set(audio_thread,
		   "frequency", ceil((gdouble) samplerate / (gdouble) buffer_size) + AGS_SOUNDCARD_DEFAULT_OVERCLOCK,
		   NULL);
    
      ags_playback_domain_set_audio_thread(playback_domain,
					   (AgsThread *) audio_thread,
					   AGS_SOUND_SCOPE_PLAYBACK);    

      /* set thread child */
      ags_thread_add_child_extended((AgsThread *) audio_loop,
				    (AgsThread *) audio_thread,
				    TRUE, TRUE);
    }

    /* sequencer ability */
    if((AGS_SOUND_ABILITY_SEQUENCER & (ability_flags)) != 0 &&
       (AGS_SOUND_ABILITY_SEQUENCER & (audio_ability_flags)) == 0){
      AgsAudioThread *audio_thread;

      audio_thread = ags_audio_thread_new(soundcard,
					  (GObject *) audio);
      ags_audio_thread_set_sound_scope(audio_thread,
				       AGS_SOUND_SCOPE_SEQUENCER);

      g_object_set(audio_thread,
		   "frequency", ceil((gdouble) samplerate / (gdouble) buffer_size) + AGS_SOUNDCARD_DEFAULT_OVERCLOCK,
		   NULL);
    
      ags_playback_domain_set_audio_thread(playback_domain,
					   (AgsThread *) audio_thread,
					   AGS_SOUND_SCOPE_SEQUENCER);

      /* set thread child */
      ags_thread_add_child_extended((AgsThread *) audio_loop,
				    (AgsThread *) audio_thread,
				    TRUE, TRUE);
    }

    /* notation ability */
    if((AGS_SOUND_ABILITY_NOTATION & (ability_flags)) != 0 &&
       (AGS_SOUND_ABILITY_NOTATION & (audio_ability_flags)) == 0){
      AgsAudioThread *audio_thread;

      audio_thread = ags_audio_thread_new(soundcard,
					  (GObject *) audio);
      ags_audio_thread_set_sound_scope(audio_thread,
				       AGS_SOUND_SCOPE_NOTATION);

      g_object_set(audio_thread,
		   "frequency", ceil((gdouble) samplerate / (gdouble) buffer_size) + AGS_SOUNDCARD_DEFAULT_OVERCLOCK,
		   NULL);
    
      ags_playback_domain_set_audio_thread(playback_domain,
					   (AgsThread *) audio_thread,
					   AGS_SOUND_SCOPE_NOTATION);    

      /* set thread child */
      ags_thread_add_child_extended((AgsThread *) audio_loop,
				    (AgsThread *) audio_thread,
				    TRUE, TRUE);
    }

    /* wave ability */
    if((AGS_SOUND_ABILITY_WAVE & (ability_flags)) != 0 &&
       (AGS_SOUND_ABILITY_WAVE & (audio_ability_flags)) == 0){
      AgsAudioThread *audio_thread;

      audio_thread = ags_audio_thread_new(soundcard,
					  (GObject *) audio);
      ags_audio_thread_set_sound_scope(audio_thread,
				       AGS_SOUND_SCOPE_WAVE);

      g_object_set(audio_thread,
		   "frequency", ceil((gdouble) samplerate / (gdouble) buffer_size) + AGS_SOUNDCARD_DEFAULT_OVERCLOCK,
		   NULL);
    
      ags_playback_domain_set_audio_thread(playback_domain,
					   (AgsThread *) audio_thread,
					   AGS_SOUND_SCOPE_WAVE);

      /* set thread child */
      ags_thread_add_child_extended((AgsThread *) audio_loop,
				    (AgsThread *) audio_thread,
				    TRUE, TRUE);
    }

    /* midi ability */
    if((AGS_SOUND_ABILITY_MIDI & (ability_flags)) != 0 &&
       (AGS_SOUND_ABILITY_MIDI & (audio_ability_flags)) == 0){
      AgsAudioThread *audio_thread;

      audio_thread = ags_audio_thread_new(soundcard,
					  (GObject *) audio);
      ags_audio_thread_set_sound_scope(audio_thread,
				       AGS_SOUND_SCOPE_MIDI);

      g_object_set(audio_thread,
		   "frequency", ceil((gdouble) samplerate / (gdouble) buffer_size) + AGS_SOUNDCARD_DEFAULT_OVERCLOCK,
		   NULL);
    
      ags_playback_domain_set_audio_thread(playback_domain,
					   (AgsThread *) audio_thread,
					   AGS_SOUND_SCOPE_MIDI);

      /* set thread child */
      ags_thread_add_child_extended((AgsThread *) audio_loop,
				    (AgsThread *) audio_thread,
				    TRUE, TRUE);
    }
  }
  
  /* channel */
  ags_audio_set_ability_flags_channel(output, ability_flags);  
  ags_audio_set_ability_flags_channel(input, ability_flags);
  
  /* set flags */
  pthread_mutex_lock(audio_mutex);

  audio->ability_flags |= ability_flags;
  
  pthread_mutex_unlock(audio_mutex);

  /* unref */
  g_object_unref(main_loop);
 }

/**
 * ags_audio_unset_ability_flags:
 * @audio: the #AgsAudio
 * @ability_flags: see enum AgsSoundAbilityFlags
 *
 * Disable an ability of AgsAudio.
 *
 * Since: 2.0.0
 */
void
ags_audio_unset_ability_flags(AgsAudio *audio, guint ability_flags)
{
  AgsChannel *output, *input;  
  AgsPlaybackDomain *playback_domain;

  AgsThread *main_loop;
  AgsThread *audio_loop;

  AgsApplicationContext *application_context;
  
  guint audio_ability_flags;
  
  pthread_mutex_t *audio_mutex;

  auto void ags_audio_unset_ability_flags_channel(AgsChannel *channel, guint ability_flags);

  void ags_audio_unset_ability_flags_channel(AgsChannel *channel, guint ability_flags){
    pthread_mutex_t *channel_mutex;

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* set ability flags */
      ags_channel_unset_ability_flags(channel, ability_flags);

      /* iterate */
      pthread_mutex_lock(channel_mutex);

      channel = channel->next;
      
      pthread_mutex_unlock(channel_mutex);
    }
  }

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  application_context = ags_application_context_get_instance();

  /* get main loop */
  main_loop = ags_concurrency_provider_get_main_loop(AGS_CONCURRENCY_PROVIDER(application_context));

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* check flags */
  pthread_mutex_lock(audio_mutex);
  
  audio_ability_flags = audio->ability_flags;

  output = audio->output;
  input = audio->input;
  
  playback_domain = AGS_PLAYBACK_DOMAIN(audio->playback_domain);

  pthread_mutex_unlock(audio_mutex);

  /* find audio loop */
  audio_loop = ags_thread_find_type(main_loop,
				    AGS_TYPE_AUDIO_LOOP);

  /* playback ability */
  if((AGS_SOUND_ABILITY_PLAYBACK & (ability_flags)) == 0 &&
     (AGS_SOUND_ABILITY_PLAYBACK & (audio_ability_flags)) != 0){
    AgsAudioThread *audio_thread;

    audio_thread = (AgsAudioThread *) ags_playback_domain_get_audio_thread(playback_domain,
									   AGS_SOUND_SCOPE_PLAYBACK);
    ags_thread_remove_child(audio_loop,
			    (AgsThread *) audio_thread);
    
    ags_playback_domain_set_audio_thread(playback_domain,
					 NULL,
					 AGS_SOUND_SCOPE_PLAYBACK);
  }

  /* notation ability */
  if((AGS_SOUND_ABILITY_NOTATION & (ability_flags)) == 0 &&
     (AGS_SOUND_ABILITY_NOTATION & (audio_ability_flags)) != 0){
    AgsAudioThread *audio_thread;
    
    audio_thread = (AgsAudioThread *) ags_playback_domain_get_audio_thread(playback_domain,
									   AGS_SOUND_SCOPE_NOTATION);
    ags_thread_remove_child(audio_loop,
			    (AgsThread *) audio_thread);    

    ags_playback_domain_set_audio_thread(playback_domain,
					 NULL,
					 AGS_SOUND_SCOPE_NOTATION);
  }

  /* sequencer ability */
  if((AGS_SOUND_ABILITY_SEQUENCER & (ability_flags)) == 0 &&
     (AGS_SOUND_ABILITY_SEQUENCER & (audio_ability_flags)) != 0){
    AgsAudioThread *audio_thread;
    
    audio_thread = (AgsAudioThread *) ags_playback_domain_get_audio_thread(playback_domain,
									   AGS_SOUND_SCOPE_SEQUENCER);
    ags_thread_remove_child(audio_loop,
			    (AgsThread *) audio_thread);    

    ags_playback_domain_set_audio_thread(playback_domain,
					 NULL,
					 AGS_SOUND_SCOPE_SEQUENCER);
  }

  /* wave ability */
  if((AGS_SOUND_ABILITY_WAVE & (ability_flags)) == 0 &&
     (AGS_SOUND_ABILITY_WAVE & (audio_ability_flags)) != 0){
    AgsAudioThread *audio_thread;
    
    audio_thread = (AgsAudioThread *) ags_playback_domain_get_audio_thread(playback_domain,
									   AGS_SOUND_SCOPE_WAVE);
    ags_thread_remove_child(audio_loop,
			    (AgsThread *) audio_thread);    

    ags_playback_domain_set_audio_thread(playback_domain,
					 NULL,
					 AGS_SOUND_SCOPE_WAVE);
  }

  /* midi ability */
  if((AGS_SOUND_ABILITY_MIDI & (ability_flags)) == 0 &&
     (AGS_SOUND_ABILITY_MIDI & (audio_ability_flags)) != 0){
    AgsAudioThread *audio_thread;
    
    audio_thread = (AgsAudioThread *) ags_playback_domain_get_audio_thread(playback_domain,
									   AGS_SOUND_SCOPE_MIDI);
    ags_thread_remove_child(audio_loop,
			    (AgsThread *) audio_thread);    

    ags_playback_domain_set_audio_thread(playback_domain,
					 NULL,
					 AGS_SOUND_SCOPE_MIDI);
  }

  /* channel */
  ags_audio_unset_ability_flags_channel(output, ability_flags);  
  ags_audio_unset_ability_flags_channel(input, ability_flags);
  
  /* unset flags */
  pthread_mutex_lock(audio_mutex);

  audio->ability_flags &= (~ability_flags);
  
  pthread_mutex_unlock(audio_mutex);

  /* unref */
  g_object_unref(main_loop);
}

/**
 * ags_audio_test_behaviour_flags:
 * @audio: the #AgsAudio
 * @behaviour_flags: the behaviour flags
 *
 * Test @behaviour_flags to be set on @audio.
 * 
 * Returns: %TRUE if flags are set, else %FALSE
 *
 * Since: 2.0.0
 */
gboolean
ags_audio_test_behaviour_flags(AgsAudio *audio, guint behaviour_flags)
{
  gboolean retval;  
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return(FALSE);
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* test */
  pthread_mutex_lock(audio_mutex);

  retval = (behaviour_flags & (audio->behaviour_flags)) ? TRUE: FALSE;
  
  pthread_mutex_unlock(audio_mutex);

  return(retval);
}

/**
 * ags_audio_set_behaviour_flags:
 * @audio: the #AgsAudio
 * @behaviour_flags: the behaviour flags
 * 
 * Set behaviour flags.
 * 
 * Since: 2.0.0
 */
void
ags_audio_set_behaviour_flags(AgsAudio *audio, guint behaviour_flags)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* set flags */
  pthread_mutex_lock(audio_mutex);

  audio->behaviour_flags |= behaviour_flags;

  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_unset_behaviour_flags:
 * @audio: the #AgsAudio
 * @behaviour_flags: the behaviour flags
 * 
 * Unset behaviour flags.
 * 
 * Since: 2.0.0
 */
void
ags_audio_unset_behaviour_flags(AgsAudio *audio, guint behaviour_flags)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* unset flags */
  pthread_mutex_lock(audio_mutex);

  audio->behaviour_flags &= (~behaviour_flags);

  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_test_staging_flags:
 * @audio: the #AgsAudio
 * @sound_scope: the #AgsSoundScope to test
 * @staging_flags: the staging flags
 *
 * Test @staging_flags to be set on @audio.
 * 
 * Returns: %TRUE if flags are set, else %FALSE
 *
 * Since: 2.0.0
 */
gboolean
ags_audio_test_staging_flags(AgsAudio *audio, gint sound_scope,
			     guint staging_flags)
{
  gboolean retval;  
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return(FALSE);
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* test */
  pthread_mutex_lock(audio_mutex);

  retval = (staging_flags & (audio->staging_flags[sound_scope])) ? TRUE: FALSE;
  
  pthread_mutex_unlock(audio_mutex);

  return(retval);
}

/**
 * ags_audio_set_staging_flags:
 * @audio: the #AgsAudio
 * @sound_scope: the #AgsSoundScope to apply, or -1 to apply to all
 * @staging_flags: the staging flags
 * 
 * Set staging flags.
 * 
 * Since: 2.0.0
 */
void
ags_audio_set_staging_flags(AgsAudio *audio, gint sound_scope,
			    guint staging_flags)
{
  guint i;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* set flags */
  pthread_mutex_lock(audio_mutex);

  if(sound_scope < 0){
    for(i = 0; i < AGS_SOUND_SCOPE_LAST; i++){
      audio->staging_flags[i] |= staging_flags;
    }
  }else if(sound_scope < AGS_SOUND_SCOPE_LAST){
    audio->staging_flags[sound_scope] |= staging_flags;
  }

  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_unset_staging_flags:
 * @audio: the #AgsAudio
 * @sound_scope: the #AgsSoundScope to apply, or -1 to apply to all
 * @staging_flags: the staging flags
 * 
 * Unset staging flags.
 * 
 * Since: 2.0.0
 */
void
ags_audio_unset_staging_flags(AgsAudio *audio, gint sound_scope,
			      guint staging_flags)
{
  guint i;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* unset flags */
  pthread_mutex_lock(audio_mutex);

  if(sound_scope < 0){
    for(i = 0; i < AGS_SOUND_SCOPE_LAST; i++){
      audio->staging_flags[i] &= (~staging_flags);
    }
  }else if(sound_scope < AGS_SOUND_SCOPE_LAST){
    audio->staging_flags[sound_scope] &= (~staging_flags);
  }

  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_find_name:
 * @audio: the #GList-struct containing #AgsAudio
 * @audio_name: the audio name to find
 * 
 * Find @audio_name in @audio.
 *
 * Returns: the next matching #GList-struct containing #AgsAudio
 * 
 * Since: 2.1.0
 */
GList*
ags_audio_find_name(GList *audio,
		    gchar *audio_name)
{
  if(audio == NULL ||
     audio_name == NULL){
    return(NULL);
  }

  while(audio != NULL){
    gboolean success;
    
    pthread_mutex_t *audio_mutex;

    /* get audio mutex */
    audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio->data);

    /* match */
    pthread_mutex_lock(audio_mutex);

    success = (AGS_AUDIO(audio->data)->audio_name != NULL &&
	       !g_strcmp0(AGS_AUDIO(audio->data)->audio_name,
			  audio_name)) ? TRUE: FALSE;
    
    pthread_mutex_unlock(audio_mutex);
    
    if(success){
      return(audio);
    }
    
    audio = audio->next;
  }
  
  return(NULL);
}

/**
 * ags_audio_set_max_audio_channels:
 * @audio: the #AgsAudio
 * @max_audio_channels: maximum of audio channels
 * 
 * Set maximum audio channels.
 * 
 * Since: 2.0.0
 */
void
ags_audio_set_max_audio_channels(AgsAudio *audio,
				 guint max_audio_channels)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* set max audio channels */
  pthread_mutex_lock(audio_mutex);

  audio->max_audio_channels = max_audio_channels;

  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_set_max_pads:
 * @audio: the #AgsAudio
 * @channel_type: the #GType of channel, either AGS_TYPE_OUTPUT or AGS_TYPE_INPUT
 * @max_pads: maximum of pads
 * 
 * Set maximum pads of @channel_type.
 * 
 * Since: 2.0.0
 */
void
ags_audio_set_max_pads(AgsAudio *audio,
		       GType channel_type,
		       guint max_pads)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* set max pads */
  pthread_mutex_lock(audio_mutex);

  if(g_type_is_a(channel_type,
		 AGS_TYPE_OUTPUT)){
    audio->max_output_pads = max_pads;
  }else if(g_type_is_a(channel_type,
		       AGS_TYPE_INPUT)){
    audio->max_input_pads = max_pads;
  }

  pthread_mutex_unlock(audio_mutex);
}

void
ags_audio_real_set_audio_channels(AgsAudio *audio,
				  guint audio_channels, guint audio_channels_old)
{
  AgsMessageDelivery *message_delivery;
  AgsMessageQueue *message_queue;

  GObject *output_soundcard, *input_soundcard;

  gboolean alloc_recycling;
  gboolean link_recycling; // affects AgsInput
  gboolean set_sync_link, set_async_link; // affects AgsOutput
  gboolean alloc_pattern;
  gboolean add_synth_generator;

  guint bank_dim[3];

  guint audio_flags;
  guint samplerate;
  guint buffer_size;
  guint format;
  guint output_pads, input_pads;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *prev_mutex, *prev_pad_mutex, *current_mutex;
  
  auto void ags_audio_set_audio_channels_init_parameters(GType type);
  auto void ags_audio_set_audio_channels_grow(GType type);
  auto void ags_audio_set_audio_channels_shrink_zero();
  auto void ags_audio_set_audio_channels_shrink();

  auto void ags_audio_set_audio_channels_shrink_notation();
  auto void ags_audio_set_audio_channels_shrink_automation();
  auto void ags_audio_set_audio_channels_shrink_wave();
  auto void ags_audio_set_audio_channels_shrink_midi();
  
  void ags_audio_set_audio_channels_init_parameters(GType type){
    alloc_pattern = FALSE;
    add_synth_generator = FALSE;
    
    if(type == AGS_TYPE_OUTPUT){
      link_recycling = FALSE;

      if((AGS_AUDIO_OUTPUT_HAS_RECYCLING & (audio_flags)) != 0){
	alloc_recycling = TRUE;
      }else{
	alloc_recycling = FALSE;

	if((AGS_AUDIO_INPUT_HAS_RECYCLING & (audio_flags)) != 0){
	  if((AGS_AUDIO_SYNC & (audio_flags)) != 0 && (AGS_AUDIO_ASYNC & (audio_flags)) == 0){
	    set_sync_link = FALSE;
	    set_async_link = TRUE;
	  }else if((AGS_AUDIO_ASYNC & (audio_flags)) != 0){
	    set_async_link = TRUE;
	    set_sync_link = FALSE;
	  }else{
#ifdef AGS_DEBUG
	    g_message("ags_audio_set_audio_channels - warning: AGS_AUDIO_SYNC nor AGS_AUDIO_ASYNC weren't defined");
#endif
	    set_sync_link = FALSE;
	    set_async_link = FALSE;
	  }
	}
      }
    }else{
      set_sync_link = FALSE;
      set_async_link = FALSE;
      
      if((AGS_AUDIO_INPUT_HAS_RECYCLING & (audio_flags)) != 0){
	alloc_recycling = TRUE;
      }else{
	alloc_recycling = FALSE;
      }
      
      if((AGS_AUDIO_ASYNC & (audio_flags)) != 0 && alloc_recycling){
	link_recycling = TRUE;
      }else{
	link_recycling = FALSE;
      }

      if((AGS_SOUND_ABILITY_SEQUENCER & (audio->ability_flags)) != 0){
	alloc_pattern = TRUE;
	
	bank_dim[0] = audio->bank_dim[0];
	bank_dim[1] = audio->bank_dim[1];
	bank_dim[2] = audio->bank_dim[2];
      }

      if((AGS_AUDIO_INPUT_HAS_SYNTH & (audio_flags)) != 0){
	add_synth_generator = TRUE;
      }
    }    
  }
  
  void ags_audio_set_audio_channels_grow(GType type){
    AgsChannel *start;
    AgsChannel *channel, *current, *pad_next, *nth_channel;
    AgsRecycling *first_recycling, *last_recycling;
    AgsPlaybackDomain *playback_domain;
    
    guint pads;
    guint i, j;

    pthread_mutex_lock(audio_mutex);

    playback_domain = (AgsPlaybackDomain *) audio->playback_domain;
    
    if(type == AGS_TYPE_OUTPUT){
      /* AGS_TYPE_OUTPUT */
      pads = audio->output_pads;
      
      start = audio->output;
    }else{
      /* AGS_TYPE_INPUT */
      pads = audio->input_pads;

      start = audio->input;
    }

    pthread_mutex_unlock(audio_mutex);

    /* grow */
    for(j = 0; j < pads; j++){
      if(audio_channels_old != 0){
	nth_channel = ags_channel_nth(start, j * audio_channels);

	pad_next = ags_channel_next_pad(nth_channel);

	g_object_unref(nth_channel);
	g_object_unref(pad_next);
      }else{
	pad_next = NULL;
      }

      for(i = audio_channels_old; i < audio_channels; i++){
	AgsPlayback *playback;
	
	channel = (AgsChannel *) g_object_new(type,
					      "audio", (GObject *) audio,
					      "output-soundcard", output_soundcard,
					      "output-soundcard-channel", i,
					      "input-soundcard", input_soundcard,
					      "input-soundcard-channel", i,
					      "samplerate", samplerate,
					      "buffer-size", buffer_size,
					      NULL);
	g_object_ref(channel);

	playback = (AgsPlayback *) channel->playback;
	g_object_set(playback,
		     "playback-domain", playback_domain,
		     NULL);
	
	if(alloc_pattern){
	  channel->pattern = g_list_alloc();
	  channel->pattern->data = (gpointer) ags_pattern_new();
	  g_object_ref(channel->pattern->data);
	  
	  ags_pattern_set_dim((AgsPattern *) channel->pattern->data,
			      bank_dim[0], bank_dim[1], bank_dim[2]);
	}

	if(add_synth_generator){
	  AgsSynthGenerator *synth_generator;

	  synth_generator = ags_synth_generator_new();
	  ags_input_add_synth_generator((AgsInput *) channel,
					(GObject *) synth_generator);
	}
	
	if(start == NULL){
	  pthread_mutex_lock(audio_mutex);
	  
	  start = channel;

	  if(type == AGS_TYPE_OUTPUT){
	    audio->output = channel;
	  }else{
	    audio->input = channel;
	  }

	  pthread_mutex_unlock(audio_mutex);
	}

	if(j * audio_channels + i != 0){
	  /* set prev */
	  channel->prev = ags_channel_nth(start, j * audio_channels + i - 1);
	  g_object_unref(channel->prev);
	  
	  /* get prev mutex */
	  prev_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel->prev);

	  /* set next and prev->next */
	  pthread_mutex_lock(prev_mutex);  

	  if(audio_channels_old != 0 &&
	     i == audio_channels - 1){
	    channel->next = pad_next;
	  }
	  
	  channel->prev->next = channel;

	  pthread_mutex_unlock(prev_mutex);
	}

	if(j != 0){
	  /* set prev pad */
	  nth_channel = ags_channel_nth(start, i);
	  
	  channel->prev_pad = ags_channel_pad_nth(nth_channel, j - 1);

	  g_object_unref(nth_channel);
	  g_object_unref(channel->prev_pad);
	  
	  /* get prev pad mutex */
	  prev_pad_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel->prev_pad);

	  /* set prev pad next pad */
	  pthread_mutex_lock(prev_pad_mutex);
	  
	  channel->prev_pad->next_pad = channel;

	  pthread_mutex_unlock(prev_pad_mutex);
	}

	/* set indices */
	channel->pad = j;
	channel->audio_channel = i;
	channel->line = j * audio_channels + i;

	/* reset nested AgsRecycling tree */
	if(alloc_recycling){
	  first_recycling =
	    last_recycling = ags_recycling_new(output_soundcard);
	  g_object_ref(first_recycling);
	  g_object_set(first_recycling,
		       "channel", channel,
		       NULL);
	  
	  ags_channel_reset_recycling(channel,
				      first_recycling, last_recycling);
	}else if(set_sync_link){
	  AgsChannel *input;

	  pthread_mutex_lock(audio_mutex);

	  input = audio->input;
	  
	  pthread_mutex_unlock(audio_mutex);

	  input = ags_channel_nth(input,
				  channel->line);
	  g_object_unref(input);
	  
	  /* set sync link */
	  if(input != NULL){
	    first_recycling = input->first_recycling;
	    last_recycling = input->last_recycling;
	    
	    ags_channel_reset_recycling(channel,
					first_recycling, last_recycling);
	  }
	}else if(set_async_link){
	  AgsChannel *start_input;
	  AgsChannel *input, *input_pad_last, *nth_input;

	  pthread_mutex_lock(audio_mutex);

	  start_input = audio->input;

	  if(start_input != NULL){
	    g_object_ref(start_input);
	  }
	  
	  pthread_mutex_unlock(audio_mutex);

	  nth_input = ags_channel_nth(start_input,
				      i);
	  g_object_unref(nth_input);
	  
	  input = nth_input;
	  
	  /* set async link */
	  if(input != NULL){
	    input_pad_last = ags_channel_pad_last(input);

	    first_recycling = input->first_recycling;
	    last_recycling = input_pad_last->last_recycling;
	    
	    ags_channel_reset_recycling(channel,
					first_recycling, last_recycling);

	    g_object_unref(input_pad_last);
	    g_object_unref(input);
	  }
	  
	  /* unref */
	  if(start_input != NULL){
	    g_object_unref(start_input);
	  }
	}
      }
    }
  }
  
  void ags_audio_set_audio_channels_shrink_zero(){
    AgsChannel *channel, *start, *channel_next;
    
    gboolean first_run;
    
    GError *error;

    pthread_mutex_lock(audio_mutex);
	  
    start =
      channel = audio->output;
    first_run = TRUE;

    error = NULL;

  ags_audio_set_audio_channel_shrink_zero0:

    while(channel != NULL){
      error =  NULL;
      ags_channel_set_link(channel, NULL,
			   &error);

      if(error != NULL){
	g_error("%s", error->message);

	g_error_free(error);
      }
      
      channel = channel->next;
    }

    channel = start;

    while(channel != NULL){
      channel_next = channel->next;

      g_object_run_dispose((GObject *) channel);
      g_object_unref((GObject *) channel);

      channel = channel_next;
    }

    if(first_run){
      start =
	channel = audio->input;
      first_run = FALSE;
      goto ags_audio_set_audio_channel_shrink_zero0;
    }

    audio->output = NULL;
    audio->input = NULL;

    pthread_mutex_unlock(audio_mutex);
  }
  
  void ags_audio_set_audio_channels_shrink(){
    AgsChannel *start;
    AgsChannel *channel, *nth_channel;
    AgsChannel *channel0, *channel1;
    AgsRecycling *recycling;

    guint pads, i, j;
    gboolean first_run;
    
    GError *error;

    pthread_mutex_lock(audio_mutex);

    start =
      channel = audio->output;
    pads = audio->output_pads;

    pthread_mutex_unlock(audio_mutex);

    first_run = TRUE;

  ags_audio_set_audio_channel_shrink0:

    for(i = 0; i < pads; i++){
      channel = ags_channel_nth(channel, audio_channels);
      g_object_unref(channel);
      
      for(j = audio_channels; j < audio_channels_old; j++){
	error = NULL;
	ags_channel_set_link(channel, NULL,
			     &error);
	
	if(error != NULL){
	  g_error("%s", error->message);

	  g_error_free(error);
	}
	
	channel = channel->next;
      }
    }

    channel = start;

    if(i < pads){
      for(i = 0; ; i++){
	for(j = 0; j < audio_channels -1; j++){
	  channel->pad = i;
	  channel->audio_channel = j;
	  channel->line = i * audio_channels + j;

	  channel = channel->next;
	}

	channel->pad = i;
	channel->audio_channel = j;
	channel->line = i * audio_channels + j;

	channel0 = channel->next;
	
	for(; j < audio_channels_old; j++){
	  channel1 = channel0->next;
      
	  g_object_run_dispose((GObject *) channel0);
	  g_object_unref((GObject *) channel0);

	  channel0 = channel1;
	}

	channel->next = channel1;

	if(channel1 != NULL){
	  channel1->prev = channel;
	}else{
	  break;
	}
	
	channel = channel1;
      }
    }

    if(first_run){
      first_run = FALSE;
      start =
	channel = audio->input;
      pads = audio->input_pads;

      goto ags_audio_set_audio_channel_shrink0;
    }
  }
    
  void ags_audio_set_audio_channels_shrink_notation(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);

    list = 
      list_start = g_list_copy(audio->notation);

    while(list != NULL){
      if(AGS_NOTATION(list->data)->audio_channel >= audio_channels){
	ags_audio_remove_notation(audio,
				  list->data);

	g_object_run_dispose((GObject *) list->data);
	g_object_unref((GObject *) list->data);
      }
      
      list = list->next;
    }

    g_list_free(list_start);

    pthread_mutex_unlock(audio_mutex);
  }

  void ags_audio_set_audio_channels_shrink_automation(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);

    list = 
      list_start = g_list_copy(audio->automation);

    while(list != NULL){
      if(AGS_AUTOMATION(list->data)->line % audio_channels_old >= audio_channels){
	ags_audio_remove_automation(audio,
				    list->data);

	g_object_run_dispose((GObject *) list->data);
	g_object_unref((GObject *) list->data);
      }
      
      list = list->next;
    }

    g_list_free(list_start);

    if(audio_channels == 0){
      audio->automation = NULL;
    }

    pthread_mutex_unlock(audio_mutex);
  }
  
  void ags_audio_set_audio_channels_shrink_wave(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);

    list = 
      list_start = g_list_copy(audio->wave);

    while(list != NULL){
      if(AGS_WAVE(list->data)->line % audio_channels_old >= audio_channels){
	ags_audio_remove_wave(audio,
			      list->data);

	g_object_run_dispose((GObject *) list->data);
	g_object_unref((GObject *) list->data);
      }
      
      list = list->next;
    }

    g_list_free(list_start);

    pthread_mutex_unlock(audio_mutex);
  }

  void ags_audio_set_audio_channels_shrink_midi(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);

    list = 
      list_start = g_list_copy(audio->midi);

    while(list != NULL){
      if(AGS_MIDI(list->data)->audio_channel >= audio_channels){
	ags_audio_remove_midi(audio,
			      list->data);

	g_object_run_dispose((GObject *) list->data);
	g_object_unref((GObject *) list->data);
      }
      
      list = list->next;
    }

    g_list_free(list_start);

    pthread_mutex_unlock(audio_mutex);
  }
  
  /*
   * entry point
   */
  if(audio_channels == audio_channels_old){
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* grow / shrink */
  pthread_mutex_lock(audio_mutex);

  audio_flags = audio->flags;

  output_soundcard = audio->output_soundcard;
  input_soundcard = audio->input_soundcard;
  
  samplerate = audio->samplerate;
  buffer_size = audio->buffer_size;
  format = audio->format;

  output_pads = audio->output_pads;
  input_pads = audio->input_pads;
  
  pthread_mutex_unlock(audio_mutex);

  if(audio_channels > audio_channels_old){
    AgsPlaybackDomain *playback_domain;
    AgsPlayback *playback;
    AgsChannel *current;

    guint i;
    
    pthread_mutex_t *current_mutex;
    pthread_mutex_t *playback_domain_mutex;
    
    /* grow input channels */
    if(audio->input_pads > 0 &&
       (AGS_AUDIO_NO_INPUT & (audio_flags)) == 0){
      ags_audio_set_audio_channels_init_parameters(AGS_TYPE_INPUT);
      ags_audio_set_audio_channels_grow(AGS_TYPE_INPUT);
    }

    /* apply new sizes */
    pthread_mutex_lock(audio_mutex);
    
    audio->input_lines = audio_channels * audio->input_pads;

    pthread_mutex_unlock(audio_mutex);

    /* grow output channels */
    if(audio->output_pads > 0 &&
       (AGS_AUDIO_NO_OUTPUT & (audio_flags)) == 0){
      ags_audio_set_audio_channels_init_parameters(AGS_TYPE_OUTPUT);
      ags_audio_set_audio_channels_grow(AGS_TYPE_OUTPUT);
    }

    /* apply new sizes and get some fields */
    pthread_mutex_lock(audio_mutex);
    
    audio->output_lines = audio_channels * audio->output_pads;

    audio->audio_channels = audio_channels;

    playback_domain = AGS_PLAYBACK_DOMAIN(audio->playback_domain);

    pthread_mutex_unlock(audio_mutex);

    /* get playback domain mutex */
    playback_domain_mutex = AGS_PLAYBACK_DOMAIN_GET_OBJ_MUTEX(playback_domain);

    /* add output playback */
    pthread_mutex_lock(audio_mutex);
    
    current = audio->output;
    
    pthread_mutex_unlock(audio_mutex);

    while(current != NULL){
      current = ags_channel_nth(current,
				audio_channels_old);
      g_object_unref(current);
      
      for(i = 0; i < audio_channels - audio_channels_old; i++){
	/* get current mutex */
	current_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(current);

	/* playback */
	pthread_mutex_lock(current_mutex);

	playback = (AgsPlayback *) current->playback;

	pthread_mutex_unlock(current_mutex);

	/* append */
	ags_playback_domain_add_playback(playback_domain,
					 (GObject *) playback, AGS_TYPE_OUTPUT);
	
	/* iterate */
	pthread_mutex_lock(current_mutex);

	current = current->next;

	pthread_mutex_unlock(current_mutex);
      }
    }

    /* add input playback */
    pthread_mutex_lock(audio_mutex);
    
    current = audio->input;
    
    pthread_mutex_unlock(audio_mutex);

    while(current != NULL){
      current = ags_channel_nth(current,
				audio_channels_old);
      g_object_unref(current);
      
      for(i = 0; i < audio_channels - audio_channels_old; i++){
	/* get current mutex */
	current_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(current);

	/* playback */
	pthread_mutex_lock(current_mutex);

	playback = (AgsPlayback *) current->playback;

	pthread_mutex_unlock(current_mutex);

	/* append */
	ags_playback_domain_add_playback(playback_domain,
					 (GObject *) playback, AGS_TYPE_INPUT);

	/* iterate */
	pthread_mutex_lock(current_mutex);

	current = current->next;

	pthread_mutex_unlock(current_mutex);
      }
    }
  }else if(audio_channels < audio_channels_old){
    AgsPlaybackDomain *playback_domain;
    AgsPlayback *playback;
    AgsChannel *current;

    GList *list, *list_start;
    
    guint i, j;
    
    pthread_mutex_t *playback_domain_mutex;

    /* shrink audio channels */
    ags_audio_set_audio_channels_shrink_automation();
    
    if((AGS_SOUND_ABILITY_NOTATION & (audio->ability_flags)) != 0){
      ags_audio_set_audio_channels_shrink_notation();
    }

    if((AGS_SOUND_ABILITY_WAVE & (audio->ability_flags)) != 0){
      ags_audio_set_audio_channels_shrink_wave();
    }

    if((AGS_SOUND_ABILITY_MIDI & (audio->ability_flags)) != 0){
      ags_audio_set_audio_channels_shrink_midi();
    }

    if(audio_channels == 0){
      ags_audio_set_audio_channels_shrink_zero();
    }else{
      ags_audio_set_audio_channels_shrink();
    }

    /* apply new sizes */
    pthread_mutex_lock(audio_mutex);

    audio->audio_channels = audio_channels;
    
    audio->input_lines = audio_channels * audio->input_pads;
    audio->output_lines = audio_channels * audio->output_pads;

    playback_domain = AGS_PLAYBACK_DOMAIN(audio->playback_domain);

    pthread_mutex_unlock(audio_mutex);

    /* get playback domain mutex */
    playback_domain_mutex = AGS_PLAYBACK_DOMAIN_GET_OBJ_MUTEX(playback_domain);

    /* shrink output playback domain */
    pthread_mutex_lock(playback_domain_mutex);

    list_start = g_list_copy(playback_domain->output_playback);
    
    pthread_mutex_unlock(playback_domain_mutex);
    
    for(j = 0; j < output_pads; j++){
      list = g_list_nth(list_start,
			(j + 1) * audio_channels);
      
      for(i = 0; i < audio_channels_old - audio_channels && list != NULL; i++){
	playback = list->data;

	if(AGS_IS_PLAYBACK(playback)){
	  ags_playback_domain_remove_playback(playback_domain,
					      (GObject *) playback, AGS_TYPE_OUTPUT);

	  g_object_run_dispose((GObject *) playback);
	  g_object_unref((GObject *) playback);
	}
	
	/* iterate */
	list = list->next;
      }
    }

    g_list_free(list_start);

    /* shrink input playback domain */
    pthread_mutex_lock(playback_domain_mutex);

    list_start = g_list_copy(playback_domain->input_playback);
    
    pthread_mutex_unlock(playback_domain_mutex);
    
    for(j = 0; j < input_pads; j++){
      list = g_list_nth(list_start,
			(j + 1) * audio_channels);
      
      for(i = 0; i < audio_channels_old - audio_channels; i++){
	playback = list->data;

	if(AGS_IS_PLAYBACK(playback)){
	  ags_playback_domain_remove_playback(playback_domain,
					      (GObject *) playback, AGS_TYPE_INPUT);
	  
	  g_object_run_dispose((GObject *) playback);
	  g_object_unref((GObject *) playback);
	}
	
	/* iterate */
	list = list->next;
      }
    }

    g_list_free(list_start);
  }
  
  /* emit message */
  message_delivery = ags_message_delivery_get_instance();

  message_queue = (AgsMessageQueue *) ags_message_delivery_find_namespace(message_delivery,
									  "libags-audio");

  if(message_queue != NULL){
    AgsMessageEnvelope *message;

    xmlDoc *doc;
    xmlNode *root_node;

    /* specify message body */
    doc = xmlNewDoc("1.0");

    root_node = xmlNewNode(NULL,
			   "ags-command");
    xmlDocSetRootElement(doc, root_node);    

    xmlNewProp(root_node,
	       "method",
	       "AgsAudio::set-audio-channels");

    /* add message */
    message = ags_message_envelope_alloc((GObject *) audio,
					 NULL,
					 doc);

    /* set parameter */
    message->n_params = 2;

    message->parameter_name = (gchar **) malloc(3 * sizeof(gchar *));
    message->value = g_new0(GValue,
			    2);

    /* audio channels */
    message->parameter_name[0] = "audio-channels";
    
    g_value_init(&(message->value[0]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[0]),
		     audio_channels);

    /* audio channels old */
    message->parameter_name[1] = "audio-channels-old";
    
    g_value_init(&(message->value[1]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[1]),
		     audio_channels_old);

    /* terminate string vector */
    message->parameter_name[2] = NULL;
    
    /* add message */
    ags_message_delivery_add_message(message_delivery,
				     "libags-audio",
				     message);
  }
}

/**
 * ags_audio_set_audio_channels:
 * @audio: the #AgsAudio
 * @audio_channels: new audio channels
 * @audio_channels_old: old audio channels
 *
 * Resize audio channels AgsInput will be allocated first.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_audio_channels(AgsAudio *audio,
			     guint audio_channels, guint audio_channels_old)
{
  pthread_mutex_t *audio_mutex;

  g_return_if_fail(AGS_IS_AUDIO(audio));

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* lock */
  pthread_mutex_lock(audio_mutex);

  audio_channels_old = audio->audio_channels;

  pthread_mutex_unlock(audio_mutex);

  /* emit */
  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[SET_AUDIO_CHANNELS], 0,
		audio_channels, audio_channels_old);
  g_object_unref((GObject *) audio);

  /* unlock */
  //TODO:JK: remove comment above
}

/*
 * resize
 * AgsInput has to be allocated first
 */
void
ags_audio_real_set_pads(AgsAudio *audio,
			GType channel_type,
			guint pads, guint pads_old)
{
  AgsChannel *channel;
  AgsPlaybackDomain *playback_domain;
  
  AgsMessageDelivery *message_delivery;
  AgsMessageQueue *message_queue;

  GObject *output_soundcard;
  GObject *input_soundcard;
  
  guint bank_dim[3];

  guint audio_flags;
  guint audio_channels;
  guint output_pads, input_pads;
  guint samplerate;
  guint buffer_size;
  guint format;
  gboolean alloc_recycling, link_recycling, set_sync_link, set_async_link;
  gboolean alloc_pattern;
  gboolean add_synth_generator;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *current_mutex;
  pthread_mutex_t *prev_mutex, *prev_pad_mutex;
  pthread_mutex_t *playback_domain_mutex;

  auto void ags_audio_set_pads_init_parameters();
  auto void ags_audio_set_pads_grow();
  auto void ags_audio_set_pads_unlink_all(AgsChannel *channel);
  auto void ags_audio_set_pads_shrink_zero(AgsChannel *channel);
  auto void ags_audio_set_pads_shrink(AgsChannel *channel);
  auto void ags_audio_set_pads_free_notation();
  auto void ags_audio_set_pads_remove_notes();
  auto void ags_audio_set_pads_shrink_automation();
  auto void ags_audio_set_pads_free_wave();
  auto void ags_audio_set_pads_free_midi();
  
  void ags_audio_set_pads_init_parameters(){
    alloc_recycling = FALSE;
    link_recycling = FALSE;
    set_sync_link = FALSE;
    set_async_link = FALSE;

    alloc_pattern = FALSE;
    add_synth_generator = FALSE;
    
    if(channel_type == AGS_TYPE_OUTPUT){
      if((AGS_AUDIO_OUTPUT_HAS_RECYCLING & (audio_flags)) != 0){
	alloc_recycling = TRUE;
      }else{
	if((AGS_AUDIO_INPUT_HAS_RECYCLING & (audio_flags)) != 0){
	  if((AGS_AUDIO_SYNC & (audio_flags)) != 0 && (AGS_AUDIO_ASYNC & (audio_flags)) == 0){
	    set_async_link = TRUE;
	  }else if((AGS_AUDIO_ASYNC & (audio_flags)) != 0){
	    set_async_link = TRUE;
	  }else{
#ifdef AGS_DEBUG
	    g_message("ags_audio_set_pads - warning: AGS_AUDIO_SYNC nor AGS_AUDIO_ASYNC weren't defined");
#endif
	  }
	}
      }
    }else{      
      if((AGS_AUDIO_INPUT_HAS_SYNTH & (audio_flags)) != 0){
	add_synth_generator = TRUE;
      }

      if((AGS_AUDIO_INPUT_HAS_RECYCLING & (audio_flags)) != 0){
	alloc_recycling = TRUE;
      }
      
      if((AGS_AUDIO_ASYNC & (audio_flags)) != 0 && alloc_recycling){
	link_recycling = TRUE;
      }

      pthread_mutex_lock(audio_mutex);
      
      if((AGS_SOUND_ABILITY_SEQUENCER & (audio->ability_flags)) != 0){
	alloc_pattern = TRUE;

	bank_dim[0] = audio->bank_dim[0];
	bank_dim[1] = audio->bank_dim[1];
	bank_dim[2] = audio->bank_dim[2];
      }      

      pthread_mutex_unlock(audio_mutex);
    }    
  }

  void ags_audio_set_pads_grow(){
    AgsChannel *start;
    AgsChannel *channel, *nth_channel;
    AgsRecycling *first_recycling, *last_recycling;
    AgsPlaybackDomain *playback_domain;
    
    guint i, j;

    pthread_mutex_lock(audio_mutex);

    playback_domain = (AgsPlaybackDomain *) audio->playback_domain;
    
    if(channel_type == AGS_TYPE_OUTPUT){
      start = audio->output;
    }else{
      start = audio->input;
    }

    pthread_mutex_unlock(audio_mutex);

    for(j = pads_old; j < pads; j++){
      AgsPlayback *playback;
      
      for(i = 0; i < audio_channels; i++){
	channel = (AgsChannel *) g_object_new(channel_type,
					      "audio", (GObject *) audio,
					      "output-soundcard", output_soundcard,
					      "output-soundcard-channel", i,
					      "input-soundcard", input_soundcard,
					      "input-soundcard-channel", i,
					      "samplerate", samplerate,
					      "buffer-size", buffer_size,
					      "format", format,
					      NULL);
	g_object_ref(channel);

	playback = (AgsPlayback *) channel->playback;
	g_object_set(playback,
		     "playback-domain", playback_domain,
		     NULL);

	if(alloc_pattern){
	  channel->pattern = g_list_alloc();
	  channel->pattern->data = (gpointer) ags_pattern_new();
	  g_object_ref(channel->pattern->data);
	  
	  ags_pattern_set_dim((AgsPattern *) channel->pattern->data,
			      bank_dim[0], bank_dim[1], bank_dim[2]);
	}

	if(add_synth_generator){
	  AgsSynthGenerator *synth_generator;

	  synth_generator = ags_synth_generator_new();
	  ags_input_add_synth_generator((AgsInput *) channel,
					(GObject *) synth_generator);
	}
	
	if(start == NULL){
	  /* set first channel in AgsAudio */
	  pthread_mutex_lock(audio_mutex);
	  
	  if(channel_type == AGS_TYPE_OUTPUT){
	    start = 
	      audio->output = channel;
	  }else{
	    start = 
	      audio->input = channel;
	  }

	  pthread_mutex_unlock(audio_mutex);
	}

	if(j * audio_channels + i != 0){
	  /* set prev */
	  channel->prev = ags_channel_nth(start, j * audio_channels + i - 1);
	  g_object_unref(channel->prev);
	  
	  /* get prev mutex */
	  prev_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel->prev);

	  /* set prev->next */
	  pthread_mutex_lock(prev_mutex);
	  
	  channel->prev->next = channel;

	  pthread_mutex_unlock(prev_mutex);
	}
	
	if(j != 0){
	  /* set prev pad */
	  nth_channel = ags_channel_nth(start, i);
	  
	  channel->prev_pad = ags_channel_pad_nth(nth_channel, j - 1);

	  g_object_unref(nth_channel);
	  g_object_unref(channel->prev_pad);
	  
	  /* get prev pad mutex */
	  prev_pad_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel->prev_pad);

	  /* set prev_pad->next_pad */
	  pthread_mutex_lock(prev_pad_mutex);
	  
	  channel->prev_pad->next_pad = channel;
	  
	  pthread_mutex_unlock(prev_pad_mutex);
	}

	/* set indices */
	channel->pad = j;
	channel->audio_channel = i;
	channel->line = j * audio_channels + i;

	/* reset nested AgsRecycling tree */
	if(alloc_recycling){
	  first_recycling =
	    last_recycling = ags_recycling_new(output_soundcard);
	  g_object_ref(first_recycling);
	  g_object_set(first_recycling,
		       "channel", channel,
		       NULL);
	  
	  ags_channel_reset_recycling(channel,
				      first_recycling, last_recycling);
	}else if(set_sync_link){
	  AgsChannel *input;

	  pthread_mutex_lock(audio_mutex);

	  input = audio->input;
	  
	  pthread_mutex_unlock(audio_mutex);

	  input = ags_channel_nth(input,
				  channel->line);
	  g_object_unref(input);
	  
	  /* set sync link */
	  if(input != NULL){
	    first_recycling = input->first_recycling;
	    last_recycling = input->last_recycling;
	    
	    ags_channel_reset_recycling(channel,
					first_recycling, last_recycling);
	  }
	}else if(set_async_link){
	  AgsChannel *start_input;
	  AgsChannel *input, *input_pad_last, *nth_input;

	  pthread_mutex_lock(audio_mutex);

	  start_input = audio->input;

	  if(start_input != NULL){
	    g_object_ref(start_input);
	  }
	  
	  pthread_mutex_unlock(audio_mutex);

	  nth_input = ags_channel_nth(start_input,
				      i);
	  g_object_unref(nth_input);
	  
	  input = nth_input;

	  /* set async link */
	  if(input != NULL){
	    input_pad_last = ags_channel_pad_last(input);

	    first_recycling = input->first_recycling;
	    last_recycling = input_pad_last->last_recycling;
	    
	    ags_channel_reset_recycling(channel,
					first_recycling, last_recycling);

	    g_object_unref(input_pad_last);
	    g_object_unref(input);
	  }

	  /* unref */
	  if(start_input != NULL){
	    g_object_unref(start_input);
	  }
	}
      }
    }
  }

  void ags_audio_set_pads_unlink_all(AgsChannel *channel){
    GError *error;

    while(channel != NULL){
      error = NULL;
      ags_channel_set_link(channel, NULL,
			   &error);

      if(error != NULL){
	g_error("%s", error->message);

	g_error_free(error);
      }

      channel = channel->next;
    }
  }
  
  void ags_audio_set_pads_shrink_zero(AgsChannel *channel){
    AgsChannel *channel_next;

    GError *error;
    
    while(channel != NULL){
      channel_next = channel->next;

#if 0
      error = NULL;
      ags_channel_set_link(channel, NULL,
			   &error);

      if(error != NULL){
	g_error("%s", error->message);

	g_error_free(error);
      }
#endif
      
      g_object_run_dispose((GObject *) channel);
      g_object_unref((GObject *) channel);

      channel = channel_next;
    }
  }

  void ags_audio_set_pads_shrink(AgsChannel *channel){
    AgsChannel *current;

    guint i;

    current = channel;
    
    if(channel != NULL &&
       channel->prev_pad != NULL){
      channel = channel->prev_pad;
    }else{
      channel = NULL;
    }
    
    ags_audio_set_pads_shrink_zero(current);

    /* remove pads */
    if(channel != NULL){
      current = channel;
      
      for(i = 0; i < audio_channels; i++){
	current->next_pad = NULL;
	
	/* iterate */
	current = current->next;
      }

      /* remove channel */
      current = ags_channel_nth(channel,
				audio_channels - 1);
      g_object_unref(current);
      
      current->next = NULL;
    }
  }

  void ags_audio_set_pads_free_notation(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);
    
    list = 
      list_start = g_list_copy(audio->notation);
    
    while(list != NULL){
      ags_audio_remove_notation(audio,
				list->data);
      
      g_object_run_dispose((GObject *) list->data);
      g_object_unref((GObject *) list->data);
      
      list = list->next;
    }

    g_list_free(audio->notation);
    g_list_free(list_start);
    
    audio->notation = NULL;

    pthread_mutex_unlock(audio_mutex);
  }
  
  void ags_audio_set_pads_remove_notes(){
    GList *notation;
    GList *note_start, *note;

    pthread_mutex_lock(audio_mutex);

    notation = audio->notation;

    while(notation != NULL){
      note =
	note_start = g_list_copy(AGS_NOTATION(notation->data)->note);

      while(note != NULL){
	if(AGS_NOTE(note->data)->y >= pads){
	  AGS_NOTATION(notation->data)->note = g_list_remove(AGS_NOTATION(notation->data)->note,
							     note->data);
	  
	  g_object_unref(note->data);
	}

	note = note->next;
      }

      g_list_free(note_start);
      
      notation = notation->next;
    }

    pthread_mutex_unlock(audio_mutex);
  }

  void ags_audio_set_pads_shrink_automation(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);

    list = 
      list_start = g_list_copy(audio->automation);
    
    while(list != NULL){
      if(AGS_AUTOMATION(list->data)->channel_type == channel_type &&
	 AGS_AUTOMATION(list->data)->line >= pads * audio_channels){
	ags_audio_remove_automation(audio,
				    list->data);
	
	g_object_run_dispose((GObject *) list->data);
	g_object_unref((GObject *) list->data);
      }
      
      list = list->next;
    }

    g_list_free(list_start);

    pthread_mutex_unlock(audio_mutex);
  }

  void ags_audio_set_pads_free_wave(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);

    list = 
      list_start = g_list_copy(audio->wave);
    
    while(list != NULL){
      ags_audio_remove_wave(audio,
			    list->data);
      
      g_object_run_dispose((GObject *) list->data);
      g_object_unref((GObject *) list->data);
      
      list = list->next;
    }

    g_list_free(audio->wave);
    g_list_free(list_start);
    
    audio->wave = NULL;

    pthread_mutex_unlock(audio_mutex);
  }
  
  void ags_audio_set_pads_free_midi(){
    GList *list_start, *list;

    pthread_mutex_lock(audio_mutex);
    
    list = 
      list_start = g_list_copy(audio->midi);
    
    while(list != NULL){
      ags_audio_remove_midi(audio,
			    list->data);

      g_object_run_dispose((GObject *) list->data);
      g_object_unref((GObject *) list->data);
      
      list = list->next;
    }

    g_list_free(audio->midi);
    g_list_free(list_start);
    
    audio->midi = NULL;

    pthread_mutex_unlock(audio_mutex);
  }

  if(!(g_type_is_a(channel_type,
		   AGS_TYPE_OUTPUT) ||
       g_type_is_a(channel_type,
		   AGS_TYPE_INPUT))){
    g_warning("unknown channel type");

    return;
  }
  
  /*
   * entry point
   */
  if(pads_old == pads){
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get some fields */
  pthread_mutex_lock(audio_mutex);

  audio_flags = audio->flags;
  
  audio_channels = audio->audio_channels;

  output_soundcard = audio->output_soundcard;
  input_soundcard = audio->input_soundcard;
  
  samplerate = audio->samplerate;
  buffer_size = audio->buffer_size;
  format = audio->format;
  
  output_pads = audio->output_pads;
  input_pads = audio->input_pads;
  
  playback_domain = AGS_PLAYBACK_DOMAIN(audio->playback_domain);
  
  pthread_mutex_unlock(audio_mutex);

  /* get playback domain mutex */
  playback_domain_mutex = AGS_PLAYBACK_DOMAIN_GET_OBJ_MUTEX(playback_domain);
  
  /* init boolean parameters */
  ags_audio_set_pads_init_parameters();

  if(g_type_is_a(channel_type,
		 AGS_TYPE_OUTPUT) &&
     (AGS_AUDIO_NO_OUTPUT & (audio_flags)) == 0){
    /* output */
    if(audio_channels != 0){
      /* grow or shrink */
      if(pads > output_pads){
	AgsChannel *current;

	guint i, j;

	/* grow channels */
	ags_audio_set_pads_grow();

	/* get some fields */
	pthread_mutex_lock(audio_mutex);
	  
	current = audio->output;

	pthread_mutex_unlock(audio_mutex);

	/* alloc playback domain */
	current = ags_channel_pad_nth(current,
				      pads_old);

	g_object_unref(current);
	
	for(j = pads_old; j < pads; j++){
	  for(i = 0; i < audio_channels; i++){
	    AgsPlayback *playback;
	    
	    /* get current mutex */
	    current_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(current);

	    /* playback */
	    pthread_mutex_lock(current_mutex);

	    playback = (AgsPlayback *) current->playback;

	    pthread_mutex_unlock(current_mutex);

	    /* append */
	    ags_playback_domain_add_playback(playback_domain,
					     (GObject *) playback, AGS_TYPE_OUTPUT);
	
	    /* iterate */
	    pthread_mutex_lock(current_mutex);

	    current = current->next;

	    pthread_mutex_unlock(current_mutex);
	  }
	}
      }else if(pads == 0){	
	GList *list, *list_start;
      
	ags_audio_set_pads_shrink_automation();      

	if((AGS_SOUND_BEHAVIOUR_DEFAULTS_TO_OUTPUT & (audio->behaviour_flags)) != 0){
	  if((AGS_SOUND_ABILITY_NOTATION & (audio->ability_flags)) != 0){
	    ags_audio_set_pads_free_notation();
	  }

	  if((AGS_SOUND_ABILITY_WAVE & (audio->ability_flags)) != 0){
	    ags_audio_set_pads_free_wave();
	  }

	  if((AGS_SOUND_ABILITY_MIDI & (audio->ability_flags)) != 0){
	    ags_audio_set_pads_free_midi();
	  }
	}
	
	pthread_mutex_lock(audio_mutex);

	channel = audio->output;

	pthread_mutex_unlock(audio_mutex);
      
	/* unlink and remove */
	ags_audio_set_pads_unlink_all(channel);
	ags_audio_set_pads_shrink_zero(channel);
      
	pthread_mutex_lock(audio_mutex);
	
	audio->output = NULL;

	pthread_mutex_unlock(audio_mutex);

	/* remove playback */
	pthread_mutex_lock(playback_domain_mutex);

	list = 
	  list_start = g_list_copy(playback_domain->output_playback);

	pthread_mutex_unlock(playback_domain_mutex);

	while(list != NULL){
	  AgsPlayback *playback;

	  playback = (AgsPlayback *) list->data;

	  if(AGS_IS_PLAYBACK(playback)){
	    /* remove playback */
	    ags_playback_domain_remove_playback(playback_domain,
						(GObject *) playback, AGS_TYPE_OUTPUT);

	    g_object_run_dispose((GObject *) playback);
	    g_object_unref((GObject *) playback);
	  }
	  
	  /* iterate */
	  list = list->next;
	}

	g_list_free(list_start);
      }else if(pads < output_pads){
	GList *list, *list_start;

	/* get some fields */
	pthread_mutex_lock(audio_mutex);
	
	channel = audio->output;

	pthread_mutex_unlock(audio_mutex);

	ags_audio_set_pads_shrink_automation();      
	ags_audio_set_pads_remove_notes();

	channel = ags_channel_pad_nth(channel,
				      pads);

	g_object_unref(channel);
	
	/* remove playback */
	pthread_mutex_lock(playback_domain_mutex);
	  
	list = 
	  list_start = g_list_copy(playback_domain->output_playback);

	pthread_mutex_unlock(playback_domain_mutex);
	  
	while(list != NULL){
	  AgsChannel *current;
	  AgsPlayback *playback;

	  guint current_line;

	  playback = (AgsPlayback *) list->data;

	  g_object_get(playback,
		       "channel", &current,
		       NULL);
	  g_object_get(current,
		       "line", &current_line,
		       NULL);

	  if(current_line >= pads * audio_channels &&
	     AGS_IS_PLAYBACK(playback)){
	    /* remove playback */
	    ags_playback_domain_remove_playback(playback_domain,
						(GObject *) playback, AGS_TYPE_OUTPUT);
	    
	    g_object_run_dispose((GObject *) playback);
	    g_object_unref((GObject *) playback);
	  }

	  g_object_unref(current);
	  
	  /* iterate */
	  list = list->next;
	}

	g_list_free(list_start);

	ags_audio_set_pads_unlink_all(channel);
	ags_audio_set_pads_shrink(channel);
      }
    }
    
    /* apply new size */
    audio->output_pads = pads;
    audio->output_lines = pads * audio_channels;

    //    if((AGS_AUDIO_SYNC & audio_flags) != 0 && (AGS_AUDIO_ASYNC & audio_flags) == 0){
    //TODO:JK: fix me
    //      input_pads = pads;
    //      audio->input_lines = pads * audio_channels;
    //    }
  }else if(g_type_is_a(channel_type,
		       AGS_TYPE_INPUT) &&
	   (AGS_AUDIO_NO_INPUT & (audio_flags)) == 0){
    /* input */
    if(audio_channels != 0){
      /* grow or shrink */
      if(pads > pads_old){
	AgsChannel *current;

	guint i, j;

	/* grow channels */
	ags_audio_set_pads_grow();

	/* get some fields */
	pthread_mutex_lock(audio_mutex);
	  
	current = audio->input;

	pthread_mutex_unlock(audio_mutex);

	/* alloc playback domain */
	current = ags_channel_pad_nth(current,
				      pads_old);

	g_object_unref(current);
	
	for(j = pads_old; j < pads; j++){
	  for(i = 0; i < audio_channels; i++){
	    AgsPlayback *playback;
	    
	    /* get current mutex */
	    current_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(current);

	    /* playback */
	    pthread_mutex_lock(current_mutex);

	    playback = (AgsPlayback *) current->playback;

	    pthread_mutex_unlock(current_mutex);

	    /* append */
	    ags_playback_domain_add_playback(playback_domain,
					     (GObject *) playback, AGS_TYPE_INPUT);
	
	    /* iterate */
	    pthread_mutex_lock(current_mutex);

	    current = current->next;

	    pthread_mutex_unlock(current_mutex);
	  }
	}
      }else if(pads == 0){
	GList *list, *list_start;
	
	ags_audio_set_pads_shrink_automation();

	if((AGS_SOUND_BEHAVIOUR_DEFAULTS_TO_INPUT & (audio->behaviour_flags)) != 0){
	  if((AGS_SOUND_ABILITY_NOTATION & (audio->ability_flags)) != 0){
	    ags_audio_set_pads_free_notation();
	  }

	  if((AGS_SOUND_ABILITY_WAVE & (audio->ability_flags)) != 0){
	    ags_audio_set_pads_free_wave();
	  }

	  if((AGS_SOUND_ABILITY_MIDI & (audio->ability_flags)) != 0){
	    ags_audio_set_pads_free_midi();
	  }
	}

	pthread_mutex_lock(audio_mutex);

	channel = audio->input;

	pthread_mutex_unlock(audio_mutex);

	/* shrink channels */
	ags_audio_set_pads_unlink_all(channel);
	ags_audio_set_pads_shrink_zero(channel);
	
	pthread_mutex_lock(audio_mutex);

	audio->input = NULL;  

	pthread_mutex_unlock(audio_mutex);

	/* remove playback */
	pthread_mutex_lock(playback_domain_mutex);

	list = 
	  list_start = g_list_copy(playback_domain->input_playback);

	pthread_mutex_unlock(playback_domain_mutex);

	while(list != NULL){
	  AgsPlayback *playback;

	  playback = (AgsPlayback *) list->data;

	  if(AGS_IS_PLAYBACK(playback)){
	    /* remove playback */
	    ags_playback_domain_remove_playback(playback_domain,
						(GObject *) playback, AGS_TYPE_INPUT);

	    g_object_run_dispose((GObject *) playback);
	    g_object_unref((GObject *) playback);
	  }
	  
	  /* iterate */
	  list = list->next;
	}

	g_list_free(list_start);
      }else if(pads < pads_old){
	GList *list, *list_start;

	/* get some fields */
	pthread_mutex_lock(audio_mutex);

	channel = audio->input;

	pthread_mutex_unlock(audio_mutex);
	
	ags_audio_set_pads_shrink_automation();
      
	channel = ags_channel_pad_nth(channel,
				      pads);

	g_object_unref(channel);
		
	/* remove playback */
	pthread_mutex_lock(playback_domain_mutex);
	  
	list = 
	  list_start = g_list_copy(playback_domain->input_playback);

	pthread_mutex_unlock(playback_domain_mutex);
	  
	while(list != NULL){
	  AgsChannel *current;
	  AgsPlayback *playback;

	  guint current_line;

	  playback = (AgsPlayback *) list->data;

	  g_object_get(playback,
		       "channel", &current,
		       NULL);
	  g_object_get(current,
		       "line", &current_line,
		       NULL);

	  if(current_line >= pads * audio_channels &&
	     AGS_IS_PLAYBACK(playback)){
	    /* remove playback */
	    ags_playback_domain_remove_playback(playback_domain,
						(GObject *) playback, AGS_TYPE_INPUT);
	    
	    g_object_run_dispose((GObject *) playback);
	    g_object_unref((GObject *) playback);
	  }

	  g_object_unref(current);
	  
	  /* iterate */
	  list = list->next;
	}

	g_list_free(list_start);

	/* shrink channels */
	ags_audio_set_pads_unlink_all(channel);
	ags_audio_set_pads_shrink(channel);
      }
    }

    /* apply new allocation */
    audio->input_pads = pads;
    audio->input_lines = pads * audio_channels;
  }
  
  /* emit message */
  message_delivery = ags_message_delivery_get_instance();

  message_queue = (AgsMessageQueue *) ags_message_delivery_find_namespace(message_delivery,
									  "libags-audio");

  if(message_queue != NULL){
    AgsMessageEnvelope *message;

    xmlDoc *doc;
    xmlNode *root_node;

    /* specify message body */
    doc = xmlNewDoc("1.0");

    root_node = xmlNewNode(NULL,
			   "ags-command");
    xmlDocSetRootElement(doc, root_node);    

    xmlNewProp(root_node,
	       "method",
	       "AgsAudio::set-pads");

    /* add message */
    message = ags_message_envelope_alloc((GObject *) audio,
					 NULL,
					 doc);
    /* set parameter */
    message->n_params = 3;

    message->parameter_name = (gchar *) malloc(4 * sizeof(gchar *));
    message->value = g_new0(GValue,
			    3);

    /* channel type */
    message->parameter_name[0] = "channel-type";
    
    g_value_init(&(message->value[0]),
		 G_TYPE_ULONG);
    g_value_set_ulong(&(message->value[0]),
		      channel_type);

    /* pads */
    message->parameter_name[1] = "pads";
    
    g_value_init(&(message->value[1]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[1]),
		     pads);

    /* pads old */
    message->parameter_name[2] = "pads-old";
    
    g_value_init(&(message->value[2]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[2]),
		     pads_old);

    /* terminate string vector */
    message->parameter_name[3] = NULL;
    
    /* add message */
    ags_message_delivery_add_message(message_delivery,
				     "libags-audio",
				     message);
  }
}

/**
 * ags_audio_set_pads:
 * @audio: the #AgsAudio
 * @channel_type: AGS_TYPE_INPUT or AGS_TYPE_OUTPUT
 * @pads: new pad count
 * @pads_old: old pad count
 *
 * Set pad count for the apropriate @channel_type
 *
 * Since: 2.0.0
 */
void
ags_audio_set_pads(AgsAudio *audio,
		   GType channel_type,
		   guint pads, guint pads_old)
{
  pthread_mutex_t *audio_mutex;
  
  g_return_if_fail(AGS_IS_AUDIO(audio));

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);
  
  /* emit */
  pthread_mutex_lock(audio_mutex);

  pads_old = ((g_type_is_a(channel_type, AGS_TYPE_OUTPUT)) ? audio->output_pads: audio->input_pads);

  pthread_mutex_unlock(audio_mutex);

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[SET_PADS], 0,
		channel_type,
		pads, pads_old);
  g_object_unref((GObject *) audio);  
}

/**
 * ags_audio_set_output_soundcard:
 * @audio: the #AgsAudio
 * @output_soundcard: an #AgsSoundcard
 *
 * Set the output soundcard object of @audio.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_output_soundcard(AgsAudio *audio,
			       GObject *output_soundcard)
{
  AgsChannel *channel;
  AgsPlaybackDomain *playback_domain;
  
  AgsThread *audio_thread;
  
  GObject *old_soundcard;
  
  GList *list;

  guint samplerate;
  guint buffer_size;
  guint format;
  guint i;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;
  pthread_mutex_t *playback_domain_mutex;
  pthread_mutex_t *play_mutex, *recall_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);
  
  /* old soundcard */
  pthread_mutex_lock(audio_mutex);

  old_soundcard = audio->output_soundcard;

  pthread_mutex_unlock(audio_mutex);
  
  if(old_soundcard == output_soundcard){
    return;
  }

  /* ref and set new soundcard */
  if(output_soundcard != NULL){
    g_object_ref(output_soundcard);
  }
  
  pthread_mutex_lock(audio_mutex);
  
  audio->output_soundcard = (GObject *) output_soundcard;

  pthread_mutex_unlock(audio_mutex);

  if(output_soundcard != NULL){
    /* get presets */    
    ags_soundcard_get_presets(AGS_SOUNDCARD(output_soundcard),
			      NULL,
			      &samplerate,
			      &buffer_size,
			      &format);

    /* apply presets */
    g_object_set(audio,
		 "samplerate", samplerate,
		 "buffer-size", buffer_size,
		 "format", format,
		 NULL);
  }

  /* output */
  pthread_mutex_lock(audio_mutex);
  
  channel = audio->output;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* reset */
    g_object_set(G_OBJECT(channel),
		 "output-soundcard", output_soundcard,
		 NULL);

    /* iterate */
    pthread_mutex_lock(channel_mutex);

    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }
  
  /* input */
  pthread_mutex_lock(audio_mutex);
  
  channel = audio->input;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* reset */
    g_object_set(G_OBJECT(channel),
		 "output-soundcard", output_soundcard,
		 NULL);

    /* iterate */
    pthread_mutex_lock(channel_mutex);

    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }
  
  /* playback domain - audio thread */
  pthread_mutex_lock(audio_mutex);

  playback_domain = AGS_PLAYBACK_DOMAIN(audio->playback_domain);

  pthread_mutex_unlock(audio_mutex);

  /* get playback domain mutex */
  playback_domain_mutex = AGS_PLAYBACK_DOMAIN_GET_OBJ_MUTEX(playback_domain);

  /* audio thread - output soundcard */
  for(i = 0; i < AGS_SOUND_SCOPE_LAST; i++){
    pthread_mutex_lock(playback_domain_mutex);
    
    audio_thread = playback_domain->audio_thread[i];

    pthread_mutex_unlock(playback_domain_mutex);
    
    if(audio_thread != NULL){
      /* set output soundcard */
      g_object_set(audio_thread,
		   "default-output-soundcard", output_soundcard,
		   NULL);
    }
  }
  
  /* get play mutex */
  play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

  /* play context */
  pthread_mutex_lock(play_mutex);
  
  list = audio->play;
  
  while(list != NULL){
    g_object_set(G_OBJECT(list->data),
		 "output-soundcard", output_soundcard,
		 NULL);
    
    list = list->next;
  }
  
  pthread_mutex_unlock(play_mutex);

  /* get recall mutex */  
  recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

  /* recall context */
  pthread_mutex_lock(recall_mutex);

  list = audio->recall;
  
  while(list != NULL){
    g_object_set(G_OBJECT(list->data),
		 "output-soundcard", output_soundcard,
		 NULL);
    
    list = list->next;
  } 

  pthread_mutex_unlock(recall_mutex);

  /* unref old soundcard */
  if(old_soundcard != NULL){
    g_object_unref(old_soundcard);
  }
}

/**
 * ags_audio_set_input_soundcard:
 * @audio: the #AgsAudio
 * @input_soundcard: an #AgsSoundcard
 *
 * Set the input soundcard object on audio.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_input_soundcard(AgsAudio *audio,
			      GObject *input_soundcard)
{
  AgsChannel *channel;

  GObject *old_soundcard;
  
  GList *list;

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;
  pthread_mutex_t *play_mutex, *recall_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */  
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);
  
  /* old soundcard */
  pthread_mutex_lock(audio_mutex);

  old_soundcard = audio->input_soundcard;

  pthread_mutex_unlock(audio_mutex);
  
  if(old_soundcard == input_soundcard){
    return;
  }

  /* ref and set new soundcard */
  if(input_soundcard != NULL){
    g_object_ref(input_soundcard);
  }
  
  pthread_mutex_lock(audio_mutex);
  
  audio->input_soundcard = (GObject *) input_soundcard;

  pthread_mutex_unlock(audio_mutex);  

  /* output */
  pthread_mutex_lock(audio_mutex);
  
  channel = audio->output;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* reset */
    g_object_set(G_OBJECT(channel),
		 "input-soundcard", input_soundcard,
		 NULL);

    /* iterate */
    pthread_mutex_lock(channel_mutex);

    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }
  
  /* input */
  pthread_mutex_lock(audio_mutex);
  
  channel = audio->input;

  pthread_mutex_unlock(audio_mutex);

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* reset */
    g_object_set(G_OBJECT(channel),
		 "input-soundcard", input_soundcard,
		 NULL);

    /* iterate */
    pthread_mutex_lock(channel_mutex);

    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }

  /* get play mutex */
  play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

  /* play context */
  pthread_mutex_lock(play_mutex);
  
  list = audio->play;
  
  while(list != NULL){
    g_object_set(G_OBJECT(list->data),
		 "input-soundcard", input_soundcard,
		 NULL);
    
    list = list->next;
  }
  
  pthread_mutex_unlock(play_mutex);

  /* get recall mutex */  
  recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

  /* recall context */
  pthread_mutex_lock(recall_mutex);

  list = audio->recall;
  
  while(list != NULL){
    g_object_set(G_OBJECT(list->data),
		 "input-soundcard", input_soundcard,
		 NULL);
    
    list = list->next;
  } 

  pthread_mutex_unlock(recall_mutex);

  /* unref old soundcard */
  if(old_soundcard != NULL){
    g_object_unref(old_soundcard);
  }
}

/**
 * ags_audio_set_output_sequencer:
 * @audio: the #AgsAudio
 * @sequencer: the output sequencer
 * 
 * Set output sequencer.
 * 
 * Since: 2.0.0
 */
void
ags_audio_set_output_sequencer(AgsAudio *audio,
			       GObject *sequencer)
{
  GObject *old_sequencer;

  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */  
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);
  
  /* old sequencer */
  pthread_mutex_lock(audio_mutex);

  old_sequencer = audio->output_sequencer;

  pthread_mutex_unlock(audio_mutex);
  
  if(old_sequencer == sequencer){
    return;
  }

  /* ref and set new sequencer */
  if(sequencer != NULL){
    g_object_ref(sequencer);
  }
  
  pthread_mutex_lock(audio_mutex);
  
  audio->output_sequencer = (GObject *) sequencer;

  pthread_mutex_unlock(audio_mutex);  

  /* unref old sequencer */
  if(old_sequencer != NULL){
    g_object_unref(old_sequencer);
  }
}

/**
 * ags_audio_set_input_sequencer:
 * @audio: the #AgsAudio
 * @sequencer: the input sequencer
 * 
 * Set input sequencer.
 * 
 * Since: 2.0.0
 */
void
ags_audio_set_input_sequencer(AgsAudio *audio,
			      GObject *sequencer)
{
  GObject *old_sequencer;

  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */  
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);
  
  /* old sequencer */
  pthread_mutex_lock(audio_mutex);

  old_sequencer = audio->input_sequencer;

  pthread_mutex_unlock(audio_mutex);
  
  if(old_sequencer == sequencer){
    return;
  }

  /* ref and set new sequencer */
  if(sequencer != NULL){
    g_object_ref(sequencer);
  }
  
  pthread_mutex_lock(audio_mutex);
  
  audio->input_sequencer = (GObject *) sequencer;

  pthread_mutex_unlock(audio_mutex);  

  /* unref old sequencer */
  if(old_sequencer != NULL){
    g_object_unref(old_sequencer);
  }
}

/**
 * ags_audio_set_samplerate:
 * @audio: the #AgsAudio
 * @samplerate: the samplerate
 *
 * Sets samplerate.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_samplerate(AgsAudio *audio, guint samplerate)
{
  AgsChannel *output, *input;
  AgsPlaybackDomain *playback_domain;
  
  AgsThread *audio_thread;
  AgsMessageDelivery *message_delivery;
  AgsMessageQueue *message_queue;

  GList *start_list, *list;

  gdouble frequency;
  guint old_samplerate;
  guint i;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *playback_domain_mutex;
  
  auto void ags_audio_set_samplerate_channel(AgsChannel *channel);

  void ags_audio_set_samplerate_channel(AgsChannel *channel){
    pthread_mutex_t *channel_mutex;
    
    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* set samplerate */
      g_object_set(channel,
		   "samplerate", samplerate,
		   NULL);

      /* iterate */
      pthread_mutex_lock(channel_mutex);
      
      channel = channel->next;

      pthread_mutex_unlock(channel_mutex);
    }
  }

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* set samplerate */
  pthread_mutex_lock(audio_mutex);

  old_samplerate = audio->samplerate;
  
  audio->samplerate = samplerate;

  frequency = ceil((gdouble) audio->samplerate / (gdouble) audio->buffer_size) + AGS_SOUNDCARD_DEFAULT_OVERCLOCK;

  output = audio->output;
  input = audio->input;
  
  playback_domain = (AgsPlaybackDomain *) audio->playback_domain;

  start_list = g_list_copy(audio->synth_generator);
  
  pthread_mutex_unlock(audio_mutex);

  if(playback_domain != NULL){
    /* get playback domain mutex */
    playback_domain_mutex = AGS_PLAYBACK_DOMAIN_GET_OBJ_MUTEX(playback_domain);

    /* audio thread - frequency */
    for(i = 0; i < AGS_SOUND_SCOPE_LAST; i++){
      pthread_mutex_lock(playback_domain_mutex);
    
      audio_thread = playback_domain->audio_thread[i];

      pthread_mutex_unlock(playback_domain_mutex);
    
      if(audio_thread != NULL){
	/* apply new frequency */
	g_object_set(audio_thread,
		     "frequency", frequency,
		     NULL);
      }
    }
  }
  
  /* set samplerate output/input */
  ags_audio_set_samplerate_channel(output);
  ags_audio_set_samplerate_channel(input);

  /* set samplerate synth generator */
  list = start_list;
  
  while(list != NULL){
    g_object_set(list->data,
		 "samplerate", samplerate,
		 NULL);
    
    list = list->next;
  }
  
  g_list_free(start_list);

  /* emit message */
  message_delivery = ags_message_delivery_get_instance();

  message_queue = (AgsMessageQueue *) ags_message_delivery_find_namespace(message_delivery,
									  "libags-audio");

  if(message_queue != NULL){
    AgsMessageEnvelope *message;

    xmlDoc *doc;
    xmlNode *root_node;

    /* specify message body */
    doc = xmlNewDoc("1.0");

    root_node = xmlNewNode(NULL,
			   "ags-command");
    xmlDocSetRootElement(doc, root_node);    

    xmlNewProp(root_node,
	       "method",
	       "AgsAudio::set-samplerate");

    /* add message */
    message = ags_message_envelope_alloc((GObject *) audio,
					 NULL,
					 doc);

    /* set parameter */
    message->n_params = 2;

    message->parameter_name = (gchar **) malloc(3 * sizeof(gchar *));
    message->value = g_new0(GValue,
			    2);

    /* samplerate */
    message->parameter_name[0] = "samplerate";
    
    g_value_init(&(message->value[0]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[0]),
		     samplerate);

    /* old samplerate */
    message->parameter_name[1] = "old-samplerate";
    
    g_value_init(&(message->value[1]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[1]),
		     old_samplerate);

    /* terminate string vector */
    message->parameter_name[2] = NULL;
    
    /* add message */
    ags_message_delivery_add_message(message_delivery,
				     "libags-audio",
				     message);
  }
}

/**
 * ags_audio_set_buffer_size:
 * @audio: the #AgsAudio
 * @buffer_size: the buffer length
 *
 * Sets buffer length.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_buffer_size(AgsAudio *audio, guint buffer_size)
{
  AgsChannel *output, *input;
  AgsPlaybackDomain *playback_domain;

  AgsThread *audio_thread;
  AgsMessageDelivery *message_delivery;
  AgsMessageQueue *message_queue;

  GList *start_list, *list;
  
  gdouble frequency;
  guint old_buffer_size;
  guint i;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *playback_domain_mutex;

  auto void ags_audio_set_buffer_size_channel(AgsChannel *channel);

  void ags_audio_set_buffer_size_channel(AgsChannel *channel){
    pthread_mutex_t *channel_mutex;

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* set buffer size */
      g_object_set(channel,
		   "buffer-size", buffer_size,
		   NULL);
      
      /* iterate */
      pthread_mutex_lock(channel_mutex);
      
      channel = channel->next;

      pthread_mutex_unlock(channel_mutex);
    }
  }

  if(!AGS_IS_AUDIO(audio)){
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* set buffer size */
  pthread_mutex_lock(audio_mutex);

  old_buffer_size = audio->buffer_size;
  
  audio->buffer_size = buffer_size;

  output = audio->output;
  input = audio->input;

  playback_domain = (AgsPlaybackDomain *) audio->playback_domain;

  start_list = g_list_copy(audio->synth_generator);

  frequency = ceil((gdouble) audio->samplerate / (gdouble) audio->buffer_size) + AGS_SOUNDCARD_DEFAULT_OVERCLOCK;
  
  pthread_mutex_unlock(audio_mutex);

  /* get playback domain mutex */
  playback_domain_mutex = AGS_PLAYBACK_DOMAIN_GET_OBJ_MUTEX(playback_domain);

  /* audio thread - frequency */
  for(i = 0; i < AGS_SOUND_SCOPE_LAST; i++){
    pthread_mutex_lock(playback_domain_mutex);
    
    audio_thread = playback_domain->audio_thread[i];

    pthread_mutex_unlock(playback_domain_mutex);

    if(audio_thread != NULL){
      /* apply new frequency */
      g_object_set(playback_domain->audio_thread[i],
		   "frequency", frequency,
		   NULL);
    }
  }  

  /* set buffer size output/input */
  ags_audio_set_buffer_size_channel(output);
  ags_audio_set_buffer_size_channel(input);

  /* set buffer size synth generator */
  list = start_list;
  
  while(list != NULL){
    g_object_set(list->data,
		 "buffer-size", buffer_size,
		 NULL);
    
    list = list->next;
  }
  
  g_list_free(start_list);

  /* emit message */
  message_delivery = ags_message_delivery_get_instance();

  message_queue = (AgsMessageQueue *) ags_message_delivery_find_namespace(message_delivery,
									  "libags-audio");

  if(message_queue != NULL){
    AgsMessageEnvelope *message;

    xmlDoc *doc;
    xmlNode *root_node;

    /* specify message body */
    doc = xmlNewDoc("1.0");

    root_node = xmlNewNode(NULL,
			   "ags-command");
    xmlDocSetRootElement(doc, root_node);    

    xmlNewProp(root_node,
	       "method",
	       "AgsAudio::set-buffer-size");

    /* add message */
    message = ags_message_envelope_alloc((GObject *) audio,
					 NULL,
					 doc);

    /* set parameter */
    message->n_params = 2;

    message->parameter_name = (gchar **) malloc(3 * sizeof(gchar *));
    message->value = g_new0(GValue,
			    2);

    /* buffer_size */
    message->parameter_name[0] = "buffer-size";
    
    g_value_init(&(message->value[0]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[0]),
		     buffer_size);

    /* old buffer_size */
    message->parameter_name[1] = "old-buffer-size";
    
    g_value_init(&(message->value[1]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[1]),
		     old_buffer_size);

    /* terminate string vector */
    message->parameter_name[2] = NULL;
    
    /* add message */
    ags_message_delivery_add_message(message_delivery,
				     "libags-audio",
				     message);
  }
}

/**
 * ags_audio_set_format:
 * @audio: the #AgsAudio
 * @format: the format
 *
 * Sets buffer length.
 *
 * Since: 2.0.0
 */
void
ags_audio_set_format(AgsAudio *audio, guint format)
{
  AgsChannel *output, *input;
  
  AgsMessageDelivery *message_delivery;
  AgsMessageQueue *message_queue;

  GList *start_list, *list;

  guint old_format;
  
  pthread_mutex_t *audio_mutex;

  auto void ags_audio_set_format_channel(AgsChannel *channel);

  void ags_audio_set_format_channel(AgsChannel *channel){
    pthread_mutex_t *channel_mutex;

    while(channel != NULL){
      /* get channel mutex */
      channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

      /* set format */
      g_object_set(channel,
		   "format", format,
		   NULL);
      
      /* iterate */
      pthread_mutex_lock(channel_mutex);
      
      channel = channel->next;

      pthread_mutex_unlock(channel_mutex);
    }
  }  

  if(!AGS_IS_AUDIO(audio)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* set format */
  pthread_mutex_lock(audio_mutex);

  old_format = audio->format;
  
  audio->format = format;

  output = audio->output;
  input = audio->input;

  start_list = g_list_copy(audio->synth_generator);

  pthread_mutex_unlock(audio_mutex);

  /* set format output/input */
  ags_audio_set_format_channel(output);
  ags_audio_set_format_channel(input);

  /* set format synth generator */
  list = start_list;
  
  while(list != NULL){
    g_object_set(list->data,
		 "format", format,
		 NULL);
    
    list = list->next;
  }
  
  g_list_free(start_list);

  /* emit message */
  message_delivery = ags_message_delivery_get_instance();

  message_queue = (AgsMessageQueue *) ags_message_delivery_find_namespace(message_delivery,
									  "libags-audio");

  if(message_queue != NULL){
    AgsMessageEnvelope *message;

    xmlDoc *doc;
    xmlNode *root_node;

    /* specify message body */
    doc = xmlNewDoc("1.0");

    root_node = xmlNewNode(NULL,
			   "ags-command");
    xmlDocSetRootElement(doc, root_node);    

    xmlNewProp(root_node,
	       "method",
	       "AgsAudio::set-format");

    /* add message */
    message = ags_message_envelope_alloc((GObject *) audio,
					 NULL,
					 doc);

    /* set parameter */
    message->n_params = 2;

    message->parameter_name = (gchar **) malloc(3 * sizeof(gchar *));
    message->value = g_new0(GValue,
			    2);

    /* format */
    message->parameter_name[0] = "format";
    
    g_value_init(&(message->value[0]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[0]),
		     format);

    /* old format */
    message->parameter_name[1] = "old-format";
    
    g_value_init(&(message->value[1]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[1]),
		     old_format);

    /* terminate string vector */
    message->parameter_name[2] = NULL;
    
    /* add message */
    ags_message_delivery_add_message(message_delivery,
				     "libags-audio",
				     message);
  }
}

/**
 * ags_audio_add_preset:
 * @audio: the #AgsAudio
 * @preset: an #AgsPreset
 *
 * Adds a preset.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_preset(AgsAudio *audio,
		     GObject *preset)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_PRESET(preset)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add preset */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->preset,
		 preset) == NULL){
    success = TRUE;
    
    g_object_ref(preset);
    audio->preset = g_list_prepend(audio->preset,
				   preset);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(preset,
		 "audio", audio,
		 NULL);
  }
}

/**
 * ags_audio_remove_preset:
 * @audio: the #AgsAudio
 * @preset: an #AgsPreset
 *
 * Removes a preset.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_preset(AgsAudio *audio,
			GObject *preset)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_PRESET(preset)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove preset */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->preset,
		 preset) != NULL){
    success = TRUE;
    
    audio->preset = g_list_remove(audio->preset,
				  preset);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(preset,
		 "audio", NULL,
		 NULL);

    g_object_unref(preset);
  }
}

/**
 * ags_audio_add_synth_generator:
 * @audio: the #AgsAudio
 * @synth_generator: an #AgsSynthGenerator
 *
 * Adds a synth generator.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_synth_generator(AgsAudio *audio,
			      GObject *synth_generator)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_SYNTH_GENERATOR(synth_generator)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add synth_generator */
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->synth_generator,
		 synth_generator) == NULL){
    g_object_ref(synth_generator);
    audio->synth_generator = g_list_prepend(audio->synth_generator,
					    synth_generator);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_remove_synth_generator:
 * @audio: the #AgsAudio
 * @synth_generator: an #AgsSynthGenerator
 *
 * Removes a synth generator.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_synth_generator(AgsAudio *audio,
				 GObject *synth_generator)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_SYNTH_GENERATOR(synth_generator)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove synth_generator */
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->synth_generator,
		 synth_generator) != NULL){
    audio->synth_generator = g_list_remove(audio->synth_generator,
					   synth_generator);
    g_object_unref(synth_generator);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_add_cursor:
 * @audio: the #AgsAudio
 * @cursor: the #GObject implementing #AgsCursor
 *
 * Adds a cursor.
 *
 * Since: 2.2.0
 */
void
ags_audio_add_cursor(AgsAudio *audio, GObject *cursor)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_CURSOR(cursor)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add cursor */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->cursor,
		 cursor) == NULL){
    success = TRUE;
    
    g_object_ref(cursor);
    audio->cursor = g_list_prepend(audio->cursor,
				   (GObject *) cursor);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(cursor,
		 "audio", audio,
		 NULL);
  }
}

/**
 * ags_audio_remove_cursor:
 * @audio: the #AgsAudio
 * @cursor: the #GObject implementing #AgsCursor
 *
 * Removes a cursor.
 *
 * Since: 2.2.0
 */
void
ags_audio_remove_cursor(AgsAudio *audio, GObject *cursor)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_CURSOR(cursor)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove cursor */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->cursor,
		 cursor) != NULL){
    success = TRUE;
    
    audio->cursor = g_list_remove(audio->cursor,
				  cursor);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(cursor,
		 "audio", NULL,
		 NULL);
    
    g_object_unref(cursor);
  }
}

/**
 * ags_audio_add_notation:
 * @audio: the #AgsAudio
 * @notation: the #AgsNotation
 *
 * Adds a notation.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_notation(AgsAudio *audio, GObject *notation)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_NOTATION(notation)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add notation */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->notation,
		 notation) == NULL){
    success = TRUE;
    
    g_object_ref(notation);
    audio->notation = ags_notation_add(audio->notation,
				       (AgsNotation *) notation);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(notation,
		 "audio", audio,
		 NULL);
  }
}

/**
 * ags_audio_remove_notation:
 * @audio: the #AgsAudio
 * @notation: the #AgsNotation
 *
 * Removes a notation.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_notation(AgsAudio *audio, GObject *notation)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_NOTATION(notation)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove notation */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->notation,
		 notation) != NULL){
    success = TRUE;
    
    audio->notation = g_list_remove(audio->notation,
				    notation);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(notation,
		 "audio", NULL,
		 NULL);
    
    g_object_unref(notation);
  }
}

/**
 * ags_audio_add_automation:
 * @audio: the #AgsAudio
 * @automation: the #AgsAutomation
 *
 * Adds an automation.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_automation(AgsAudio *audio, GObject *automation)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_AUTOMATION(automation)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add automation */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->automation,
		 automation) == NULL){
    success = TRUE;
    
    g_object_ref(automation);
    audio->automation = ags_automation_add(audio->automation,
					   (AgsAutomation *) automation);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(automation,
		 "audio", audio,
		 NULL);
  }
}

/**
 * ags_audio_remove_automation:
 * @audio: the #AgsAudio
 * @automation: the #AgsAutomation
 *
 * Removes an automation.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_automation(AgsAudio *audio, GObject *automation)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_AUTOMATION(automation)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove automation */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->automation,
		 automation) != NULL){
    success = TRUE;
    
    audio->automation = g_list_remove(audio->automation,
				      automation);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(automation,
		 "audio", NULL,
		 NULL);

    g_object_unref(automation);
  }
}

/**
 * ags_audio_add_wave:
 * @audio: the #AgsAudio
 * @wave: the #AgsWave
 *
 * Adds a wave.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_wave(AgsAudio *audio, GObject *wave)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_WAVE(wave)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add wave */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->wave,
		 wave) == NULL){
    success = TRUE;
    
    g_object_ref(wave);
    audio->wave = ags_wave_add(audio->wave,
			       (AgsWave *) wave);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(wave,
		 "audio", audio,
		 NULL);
  }
}

/**
 * ags_audio_remove_wave:
 * @audio: the #AgsAudio
 * @wave: the #AgsWave
 *
 * Removes a wave.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_wave(AgsAudio *audio, GObject *wave)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_WAVE(wave)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove wave */
  success = FALSE;
  
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->wave,
		 wave) != NULL){
    success = TRUE;
    
    audio->wave = g_list_remove(audio->wave,
				wave);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(wave,
		 "audio", NULL,
		 NULL);

    g_object_unref(wave);
  }
}

/**
 * ags_audio_add_midi:
 * @audio: the #AgsAudio
 * @midi: the #AgsMidi
 *
 * Adds a midi.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_midi(AgsAudio *audio, GObject *midi)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_MIDI(midi)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add midi */
  success = FALSE;

  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->midi,
		 midi) == NULL){
    success = TRUE;
    
    g_object_ref(midi);
    audio->midi = ags_midi_add(audio->midi,
			       (AgsMidi *) midi);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(midi,
		 "audio", audio,
		 NULL);
  }
}

/**
 * ags_audio_remove_midi:
 * @audio: the #AgsAudio
 * @midi: the #AgsMidi
 *
 * Removes a midi.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_midi(AgsAudio *audio, GObject *midi)
{
  gboolean success;
  
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_MIDI(midi)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove midi */
  success = FALSE;

  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->midi,
		 midi) != NULL){
    success = TRUE;
    
    audio->midi = g_list_remove(audio->midi,
				midi);
  }
  
  pthread_mutex_unlock(audio_mutex);

  if(success){
    g_object_set(midi,
		 "audio", NULL,
		 NULL);

    g_object_unref(midi);
  }
}

/**
 * ags_audio_add_recall_id:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID
 *
 * Adds a recall id.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_recall_id(AgsAudio *audio, GObject *recall_id)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECALL_ID(recall_id)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add recall id */
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->recall_id,
		 recall_id) == NULL){
    g_object_ref(recall_id);
    audio->recall_id = g_list_prepend(audio->recall_id,
				      recall_id);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_remove_recall_id:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID
 *
 * Removes a recall id.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_recall_id(AgsAudio *audio, GObject *recall_id)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECALL_ID(recall_id)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove recall id */
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->recall_id,
		 recall_id) != NULL){
    audio->recall_id = g_list_remove(audio->recall_id,
				     recall_id);
    g_object_unref(recall_id);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_add_recycling_context:
 * @audio: the #AgsAudio
 * @recycling_context: the #AgsRecyclingContext
 *
 * Adds a recycling context.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_recycling_context(AgsAudio *audio, GObject *recycling_context)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECYCLING_CONTEXT(recycling_context)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add recycling context */
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->recycling_context,
		 recycling_context) == NULL){
    g_object_ref(recycling_context);
    audio->recycling_context = g_list_prepend(audio->recycling_context,
					      recycling_context);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_remove_recycling_context:
 * @audio: the #AgsAudio
 * @recycling_context: the #AgsRecyclingContext
 *
 * Removes a recycling context.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_recycling_context(AgsAudio *audio, GObject *recycling_context)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECYCLING_CONTEXT(recycling_context)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove recycling container */
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->recycling_context,
		 recycling_context) != NULL){
    audio->recycling_context = g_list_remove(audio->recycling_context,
					     recycling_context);
    g_object_unref(recycling_context);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_add_recall_container:
 * @audio: the #AgsAudio
 * @recall_container: the #AgsRecallContainer
 *
 * Adds a recall container.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_recall_container(AgsAudio *audio, GObject *recall_container)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECALL_CONTAINER(recall_container)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* add recall container */
  pthread_mutex_lock(audio_mutex);
  
  if(g_list_find(audio->recall_container,
		 recall_container) == NULL){
    g_object_ref(recall_container);
    audio->recall_container = g_list_prepend(audio->recall_container,
					     recall_container);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_remove_recall_container:
 * @audio: the #AgsAudio
 * @recall_container: the #AgsRecallContainer
 *
 * Removes a recall container.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_recall_container(AgsAudio *audio, GObject *recall_container)
{
  pthread_mutex_t *audio_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECALL_CONTAINER(recall_container)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* remove recall container */
  pthread_mutex_lock(audio_mutex);

  if(g_list_find(audio->recall_container,
		 recall_container) != NULL){
    audio->recall_container = g_list_remove(audio->recall_container,
					    recall_container);
    g_object_unref(recall_container);
  }
  
  pthread_mutex_unlock(audio_mutex);
}

/**
 * ags_audio_add_recall:
 * @audio: the #AgsAudio
 * @recall: the #AgsRecall
 * @play_context: if %TRUE play context, else if %FALSE recall context
 *
 * Adds a recall to @audio.
 *
 * Since: 2.0.0
 */
void
ags_audio_add_recall(AgsAudio *audio,
		     GObject *recall, gboolean play_context)
{
  gboolean success;
  
  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECALL(recall)){
    return;
  }

  success = FALSE;
  
  if(play_context){
    pthread_mutex_t *play_mutex;

    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* add recall */
    pthread_mutex_lock(play_mutex);
    
    if(g_list_find(audio->play, recall) == NULL){
      success = TRUE;

      g_object_ref(G_OBJECT(recall));
    
      audio->play = g_list_prepend(audio->play,
				   recall);
    }

    pthread_mutex_unlock(play_mutex);
  }else{
    pthread_mutex_t *recall_mutex;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* add recall */
    pthread_mutex_lock(recall_mutex);

    if(g_list_find(audio->recall, recall) == NULL){
      success = TRUE;
      
      g_object_ref(G_OBJECT(recall));
    
      audio->recall = g_list_prepend(audio->recall,
				     recall);
    }

    pthread_mutex_unlock(recall_mutex);
  }

  if(success){
    if(AGS_IS_RECALL_AUDIO(recall) ||
       AGS_IS_RECALL_AUDIO_RUN(recall)){
      g_object_set(recall,
		   "audio", audio,
		   NULL);
    }
  }
}

/**
 * ags_audio_remove_recall:
 * @audio: the #AgsAudio
 * @recall: the #AgsRecall
 * @play_context: if %TRUE play context, else if %FALSE recall context
 *
 * Removes a recall from @audio.
 *
 * Since: 2.0.0
 */
void
ags_audio_remove_recall(AgsAudio *audio, GObject *recall, gboolean play_context)
{
  gboolean success;
  
  if(!AGS_IS_AUDIO(audio) ||
     !AGS_IS_RECALL(recall)){
    return;
  }

  success = FALSE;
  
  if(play_context){
    pthread_mutex_t *play_mutex;

    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* add recall */
    pthread_mutex_lock(play_mutex);
    
    if(g_list_find(audio->play, recall) != NULL){
      audio->play = g_list_remove(audio->play,
				  recall);
    }

    pthread_mutex_unlock(play_mutex);
  }else{
    pthread_mutex_t *recall_mutex;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* add recall */
    pthread_mutex_lock(recall_mutex);

    if(g_list_find(audio->recall, recall) != NULL){
      audio->recall = g_list_remove(audio->recall,
				    recall);
    }

    pthread_mutex_unlock(recall_mutex);
  }

#if 0
  if(success){
    if(AGS_IS_RECALL_AUDIO(recall) ||
       AGS_IS_RECALL_AUDIO_RUN(recall)){
      g_object_set(recall,
		   "audio", NULL,
		   NULL);
    }

    g_object_unref(G_OBJECT(recall));
  }
#endif
}

void
ags_audio_real_duplicate_recall(AgsAudio *audio,
				AgsRecallID *recall_id,
				guint pad, guint audio_channel,
				guint line)
{
  AgsRecall *recall, *copy_recall;
  AgsRecyclingContext *parent_recycling_context, *recycling_context;

  GList *list_start, *list;

  guint sound_scope;
  guint current_staging_flags;
  gboolean play_context;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *recall_id_mutex;
  pthread_mutex_t *recycling_context_mutex;

  if(!AGS_IS_RECALL_ID(recall_id)){
    return;
  }

  /* get recall id mutex */
  recall_id_mutex = AGS_RECALL_ID_GET_OBJ_MUTEX(recall_id);

  /* get some fields */
  pthread_mutex_lock(recall_id_mutex);

  recycling_context = recall_id->recycling_context;

  sound_scope = recall_id->sound_scope;

  pthread_mutex_unlock(recall_id_mutex);

  if(sound_scope == -1){
    g_critical("can only duplicate for specific sound scope");
    
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get staging flags */
  pthread_mutex_lock(audio_mutex);

  current_staging_flags = audio->staging_flags[sound_scope];
  
  pthread_mutex_unlock(audio_mutex);

  if(!AGS_IS_RECYCLING_CONTEXT(recycling_context) ||
     (AGS_SOUND_STAGING_RUN_INIT_PRE & (current_staging_flags)) != 0){
    return;
  }

  /* get recycling context mutex */
  recycling_context_mutex = AGS_RECYCLING_CONTEXT_GET_OBJ_MUTEX(recycling_context);

  /* get parent recycling context */
  pthread_mutex_lock(recycling_context_mutex);

  parent_recycling_context = recycling_context->parent;

  pthread_mutex_unlock(recycling_context_mutex);

  /* get the appropriate list */
  if(parent_recycling_context == NULL){
    pthread_mutex_t *play_mutex;

    play_context = TRUE;
    
    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* copy play context */
    pthread_mutex_lock(play_mutex);

    list_start = g_list_copy(audio->play);

    pthread_mutex_unlock(play_mutex);

    /* reverse play context */
    list =
      list_start = g_list_reverse(list_start);
  }else{
    pthread_mutex_t *recall_mutex;

    play_context = FALSE;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* copy recall context */
    pthread_mutex_lock(recall_mutex);

    list_start = g_list_copy(audio->recall);
    
    pthread_mutex_unlock(recall_mutex);

    /* reverse recall context */
    list =
      list_start = g_list_reverse(list_start);
  }

  /* notify run */  
  //  ags_recall_notify_dependency(AGS_RECALL(list->data), AGS_RECALL_NOTIFY_RUN, 1);
  
  /* return if already played */
  pthread_mutex_lock(recall_id_mutex);

  if(ags_recall_id_check_state_flags(recall_id, AGS_SOUND_STATE_IS_WAITING) ||
     ags_recall_id_check_state_flags(recall_id, AGS_SOUND_STATE_IS_ACTIVE) ||
     ags_recall_id_check_state_flags(recall_id, AGS_SOUND_STATE_IS_PROCESSING) ||
     ags_recall_id_check_state_flags(recall_id, AGS_SOUND_STATE_IS_TERMINATING)){
    pthread_mutex_unlock(recall_id_mutex);
    
    return;
  }

  ags_recall_id_set_state_flags(recall_id,
				AGS_SOUND_STATE_IS_WAITING);

  pthread_mutex_unlock(recall_id_mutex);
  
  /* duplicate */
  while(list != NULL){
    pthread_mutex_t *current_recall_mutex;
    
    recall = AGS_RECALL(list->data);

    /* get current recall mutex */
    current_recall_mutex = AGS_RECALL_GET_OBJ_MUTEX(recall);

    /* some checks */
    pthread_mutex_lock(current_recall_mutex);
    
    if(AGS_IS_RECALL_AUDIO(recall) ||
       (AGS_RECALL_TEMPLATE & (recall->flags)) == 0 ||
       recall->recall_id != NULL ||
       !ags_recall_match_ability_flags_to_scope(recall,
						sound_scope)){
      list = list->next;

      pthread_mutex_unlock(current_recall_mutex);
      
      continue;
    }  

    pthread_mutex_unlock(current_recall_mutex);

    /* duplicate the recall */
    copy_recall = ags_recall_duplicate(recall, recall_id,
				       NULL, NULL, NULL);
      
    if(copy_recall == NULL){
      /* iterate */    
      list = list->next;

      continue;
    }

    g_object_set(copy_recall,
		 "pad", pad,
		 "audio-channel", audio_channel,
		 "line", line,
		 NULL);
    
#ifdef AGS_DEBUG
    g_message("recall duplicated: %s %s", G_OBJECT_TYPE_NAME(audio), G_OBJECT_TYPE_NAME(copy_recall));
#endif

    /* set appropriate sound scope */
    copy_recall->sound_scope = sound_scope;
      
    /* append to AgsAudio */
    ags_audio_add_recall(audio,
			 (GObject *) copy_recall,
			 play_context);
    g_signal_connect(copy_recall, "done",
		     G_CALLBACK(ags_audio_recall_done_callback), audio);
    
    /* connect */
    ags_connectable_connect(AGS_CONNECTABLE(copy_recall));

    /* notify run */
    ags_recall_notify_dependency(copy_recall, AGS_RECALL_NOTIFY_RUN, 1);

    /* iterate */
    list = list->next;
  }

  g_list_free(list_start);
}


/**
 * ags_audio_duplicate_recall:
 * @audio: the #AgsAudio
 * @recall_id: an #AgsRecallID
 * @pad: the pad
 * @audio_channel: the audio channel
 * @line: the line
 * 
 * Duplicate all #AgsRecall templates of @audio.
 *
 * Since: 2.0.0
 */
void
ags_audio_duplicate_recall(AgsAudio *audio,
			   AgsRecallID *recall_id,
			   guint pad, guint audio_channel,
			   guint line)
{  
  g_return_if_fail(AGS_IS_AUDIO(audio));

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[DUPLICATE_RECALL], 0,
		recall_id,
		pad, audio_channel,
		line);
  g_object_unref((GObject *) audio);
}

void
ags_audio_real_resolve_recall(AgsAudio *audio,
			      AgsRecallID *recall_id)
{
  AgsRecall *recall;
  AgsRecyclingContext *parent_recycling_context, *recycling_context;

  GList *list_start, *list;

  guint sound_scope;
  guint current_staging_flags;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *recall_id_mutex;
  pthread_mutex_t *recycling_context_mutex;

  if(!AGS_IS_RECALL_ID(recall_id)){
    return;
  }

  /* get recall id mutex */
  recall_id_mutex = AGS_RECALL_ID_GET_OBJ_MUTEX(recall_id);

  /* get some fields */
  pthread_mutex_lock(recall_id_mutex);

  recycling_context = recall_id->recycling_context;

  sound_scope = recall_id->sound_scope;

  pthread_mutex_unlock(recall_id_mutex);

  if(sound_scope == -1){
    g_critical("can only resolve for specific sound scope");
    
    return;
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get staging flags */
  pthread_mutex_lock(audio_mutex);

  current_staging_flags = audio->staging_flags[sound_scope];
  
  pthread_mutex_unlock(audio_mutex);

  if(!AGS_IS_RECYCLING_CONTEXT(recycling_context) ||
     (AGS_SOUND_STAGING_RUN_INIT_PRE & (current_staging_flags)) != 0){
    return;
  }

  /* get recycling context mutex */
  recycling_context_mutex = AGS_RECYCLING_CONTEXT_GET_OBJ_MUTEX(recycling_context);

  /* get parent recycling context */
  pthread_mutex_lock(recycling_context_mutex);

  parent_recycling_context = recycling_context->parent;

  pthread_mutex_unlock(recycling_context_mutex);
  
  /* get the appropriate lists */
  if(parent_recycling_context == NULL){
    pthread_mutex_t *play_mutex;

    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* copy play context */
    pthread_mutex_lock(play_mutex);

    list = 
      list_start = g_list_copy(audio->play);

    pthread_mutex_unlock(play_mutex);

    /* reverse play context */
    list =
      list_start = g_list_reverse(list_start);
  }else{
    pthread_mutex_t *recall_mutex;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* copy recall context */
    pthread_mutex_lock(recall_mutex);

    list = 
      list_start = g_list_copy(audio->recall);
    
    pthread_mutex_unlock(recall_mutex);

    /* reverse recall context */
    list =
      list_start = g_list_reverse(list_start);
  }

  /* resolve */
  while((list = ags_recall_find_recycling_context(list,
						  (GObject *) recycling_context)) != NULL){
    recall = AGS_RECALL(list->data);
    
    ags_recall_resolve_dependency(recall);

    list = list->next;
  }

  g_list_free(list_start);
}

/**
 * ags_audio_resolve_recall:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID to use
 *
 * Performs resolving of recalls.
 *
 * Since: 2.0.0
 */
void
ags_audio_resolve_recall(AgsAudio *audio,
			 AgsRecallID *recall_id)
{  
  g_return_if_fail(AGS_IS_AUDIO(audio));

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[RESOLVE_RECALL], 0,
		recall_id);
  g_object_unref((GObject *) audio);
}

void
ags_audio_real_init_recall(AgsAudio *audio,
			   AgsRecallID *recall_id, guint staging_flags)
{
  AgsRecall *recall;
  AgsRecyclingContext *parent_recycling_context, *recycling_context;
  
  GList *list_start, *list;

  guint sound_scope;
  guint current_staging_flags;
  static const guint staging_mask = (AGS_SOUND_STAGING_CHECK_RT_DATA |
				     AGS_SOUND_STAGING_RUN_INIT_PRE |
				     AGS_SOUND_STAGING_RUN_INIT_INTER |
				     AGS_SOUND_STAGING_RUN_INIT_POST);

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *recall_id_mutex;
  pthread_mutex_t *recycling_context_mutex;

  if(!AGS_IS_RECALL_ID(recall_id)){
    return;
  }
  
  /* get recall id mutex */
  recall_id_mutex = AGS_RECALL_ID_GET_OBJ_MUTEX(recall_id);

  /* get some fields */
  pthread_mutex_lock(recall_id_mutex);

  recycling_context = recall_id->recycling_context;

  sound_scope = recall_id->sound_scope;

  pthread_mutex_unlock(recall_id_mutex);

  if(sound_scope == -1){
    g_critical("can only init for specific sound scope");
    
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get staging flags */
  pthread_mutex_lock(audio_mutex);

  current_staging_flags = audio->staging_flags[sound_scope];
  
  pthread_mutex_unlock(audio_mutex);

  if(!AGS_IS_RECYCLING_CONTEXT(recycling_context) ||
     (AGS_SOUND_STAGING_RUN_INIT_PRE & (current_staging_flags)) != 0){
    return;
  }

  /* get recycling context mutex */
  recycling_context_mutex = AGS_RECYCLING_CONTEXT_GET_OBJ_MUTEX(recycling_context);

  /* get parent recycling context */
  pthread_mutex_lock(recycling_context_mutex);

  parent_recycling_context = recycling_context->parent;

  pthread_mutex_unlock(recycling_context_mutex);

  /* get the appropriate lists */
  if(parent_recycling_context == NULL){
    pthread_mutex_t *play_mutex;

    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* copy play context */
    pthread_mutex_lock(play_mutex);

    list = 
      list_start = g_list_copy(audio->play);

    pthread_mutex_unlock(play_mutex);

    /* reverse play context */
    list =
      list_start = g_list_reverse(list_start);
  }else{
    pthread_mutex_t *recall_mutex;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* copy recall context */
    pthread_mutex_lock(recall_mutex);

    list = 
      list_start = g_list_copy(audio->recall);
    
    pthread_mutex_unlock(recall_mutex);

    /* reverse recall context */
    list =
      list_start = g_list_reverse(list_start);
  }
  
  /* init  */
  staging_flags = staging_mask & staging_flags;
  
  while((list = ags_recall_find_recycling_context(list,
						  (GObject *) recycling_context)) != NULL){
    recall = AGS_RECALL(list->data);
    
    /* run init stages */
    ags_recall_set_staging_flags(recall,
				 staging_flags);

    list = list->next;
  }
  
  g_list_free(list_start);

#if 0
  ags_audio_set_staging_flags(audio, sound_scope,
			      staging_flags);
#endif
}

/**
 * ags_audio_init_recall:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID to use or #NULL
 * @staging_flags: staging flags, see #AgsSoundStagingFlags-enum
 *
 * Initializes the recalls of @audio
 *
 * Since: 2.0.0
 */
void
ags_audio_init_recall(AgsAudio *audio,
		      AgsRecallID *recall_id, guint staging_flags)
{
  g_return_if_fail(AGS_IS_AUDIO(audio));

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[INIT_RECALL], 0,
		recall_id, staging_flags);
  g_object_unref((GObject *) audio);
}

void
ags_audio_real_play_recall(AgsAudio *audio,
			   AgsRecallID *recall_id, guint staging_flags)
{
  AgsRecall *recall;
  AgsRecyclingContext *parent_recycling_context, *recycling_context;
  
  GList *list_start, *list;

  guint sound_scope;
  guint current_staging_flags;
  static const guint staging_mask = (AGS_SOUND_STAGING_RUN_INIT_PRE |
				     AGS_SOUND_STAGING_RUN_INIT_INTER |
				     AGS_SOUND_STAGING_RUN_INIT_POST |
				     AGS_SOUND_STAGING_FEED_INPUT_QUEUE |
				     AGS_SOUND_STAGING_AUTOMATE |
				     AGS_SOUND_STAGING_RUN_PRE |
				     AGS_SOUND_STAGING_RUN_INTER |
				     AGS_SOUND_STAGING_RUN_POST |
				     AGS_SOUND_STAGING_DO_FEEDBACK |
				     AGS_SOUND_STAGING_FEED_OUTPUT_QUEUE |
				     AGS_SOUND_STAGING_FINI);

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *recall_id_mutex;
  pthread_mutex_t *recycling_context_mutex;
  
  if(!AGS_IS_RECALL_ID(recall_id)){
    return;
  }
  
  /* get recall id mutex */
  recall_id_mutex = AGS_RECALL_ID_GET_OBJ_MUTEX(recall_id);

  /* get some fields */
  pthread_mutex_lock(recall_id_mutex);

  recycling_context = recall_id->recycling_context;

  sound_scope = recall_id->sound_scope;

  pthread_mutex_unlock(recall_id_mutex);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get staging flags */
  pthread_mutex_lock(audio_mutex);

  current_staging_flags = audio->staging_flags[sound_scope];
  
  pthread_mutex_unlock(audio_mutex);

#if 0
  if(!AGS_IS_RECYCLING_CONTEXT(recycling_context) ||
     (AGS_SOUND_STAGING_RUN_INIT_PRE & (current_staging_flags)) == 0 ||
     (AGS_SOUND_STAGING_RUN_INIT_INTER & (current_staging_flags)) == 0 ||
     (AGS_SOUND_STAGING_RUN_INIT_POST & (current_staging_flags)) == 0){
    return;
  }
#endif
  
  /* get recycling context mutex */
  recycling_context_mutex = AGS_RECYCLING_CONTEXT_GET_OBJ_MUTEX(recycling_context);

  /* get parent recycling context */
  pthread_mutex_lock(recycling_context_mutex);

  parent_recycling_context = recycling_context->parent;

  pthread_mutex_unlock(recycling_context_mutex);

  /* get the appropriate lists */
  if(parent_recycling_context == NULL){
    pthread_mutex_t *play_mutex;

    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* copy play context */
    pthread_mutex_lock(play_mutex);

    list_start = g_list_copy_deep(audio->play,
				  (GCopyFunc) g_object_ref,
				  NULL);

    pthread_mutex_unlock(play_mutex);

    /* reverse play context */
    list =
      list_start = g_list_reverse(list_start);
  }else{
    pthread_mutex_t *recall_mutex;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* copy recall context */
    pthread_mutex_lock(recall_mutex);

    list_start = g_list_copy_deep(audio->recall,
				  (GCopyFunc) g_object_ref,
				  NULL);
    
    pthread_mutex_unlock(recall_mutex);

    /* reverse recall context */
    list =
      list_start = g_list_reverse(list_start);
  }

  /* automate and play  */
  staging_flags = staging_mask & staging_flags;
  
  if((AGS_SOUND_STAGING_AUTOMATE & (staging_flags)) != 0){
    while(list != NULL){
      recall = AGS_RECALL(list->data);
      
      /* play stages */
      if(AGS_IS_RECALL_AUDIO(recall)){
	ags_recall_set_staging_flags(recall,
				     AGS_SOUND_STAGING_AUTOMATE);
	ags_recall_unset_staging_flags(recall,
				       AGS_SOUND_STAGING_AUTOMATE);
      }
      
      list = list->next;
    }
  }

  staging_flags &= (~AGS_SOUND_STAGING_AUTOMATE);
  list = list_start;

  while((list = ags_recall_find_recycling_context(list,
						  (GObject *) recycling_context)) != NULL){
    recall = AGS_RECALL(list->data);
    
    /* play stages */
    ags_recall_set_staging_flags(recall,
				 staging_flags);
    ags_recall_unset_staging_flags(recall,
				   staging_flags);

    list = list->next;
  }
  
  g_list_free_full(list_start,
		   g_object_unref);

  //FIXME:JK: uncomment
  //  ags_audio_set_staging_flags(audio, sound_scope,
  //			      staging_flags);
}

/**
 * ags_audio_play_recall:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID to apply to
 * @staging_flags: staging flags, see #AgsSoundStagingFlags-enum
 *
 * Performs play for the specified @staging_flags.
 *
 * Since: 2.0.0
 */
void
ags_audio_play_recall(AgsAudio *audio,
		      AgsRecallID *recall_id, guint staging_flags)
{
  g_return_if_fail(AGS_IS_AUDIO(audio) && AGS_IS_RECALL_ID(recall_id));

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[PLAY_RECALL], 0,
		recall_id, staging_flags);
  g_object_unref((GObject *) audio);
}

void
ags_audio_real_done_recall(AgsAudio *audio,
			   AgsRecallID *recall_id)
{
  AgsRecall *recall;
  AgsRecyclingContext *parent_recycling_context, *recycling_context;
  
  GList *list_start, *list;

  guint sound_scope;
  guint current_staging_flags;
  static const guint staging_flags = (AGS_SOUND_STAGING_DONE);  

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *recall_id_mutex;
  pthread_mutex_t *recycling_context_mutex;

  if(!AGS_IS_RECALL_ID(recall_id)){
    return;
  }
  
  /* get recall id mutex */
  recall_id_mutex = AGS_RECALL_ID_GET_OBJ_MUTEX(recall_id);

  /* get some fields */
  pthread_mutex_lock(recall_id_mutex);

  recycling_context = recall_id->recycling_context;

  sound_scope = recall_id->sound_scope;

  pthread_mutex_unlock(recall_id_mutex);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get staging flags */
  pthread_mutex_lock(audio_mutex);

  current_staging_flags = audio->staging_flags[sound_scope];
  
  pthread_mutex_unlock(audio_mutex);

  if(!AGS_IS_RECYCLING_CONTEXT(recycling_context) ||
     (AGS_SOUND_STAGING_RUN_INIT_PRE & (current_staging_flags)) == 0 ||
     (AGS_SOUND_STAGING_RUN_INIT_INTER & (current_staging_flags)) == 0 ||
     (AGS_SOUND_STAGING_RUN_INIT_POST & (current_staging_flags)) == 0){
    return;
  }

  /* get recycling context mutex */
  recycling_context_mutex = AGS_RECYCLING_CONTEXT_GET_OBJ_MUTEX(recycling_context);

  /* get parent recycling context */
  pthread_mutex_lock(recycling_context_mutex);

  parent_recycling_context = recycling_context->parent;

  pthread_mutex_unlock(recycling_context_mutex);

  /* get the appropriate lists */
  if(parent_recycling_context == NULL){
    pthread_mutex_t *play_mutex;

    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* copy play context */
    pthread_mutex_lock(play_mutex);

    list = 
      list_start = g_list_copy(audio->play);

    pthread_mutex_unlock(play_mutex);

    /* reverse play context */
    list =
      list_start = g_list_reverse(list_start);
  }else{
    pthread_mutex_t *recall_mutex;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* copy recall context */
    pthread_mutex_lock(recall_mutex);

    list = 
      list_start = g_list_copy(audio->recall);
    
    pthread_mutex_unlock(recall_mutex);

    /* reverse recall context */
    list =
      list_start = g_list_reverse(list_start);
  }

  /* done  */  
  while((list = ags_recall_find_recycling_context(list,
						  (GObject *) recycling_context)) != NULL){
    recall = AGS_RECALL(list->data);
    
    /* done stages */
    ags_recall_set_staging_flags(recall,
				 staging_flags);

    list = list->next;
  }
  
  g_list_free(list_start);

  ags_audio_set_staging_flags(audio, sound_scope,
			      staging_flags);
}

/**
 * ags_audio_done_recall:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID to apply to
 *
 * Done processing audio data.
 *
 * Since: 2.0.0
 */
void
ags_audio_done_recall(AgsAudio *audio,
		      AgsRecallID *recall_id)
{
  g_return_if_fail(AGS_IS_AUDIO(audio));

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[DONE_RECALL], 0,
		recall_id);
  g_object_unref((GObject *) audio);
}

void
ags_audio_real_cancel_recall(AgsAudio *audio,
			     AgsRecallID *recall_id)
{
  AgsRecall *recall;
  AgsRecyclingContext *parent_recycling_context, *recycling_context;
  
  GList *list_start, *list;

  guint sound_scope;
  guint current_staging_flags;
  static const guint staging_flags = (AGS_SOUND_STAGING_CANCEL);

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *recall_id_mutex;
  pthread_mutex_t *recycling_context_mutex;

  if(!AGS_IS_RECALL_ID(recall_id)){
    return;
  }
  
  /* get recall id mutex */
  recall_id_mutex = AGS_RECALL_ID_GET_OBJ_MUTEX(recall_id);

  /* get some fields */
  pthread_mutex_lock(recall_id_mutex);

  recycling_context = recall_id->recycling_context;

  sound_scope = recall_id->sound_scope;

  pthread_mutex_unlock(recall_id_mutex);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get staging flags */
  pthread_mutex_lock(audio_mutex);

  current_staging_flags = audio->staging_flags[sound_scope];
  
  pthread_mutex_unlock(audio_mutex);

  if(!AGS_IS_RECYCLING_CONTEXT(recycling_context) ||
     (AGS_SOUND_STAGING_RUN_INIT_PRE & (current_staging_flags)) == 0 ||
     (AGS_SOUND_STAGING_RUN_INIT_INTER & (current_staging_flags)) == 0 ||
     (AGS_SOUND_STAGING_RUN_INIT_POST & (current_staging_flags)) == 0){
    return;
  }

  /* get recycling context mutex */
  recycling_context_mutex = AGS_RECYCLING_CONTEXT_GET_OBJ_MUTEX(recycling_context);

  /* get parent recycling context */
  pthread_mutex_lock(recycling_context_mutex);

  parent_recycling_context = recycling_context->parent;

  pthread_mutex_unlock(recycling_context_mutex);

  /* get the appropriate lists */
  if(parent_recycling_context == NULL){
    pthread_mutex_t *play_mutex;

    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* copy play context */
    pthread_mutex_lock(play_mutex);

    list = 
      list_start = g_list_copy(audio->play);

    pthread_mutex_unlock(play_mutex);

    /* reverse play context */
    list =
      list_start = g_list_reverse(list_start);
  }else{
    pthread_mutex_t *recall_mutex;

    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* copy recall context */
    pthread_mutex_lock(recall_mutex);

    list = 
      list_start = g_list_copy(audio->recall);
    
    pthread_mutex_unlock(recall_mutex);

    /* reverse recall context */
    list =
      list_start = g_list_reverse(list_start);
  }

  /* cancel  */
  while((list = ags_recall_find_recycling_context(list,
						  (GObject *) recycling_context)) != NULL){
    recall = AGS_RECALL(list->data);
    
    /* cancel stages */
    ags_recall_set_staging_flags(recall,
				 staging_flags);

    list = list->next;
  }

  g_list_free(list_start);

  ags_audio_set_staging_flags(audio, sound_scope,
			      staging_flags);
}

/**
 * ags_audio_cancel_recall:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID to apply to
 *
 * Cancel processing audio data.
 *
 * Since: 2.0.0
 */
void
ags_audio_cancel_recall(AgsAudio *audio,
			AgsRecallID *recall_id)
{
  g_return_if_fail(AGS_IS_AUDIO(audio));

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[CANCEL_RECALL], 0,
		recall_id);
  g_object_unref((GObject *) audio);
}

void
ags_audio_real_cleanup_recall(AgsAudio *audio,
			      AgsRecallID *recall_id)
{
  AgsRecall *recall;
  AgsRecyclingContext *parent_recycling_context, *recycling_context;
  
  GList *list_start, *list;
  GList *match_start, *match;

  gint sound_scope;
  guint current_staging_flags;
  gboolean play_context;
  gboolean remove_recycling_context;
  
  static const guint staging_mask = (AGS_SOUND_STAGING_CHECK_RT_DATA |
				     AGS_SOUND_STAGING_RUN_INIT_PRE |
				     AGS_SOUND_STAGING_RUN_INIT_INTER |
				     AGS_SOUND_STAGING_RUN_INIT_POST);

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *recall_id_mutex;
  pthread_mutex_t *recycling_context_mutex;

  if(!AGS_IS_RECALL_ID(recall_id)){
    return;
  }
  
  /* get recall id mutex */
  recall_id_mutex = AGS_RECALL_ID_GET_OBJ_MUTEX(recall_id);

  /* get some fields */
  pthread_mutex_lock(recall_id_mutex);

  recycling_context = recall_id->recycling_context;

  sound_scope = recall_id->sound_scope;

  pthread_mutex_unlock(recall_id_mutex);

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get staging flags */
  pthread_mutex_lock(audio_mutex);

  current_staging_flags = audio->staging_flags[sound_scope];
  
  pthread_mutex_unlock(audio_mutex);

  /* get recycling context mutex */
  recycling_context_mutex = AGS_RECYCLING_CONTEXT_GET_OBJ_MUTEX(recycling_context);

  /* get parent recycling context */
  pthread_mutex_lock(recycling_context_mutex);

  parent_recycling_context = recycling_context->parent;

  remove_recycling_context = FALSE;
  
  if((AgsRecallID *) recycling_context->recall_id == recall_id){
    remove_recycling_context = TRUE;
  }
  
  pthread_mutex_unlock(recycling_context_mutex);

  /* get the appropriate lists */
  if(parent_recycling_context == NULL){
    pthread_mutex_t *play_mutex;

    play_context = TRUE;
    
    /* get play mutex */
    play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* copy play context */
    pthread_mutex_lock(play_mutex);

    list = 
      list_start = g_list_copy(audio->play);

    pthread_mutex_unlock(play_mutex);

    /* reverse play context */
    list =
      list_start = g_list_reverse(list_start);
  }else{
    pthread_mutex_t *recall_mutex;

    play_context = FALSE;
    
    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* copy recall context */
    pthread_mutex_lock(recall_mutex);

    list = 
      list_start = g_list_copy(audio->recall);
    
    pthread_mutex_unlock(recall_mutex);

    /* reverse recall context */
    list =
      list_start = g_list_reverse(list_start);
  }

  /* cleanup  */
  match_start = NULL;

  while((list = ags_recall_find_recycling_context(list,
						  (GObject *) recycling_context)) != NULL){
    recall = AGS_RECALL(list->data);
    
    /* remove recall */
    ags_audio_remove_recall(audio, (GObject *) recall, play_context);
    match_start = g_list_prepend(match_start,
				 recall);

    list = list->next;
  }

  g_list_free(list_start);

  /* destroy  */
  match = match_start;
  
  while(match != NULL){
    recall = AGS_RECALL(match->data);

    /* destroy */
    ags_connectable_disconnect(AGS_CONNECTABLE(recall));
    g_object_run_dispose((GObject *) recall);
    g_object_unref((GObject *) recall);
    
    match = match->next;
  }

  g_list_free(match_start);
}

/**
 * ags_audio_cleanup_recall:
 * @audio: the #AgsAudio
 * @recall_id: the #AgsRecallID to apply to
 *
 * Cleanup processing audio data.
 *
 * Since: 2.0.0
 */
void
ags_audio_cleanup_recall(AgsAudio *audio,
			 AgsRecallID *recall_id)
{
  g_return_if_fail(AGS_IS_AUDIO(audio));

  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[CLEANUP_RECALL], 0,
		recall_id);
  g_object_unref((GObject *) audio);
}

void
ags_audio_recall_done_callback(AgsRecall *recall,
			       AgsAudio *audio)
{
  GList *start_recall_id;
  
  gint sound_scope;
  
  if(AGS_IS_COUNT_BEATS_AUDIO_RUN(recall)){
    sound_scope = ags_recall_get_sound_scope(recall);

    start_recall_id = ags_audio_check_scope(audio, sound_scope);
    ags_audio_stop(audio,
		   start_recall_id, sound_scope);

    g_list_free_full(start_recall_id,
		     g_object_unref);
  }
}

GList*
ags_audio_real_start(AgsAudio *audio,
		     gint sound_scope)
{
  AgsChannel *channel;
  AgsRecycling *recycling;
  AgsPlayback *playback;
  AgsRecallID *current_recall_id;
  AgsRecyclingContext *current_recycling_context;
  
  AgsMessageDelivery *message_delivery;
  AgsMessageQueue *message_queue;

  GList *recall_id;

  guint audio_flags;
  guint audio_channels;
  guint output_pads;
  guint i;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;
  pthread_mutex_t *playback_mutex;

  if(sound_scope < 0 ||
     sound_scope >= AGS_SOUND_SCOPE_LAST){
    return(NULL);
  }
  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get some fields */
  pthread_mutex_lock(audio_mutex);

  channel = audio->output;

  audio_flags = audio->flags;
  
  pthread_mutex_unlock(audio_mutex);

  if((AGS_AUDIO_OUTPUT_HAS_RECYCLING & (audio_flags)) == 0){
    return(NULL);
  }
  
  /* initialize channel */
  recall_id = NULL;

  while(channel != NULL){
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* recycling context */
    current_recycling_context = ags_recycling_context_new(1);
    ags_audio_add_recycling_context(audio,
				    (GObject *) current_recycling_context);

    /* get recycling */
    pthread_mutex_lock(channel_mutex);

    recycling = channel->first_recycling;
    
    pthread_mutex_unlock(channel_mutex);

    /* set recycling */
    ags_recycling_context_replace(current_recycling_context,
				  recycling,
				  0);

    /* create recall id */
    current_recall_id = g_object_new(AGS_TYPE_RECALL_ID,
				     "recycling-context", current_recycling_context,
				     NULL);
    ags_recall_id_set_sound_scope(current_recall_id,  sound_scope);
    ags_audio_add_recall_id(audio,
			    (GObject *) current_recall_id);
    
    recall_id = g_list_prepend(recall_id,
			       current_recall_id);
        
    /* get playback */
    pthread_mutex_lock(channel_mutex);

    playback = (AgsPlayback *) channel->playback;
    
    pthread_mutex_unlock(channel_mutex);

    /* get playback mutex */
    playback_mutex = AGS_PLAYBACK_GET_OBJ_MUTEX(playback);
    
    /* set playback's recall id */
    pthread_mutex_lock(playback_mutex);
    
    playback->recall_id[sound_scope] = current_recall_id;

    pthread_mutex_unlock(playback_mutex);
    
    /* reset recall id */
    ags_channel_recursive_run_stage(channel,
				    sound_scope, (AGS_SOUND_STAGING_CHECK_RT_DATA |
						  AGS_SOUND_STAGING_RUN_INIT_PRE |
						  AGS_SOUND_STAGING_RUN_INIT_INTER |
						  AGS_SOUND_STAGING_RUN_INIT_POST));

    /* iterate */
    pthread_mutex_lock(channel_mutex);
    
    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }
  
  recall_id = g_list_reverse(recall_id);
  
  /* emit message */
  message_delivery = ags_message_delivery_get_instance();

  message_queue = (AgsMessageQueue *) ags_message_delivery_find_namespace(message_delivery,
									  "libags-audio");

  if(message_queue != NULL){
    AgsMessageEnvelope *message;

    xmlDoc *doc;
    xmlNode *root_node;

    /* specify message body */
    doc = xmlNewDoc("1.0");

    root_node = xmlNewNode(NULL,
			   "ags-command");
    xmlDocSetRootElement(doc, root_node);    

    xmlNewProp(root_node,
	       "method",
	       "AgsAudio::start");

    /* add message */
    message = ags_message_envelope_alloc((GObject *) audio,
					 NULL,
					 doc);

    /* set parameter */
    message->n_params = 2;

    message->parameter_name = (gchar **) malloc(3 * sizeof(gchar *));
    message->value = g_new0(GValue,
			    2);

    /* recall id */
    message->parameter_name[0] = "recall-id";
    
    g_value_init(&(message->value[0]),
		 G_TYPE_POINTER);
    g_value_set_pointer(&(message->value[0]),
			recall_id);

    /* sound scope */
    message->parameter_name[1] = "sound-scope";
    
    g_value_init(&(message->value[1]),
		 G_TYPE_UINT);
    g_value_set_uint(&(message->value[1]),
		     sound_scope);

    /* terminate string vector */
    message->parameter_name[2] = NULL;
    
    /* add message */
    ags_message_delivery_add_message(message_delivery,
				     "libags-audio",
				     message);
  }

  return(recall_id);
}

/**
 * ags_audio_start:
 * @audio: the #AgsAudio
 * @sound_scope: the scope
 *
 * Is emitted as audio is started.
 *
 * Returns: the #GList-struct containing #AgsRecallID
 * 
 * Since: 2.0.0
 */
GList*
ags_audio_start(AgsAudio *audio,
		gint sound_scope)
{
  GList *list;
  
  g_return_val_if_fail(AGS_IS_AUDIO(audio), NULL);
  
  /* emit */
  list = NULL;
  
  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[START], 0,
		sound_scope,
		&list);
  g_object_unref((GObject *) audio);

  return(list);
}

void
ags_audio_real_stop(AgsAudio *audio,
		    GList *recall_id, gint sound_scope)
{
  AgsChannel *channel;
  AgsPlayback *playback;

  AgsMessageDelivery *message_delivery;
  AgsMessageQueue *message_queue;
  
  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;
  pthread_mutex_t *playback_mutex;

  if(recall_id == NULL ||
     sound_scope < 0 ||
     sound_scope >= AGS_SOUND_SCOPE_LAST){
    return;
  }
  
#if 0  
  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get some fields */
  pthread_mutex_lock(audio_mutex);

  channel = audio->output;
  
  pthread_mutex_unlock(audio_mutex);
  
  while(channel != NULL){
    AgsRecallID *current_recall_id;
    
    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* get current recall id */
    pthread_mutex_lock(channel_mutex);
    
    playback = (AgsPlayback *) channel->playback;

    pthread_mutex_unlock(channel_mutex);

    /* get playback mutex */
    playback_mutex = AGS_PLAYBACK_GET_OBJ_MUTEX(playback);

    /* recursive cancel */
    ags_channel_recursive_run_stage(channel,
				    sound_scope, (AGS_SOUND_STAGING_CANCEL |
						  AGS_SOUND_STAGING_REMOVE));

    /* get/set recall id */
    pthread_mutex_lock(playback_mutex);
    
    current_recall_id = playback->recall_id[sound_scope];
    
    playback->recall_id[sound_scope] = NULL;

    pthread_mutex_unlock(playback_mutex);

    /* iterate */
    pthread_mutex_lock(channel_mutex);    
    
    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }
#endif
  
  /* emit message */
  message_delivery = ags_message_delivery_get_instance();

  message_queue = (AgsMessageQueue *) ags_message_delivery_find_namespace(message_delivery,
									  "libags-audio");

  if(message_queue != NULL){
    AgsMessageEnvelope *message;

    xmlDoc *doc;
    xmlNode *root_node;

    /* specify message body */
    doc = xmlNewDoc("1.0");

    root_node = xmlNewNode(NULL,
			   "ags-command");
    xmlDocSetRootElement(doc, root_node);    

    xmlNewProp(root_node,
	       "method",
	       "AgsAudio::stop");

    /* add message */
    message = ags_message_envelope_alloc((GObject *) audio,
					 NULL,
					 doc);

    /* set parameter */
    message->n_params = 2;

    message->parameter_name = (gchar **) malloc(3 * sizeof(gchar *));
    message->value = g_new0(GValue,
			    2);

    /* recall id */
    message->parameter_name[0] = "recall-id";
    
    g_value_init(&(message->value[0]),
		 G_TYPE_POINTER);
    g_value_set_pointer(&(message->value[0]),
			recall_id);

    /* sound scope */
    message->parameter_name[1] = "sound-scope";
    
    g_value_init(&(message->value[1]),
		 G_TYPE_INT);
    g_value_set_int(&(message->value[1]),
		    sound_scope);

    /* terminate string vector */
    message->parameter_name[2] = NULL;

    /* add message */
    ags_message_delivery_add_message(message_delivery,
				     "libags-audio",
				     message);
  }
}

/**
 * ags_audio_stop:
 * @audio: the #AgsAudio
 * @recall_id: the current #AgsRecallID
 * @sound_scope: the scope
 *
 * Is emitted as playing audio is stopped.
 *
 * Since: 2.0.0
 */
void
ags_audio_stop(AgsAudio *audio,
	       GList *recall_id, gint sound_scope)
{
  g_return_if_fail(AGS_IS_AUDIO(audio));
  
  /* emit */
  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[STOP], 0,
		recall_id, sound_scope);
  g_object_unref((GObject *) audio);
}

GList*
ags_audio_real_check_scope(AgsAudio *audio, gint sound_scope)
{
  GList *list_start, *list;
  GList *recall_id;

  gint i;

  /* get recall id */
  g_object_get(audio,
	       "recall-id", &list_start,
	       NULL);

  /* iterate recall id */
  list = list_start;

  recall_id = NULL;

  if(sound_scope >= 0){
    while(list != NULL){
      /* check sound scope */
      if(ags_recall_id_check_sound_scope(list->data, sound_scope)){
	recall_id = g_list_prepend(recall_id,
				   list->data);
      }

      /* iterate */
      list = list->next;
    }
  }else{
    for(i = 0; i < AGS_SOUND_SCOPE_LAST; i++){
      list = list_start;

      while(list != NULL){
	/* check sound scope */
	if(ags_recall_id_check_sound_scope(list->data, i)){
	  recall_id = g_list_prepend(recall_id,
				     list->data);
	}

	/* iterate */
	list = list->next;
      }
    }
  }

  /* reverse recall id */
  recall_id = g_list_reverse(recall_id);
  g_list_foreach(recall_id,
		 (GFunc) g_object_ref,
		 NULL);

  /* unref */
  g_list_free_full(list_start,
		   g_object_unref);
  
  return(recall_id);
}

/**
 * ags_audio_check_scope:
 * @audio: the #AgsAudio
 * @sound_scope: the scope
 *
 * Check scope's recall id.
 * 
 * Returns: the scope's recall id of @audio
 * 
 * Since: 2.0.0
 */
GList*
ags_audio_check_scope(AgsAudio *audio, gint sound_scope)
{
  GList *recall_id;

  g_return_val_if_fail(AGS_IS_AUDIO(audio),
		       NULL);
  
  /* emit */
  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[CHECK_SCOPE], 0,
		sound_scope,
		&recall_id);
  g_object_unref((GObject *) audio);

  return(recall_id);
}

/**
 * ags_audio_collect_all_audio_ports:
 * @audio: the #AgsAudio
 *
 * Retrieve all ports of #AgsAudio.
 *
 * Returns: a new #GList containing #AgsPort
 *
 * Since: 2.0.0
 */
GList*
ags_audio_collect_all_audio_ports(AgsAudio *audio)
{
  GList *recall_start, *recall;
  GList *list;

  pthread_mutex_t *recall_mutex, *play_mutex;
  pthread_mutex_t *mutex;

  if(!AGS_IS_AUDIO(audio)){
    return(NULL);
  }

  list = NULL;
 
  /* get play mutex */  
  play_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

  /* collect port of playing recall */
  pthread_mutex_lock(play_mutex);

  recall =
    recall_start = g_list_copy_deep(audio->play,
				    (GCopyFunc) g_object_ref,
				    NULL);

  pthread_mutex_unlock(play_mutex);
   
  while(recall != NULL){
    AgsRecall *current;

    current = AGS_RECALL(recall->data);
    
    /* get mutex */  
    mutex = AGS_RECALL_GET_OBJ_MUTEX(current);

    /* concat port */
    pthread_mutex_lock(mutex);
    
    if(current->port != NULL){
      if(list == NULL){
	list = g_list_copy_deep(current->port,
				(GCopyFunc) g_object_ref,
				NULL);
      }else{
	if(current->port != NULL){
	  list = g_list_concat(list,
			       g_list_copy_deep(current->port,
						(GCopyFunc) g_object_ref,
						NULL));
	}
      }
    }

    pthread_mutex_unlock(mutex);

    /* iterate */
    recall = recall->next;
  }

  g_list_free_full(recall_start,
		   g_object_unref);
  
  /* get recall mutex */  
  recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

  /* the same for true recall */
  pthread_mutex_lock(recall_mutex);

  recall =
    recall_start = g_list_copy_deep(audio->recall,
				    (GCopyFunc) g_object_ref,
				    NULL);

  pthread_mutex_unlock(recall_mutex);
   
  while(recall != NULL){
    AgsRecall *current;

    current = AGS_RECALL(recall->data);
    
    /* get mutex */  
    mutex = AGS_RECALL_GET_OBJ_MUTEX(current);

    /* concat port */
    pthread_mutex_lock(mutex);
    
    if(current->port != NULL){
      if(list == NULL){
	list = g_list_copy_deep(current->port,
				(GCopyFunc) g_object_ref,
				NULL);
      }else{
	if(current->port != NULL){
	  list = g_list_concat(list,
			       g_list_copy_deep(current->port,
						(GCopyFunc) g_object_ref,
						NULL));
	}
      }
    }

    pthread_mutex_unlock(mutex);

    /* iterate */
    recall = recall->next;
  }
   
  g_list_free_full(recall_start,
		   g_object_unref);

  /* reverse result */
  list = g_list_reverse(list);
  
  return(list);
}

/**
 * ags_audio_collect_all_audio_ports_by_specifier_and_scope:
 * @audio: an #AgsAudio
 * @specifier: the port's name
 * @play_context: either %TRUE for play or %FALSE for recall
 *
 * Retrieve specified port of #AgsAudio
 *
 * Returns: a #GList-struct of #AgsPort if found, otherwise %NULL
 *
 * Since: 2.0.0
 */
GList*
ags_audio_collect_all_audio_ports_by_specifier_and_context(AgsAudio *audio,
							   gchar *specifier,
							   gboolean play_context)
{
  GList *recall_start, *recall;
  GList *port_start, *port;
  GList *list;
  
  pthread_mutex_t *recall_mutex;

  if(!AGS_IS_AUDIO(audio)){
    return(NULL);
  }
  
  if(play_context){
    /* get play mutex */
    recall_mutex = AGS_AUDIO_GET_PLAY_MUTEX(audio);

    /* get recall */
    pthread_mutex_lock(recall_mutex);

    recall =
      recall_start = g_list_copy_deep(audio->play,
				      (GCopyFunc) g_object_ref,
				      NULL);
    
    pthread_mutex_unlock(recall_mutex);
  }else{
    /* get recall mutex */
    recall_mutex = AGS_AUDIO_GET_RECALL_MUTEX(audio);

    /* get recall */
    pthread_mutex_lock(recall_mutex);

    recall =
      recall_start = g_list_copy_deep(audio->recall,
				      (GCopyFunc) g_object_ref,
				      NULL);
    
    pthread_mutex_unlock(recall_mutex);
  }
  
  /* collect port of playing recall */
  list = NULL;

  while(recall != NULL){
    AgsRecall *current;

    pthread_mutex_t *mutex;

    current = AGS_RECALL(recall->data);
    
    /* get mutex */  
    mutex = AGS_RECALL_GET_OBJ_MUTEX(current);

    /* get port */
    pthread_mutex_lock(mutex);
    
    port =
      port_start = g_list_copy_deep(current->port,
				    (GCopyFunc) g_object_ref,
				    NULL);

    pthread_mutex_unlock(mutex);

    /* check specifier */
    while((port = ags_port_find_specifier(port, specifier)) != NULL){
      AgsPort *current;

      current = AGS_PORT(port->data);

      g_object_ref(current);
      list = g_list_prepend(list,
			    current);

      /* iterate - port */
      port = port->next;
    }

    g_list_free_full(port_start,
		     g_object_unref);

    /* iterate - recall */
    recall = recall->next;
  }

  g_list_free_full(recall_start,
		   g_object_unref);

  /* reverse result */
  list = g_list_reverse(list);
  
  return(list);
}

/**
 * ags_audio_open_audio_file_as_channel:
 * @audio: the #AgsAudio
 * @filename: the files to open
 * @overwrite_channels: if existing channels should be assigned
 * @create_channels: if new channels should be created as not fitting if combined with @overwrite_channels
 *
 * Open some files.
 *
 * Since: 2.0.0
 */
void
ags_audio_open_audio_file_as_channel(AgsAudio *audio,
				     GSList *filename,
				     gboolean overwrite_channels,
				     gboolean create_channels)
{
  AgsChannel *channel;
  AgsAudioFile *audio_file;

  GObject *soundcard;
  
  GList *audio_signal;

  guint input_pads;
  guint audio_channels;
  guint i, j;
  guint list_length;

  GError *error;

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;
  pthread_mutex_t *recycling_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     filename == NULL ||
     (!overwrite_channels &&
      !create_channels)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get audio fields */
  pthread_mutex_lock(audio_mutex);

  channel = audio->input;
  soundcard = audio->output_soundcard;
  
  input_pads = audio->input_pads;
  audio_channels = audio->audio_channels;
  
  pthread_mutex_unlock(audio_mutex);
  
  /* overwriting existing channels */
  if(overwrite_channels){
    if(channel != NULL){
      for(i = 0; i < input_pads && filename != NULL; i++){
	audio_file = ags_audio_file_new((gchar *) filename->data,
					soundcard,
					-1);
	
	if(!ags_audio_file_open(audio_file)){
	  filename = filename->next;
	  
	  continue;
	}

	ags_audio_file_read_audio_signal(audio_file);
	ags_audio_file_close(audio_file);
	
	audio_signal = audio_file->audio_signal;
	
	for(j = 0; j < audio_channels && audio_signal != NULL; j++){
	  AgsRecycling *recycling;

	  AgsFileLink *file_link;
	  
	  /* reset link */
	  error = NULL;

	  ags_channel_set_link(channel, NULL,
			       &error);

	  if(error != NULL){
	    g_warning("%s", error->message);


	    g_error_free(error);
	  }

	  /* get channel mutex */
	  channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

	  /* get recycling */
	  pthread_mutex_lock(channel_mutex);

	  recycling = channel->first_recycling;

	  file_link = AGS_INPUT(channel)->file_link;
	  
	  pthread_mutex_unlock(channel_mutex);

	  /* set filename and channel */
	  if(file_link == NULL){
	    file_link = g_object_new(AGS_TYPE_AUDIO_FILE_LINK,
				     NULL);
      
	    g_object_set(channel,
			 "file-link", file_link,
			 NULL);

	    g_object_unref(file_link);
	  }
	  
	  g_object_set(file_link,
		       "filename", filename->data,
		       "audio-channel", j,
		       NULL);
	  
	  /* get recycling mutex */
	  recycling_mutex = AGS_RECYCLING_GET_OBJ_MUTEX(recycling);

	  /* replace template audio signal */
	  AGS_AUDIO_SIGNAL(audio_signal->data)->flags |= AGS_AUDIO_SIGNAL_TEMPLATE;
	  g_object_set(AGS_AUDIO_SIGNAL(audio_signal->data),
		       "recycling", recycling,
		       NULL);

	  ags_recycling_add_audio_signal(recycling,
					 audio_signal->data);

	  /* iterate */
	  audio_signal = audio_signal->next;

	  pthread_mutex_lock(channel_mutex);

	  channel = channel->next;

	  pthread_mutex_unlock(channel_mutex);
	}

	if(audio_file->file_audio_channels < audio_channels){
	  channel = ags_channel_nth(channel,
				    audio_channels - audio_file->file_audio_channels);
	  g_object_unref(channel);
	}

	g_object_run_dispose(audio_file);
	g_object_unref(audio_file);
	
	filename = filename->next;
      }
    }
  }

  /* appending to channels */
  if(create_channels && filename != NULL){
    list_length = g_slist_length(filename);
    
    ags_audio_set_pads((AgsAudio *) audio, AGS_TYPE_INPUT,
		       list_length + AGS_AUDIO(audio)->input_pads, 0);
    
    channel = ags_channel_nth(AGS_AUDIO(audio)->input,
			      (AGS_AUDIO(audio)->input_pads - list_length) * AGS_AUDIO(audio)->audio_channels);
    g_object_unref(channel);
    
    while(filename != NULL){
      audio_file = ags_audio_file_new((gchar *) filename->data,
				      soundcard,
				      -1);
      
      if(!ags_audio_file_open(audio_file)){
	filename = filename->next;
	continue;
      }
      
      ags_audio_file_read_audio_signal(audio_file);
      ags_audio_file_close(audio_file);
	
      audio_signal = audio_file->audio_signal;
      
      for(j = 0; j < audio_channels && audio_signal != NULL; j++){
	AgsRecycling *recycling;
	
	AgsFileLink *file_link;
	
	/* get channel mutex */
	channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

	/* get recycling */
	pthread_mutex_lock(channel_mutex);

	recycling = channel->first_recycling;

	file_link = AGS_INPUT(channel)->file_link;
	
	pthread_mutex_unlock(channel_mutex);

	/* set filename and channel */
	if(file_link == NULL){
	  file_link = g_object_new(AGS_TYPE_AUDIO_FILE_LINK,
				   NULL);
      
	  g_object_set(channel,
		       "file-link", file_link,
		       NULL);

	  g_object_unref(file_link);
	}
	
	g_object_set(file_link,
		     "filename", filename->data,
		     "audio-channel", j,
		     NULL);

	/* get recycling mutex */
	recycling_mutex = AGS_RECYCLING_GET_OBJ_MUTEX(recycling);

	/* replace template audio signal */
	pthread_mutex_lock(recycling_mutex);
	  
	AGS_AUDIO_SIGNAL(audio_signal->data)->flags |= AGS_AUDIO_SIGNAL_TEMPLATE;
	AGS_AUDIO_SIGNAL(audio_signal->data)->recycling = (GObject *) recycling;
	
	ags_recycling_add_audio_signal(recycling,
				       audio_signal->data);
	
	pthread_mutex_unlock(recycling_mutex);

	/* iterate */
	audio_signal = audio_signal->next;

	pthread_mutex_lock(channel_mutex);

	channel = channel->next;

	pthread_mutex_unlock(channel_mutex);
      }
      
      if(audio_channels > audio_file->file_audio_channels){
	channel = ags_channel_nth(channel,
				  audio_channels - audio_file->file_audio_channels);
	g_object_unref(channel);
      }

      g_object_run_dispose(audio_file);
      g_object_unref(audio_file);
      
      filename = filename->next;
    }
  }
}

void
ags_audio_open_audio_file_as_wave(AgsAudio *audio,
				  const gchar *filename,
				  gboolean overwrite_channels,
				  gboolean create_channels)
{
  AgsChannel *channel;
  AgsAudioFile *audio_file;

  GObject *soundcard;
  
  GList *wave;

  guint audio_channels;

  GError *error;

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;

  if(!AGS_IS_AUDIO(audio) ||
     filename == NULL ||
     (!overwrite_channels &&
      !create_channels)){
    return;
  }

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get audio fields */
  pthread_mutex_lock(audio_mutex);

  audio_channels = audio->audio_channels;

  soundcard = audio->output_soundcard;

  pthread_mutex_unlock(audio_mutex);

  /* open audio file */
  audio_file = ags_audio_file_new(filename,
				  soundcard,
				  -1);

  if(!ags_audio_file_open(audio_file)){
    return;
  }

  ags_audio_file_read_wave(audio_file,
			   0,
			   0.0, 0);
  ags_audio_file_close(audio_file);

  //TODO:JK: implement me
}

void
ags_audio_open_midi_file_as_midi(AgsAudio *audio,
				 const gchar *filename,
				 const gchar *instrument,
				 const gchar *track_name,
				 guint midi_channel)
{
  //TODO:JK: implement me
}

void
ags_audio_open_midi_file_as_notation(AgsAudio *audio,
				     const gchar *filename,
				     const gchar *instrument,
				     const gchar *track_name,
				     guint midi_channel)
{
  //TODO:JK: implement me
}

/**
 * ags_audio_recursive_set_property:
 * @audio: the #AgsAudio
 * @n_params: the count of elements in following arrays
 * @parameter_name: the parameter's name array
 * @value: the value array
 *
 * Recursive set property for #AgsAudio.
 *
 * Since: 2.0.0
 */
void
ags_audio_recursive_set_property(AgsAudio *audio,
				 gint n_params,
				 const gchar *parameter_name[], const GValue value[])
{
  AgsChannel *channel;
  
  auto void ags_audio_set_property(AgsAudio *audio,
				   gint n_params,
				   const gchar *parameter_name[], const GValue value[]);
  auto void ags_audio_recursive_set_property_down(AgsChannel *channel,
						  gint n_params,
						  const gchar *parameter_name[], const GValue value[]);
  auto void ags_audio_recursive_set_property_down_input(AgsChannel *channel,
							gint n_params,
							const gchar *parameter_name[], const GValue value[]);

  void ags_audio_set_property(AgsAudio *audio,
			      gint n_params,
			      const gchar *parameter_name[], const GValue value[]){
    guint i;

    for(i = 0; i < n_params; i++){
      g_object_set_property(G_OBJECT(audio),
			    parameter_name[i], &(value[i]));
    }
  }
  
  void ags_audio_recursive_set_property_down(AgsChannel *channel,
					     gint n_params,
					     const gchar *parameter_name[], const GValue value[]){
    if(channel == NULL){
      return;
    }

    ags_audio_set_property(AGS_AUDIO(channel->audio),
			   n_params,
			   parameter_name, value);
    
    ags_audio_recursive_set_property_down_input(channel,
						n_params,
						parameter_name, value);
  }
    
  void ags_audio_recursive_set_property_down_input(AgsChannel *channel,
						   gint n_params,
						   const gchar *parameter_name[], const GValue value[]){
    AgsAudio *audio;
    AgsChannel *input;
    
    if(channel == NULL){
      return;
    }

    audio = (AgsAudio *) channel->audio;

    if(audio == NULL){
      return;
    }
    
    input = ags_channel_nth(audio->input,
			    channel->audio_channel);
    g_object_unref(input);
    
    while(input != NULL){      
      ags_audio_recursive_set_property_down(input->link,
					    n_params,
					    parameter_name, value);

      input = input->next;
    }
  }

  if(audio == NULL){
    return;
  }

  ags_audio_set_property(audio,
			 n_params,
			 parameter_name, value);

  if(audio->input != NULL){
    channel = audio->input;
    
    while(channel != NULL){
      ags_audio_recursive_set_property_down(channel->link,
					    n_params,
					    parameter_name, value);

      channel = channel->next;
    }
  }
}

void
ags_audio_real_recursive_run_stage(AgsAudio *audio,
				   gint sound_scope, guint staging_flags)
{
  AgsChannel *channel;

  pthread_mutex_t *audio_mutex;
  pthread_mutex_t *channel_mutex;

  /* get audio mutex */
  audio_mutex = AGS_AUDIO_GET_OBJ_MUTEX(audio);

  /* get channel */
  pthread_mutex_lock(audio_mutex);

  channel = audio->output;

  pthread_mutex_unlock(audio_mutex);

  /* initialize return value */
  while(channel != NULL){
    /* reset recursive channel stage */
    ags_channel_recursive_run_stage(channel,
				    sound_scope, staging_flags);

    /* get channel mutex */
    channel_mutex = AGS_CHANNEL_GET_OBJ_MUTEX(channel);

    /* iterate */
    pthread_mutex_lock(channel_mutex);

    channel = channel->next;

    pthread_mutex_unlock(channel_mutex);
  }
}

/**
 * ags_audio_recursive_run_stage:
 * @audio: the #AgsAudio object
 * @sound_scope: the scope to reset
 * @staging_flags: the stage to enable
 *
 * Resets @audio's @sound_scope specified by @staging_flags.
 *
 * Since: 2.0.0
 */
void
ags_audio_recursive_run_stage(AgsAudio *audio,
			      gint sound_scope, guint staging_flags)
{  
  g_return_if_fail(AGS_IS_AUDIO(audio));
  
  /* emit */  
  g_object_ref((GObject *) audio);
  g_signal_emit(G_OBJECT(audio),
		audio_signals[RECURSIVE_RUN_STAGE], 0,
		sound_scope, staging_flags);
  g_object_unref((GObject *) audio);
}

/**
 * ags_audio_new:
 * @output_soundcard: the #AgsSoundcard to use for output
 *
 * Creates an #AgsAudio, with defaults of @soundcard.
 *
 * Returns: a new #AgsAudio
 *
 * Since: 2.0.0
 */
AgsAudio*
ags_audio_new(GObject *output_soundcard)
{
  AgsAudio *audio;

  audio = (AgsAudio *) g_object_new(AGS_TYPE_AUDIO,
				    "output-soundcard", output_soundcard,
				    NULL);

  return(audio);
}
